/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp D�hring, October 2012
 ***************************************************************************/

package importer;

import java.lang.management.ManagementFactory;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.analysis.plugin.filter.AbstractFilterPlugin;
import kieker.analysis.plugin.filter.flow.TraceEventRecords;
import kieker.common.configuration.Configuration;
import kieker.common.record.flow.trace.AbstractTraceEvent;
import kieker.common.record.flow.trace.ConstructionEvent;
import kieker.common.record.flow.trace.Trace;
import kieker.common.record.flow.trace.concurrency.JoinEvent;
import kieker.common.record.flow.trace.concurrency.SplitEvent;
import kieker.common.record.flow.trace.concurrency.monitor.MonitorEntryEvent;
import kieker.common.record.flow.trace.concurrency.monitor.MonitorExitEvent;
import kieker.common.record.flow.trace.concurrency.monitor.MonitorNotifyAllEvent;
import kieker.common.record.flow.trace.concurrency.monitor.MonitorNotifyEvent;
import kieker.common.record.flow.trace.concurrency.monitor.MonitorRequestEvent;
import kieker.common.record.flow.trace.concurrency.monitor.MonitorWaitEvent;
import kieker.common.record.flow.trace.operation.AbstractOperationEvent;
import kieker.common.record.flow.trace.operation.AfterOperationEvent;
import kieker.common.record.flow.trace.operation.AfterOperationFailedEvent;
import kieker.common.record.flow.trace.operation.BeforeOperationEvent;
import kieker.common.record.flow.trace.operation.CallOperationEvent;
import kieker.common.record.flow.trace.operation.constructor.AfterConstructorEvent;
import kieker.common.record.flow.trace.operation.constructor.AfterConstructorFailedEvent;
import kieker.common.record.flow.trace.operation.constructor.BeforeConstructorEvent;
import kieker.common.record.flow.trace.operation.constructor.CallConstructorEvent;
import kieker.common.record.flow.trace.operation.object.BeforeOperationObjectEvent;
import kieker.common.record.flow.trace.operation.object.CallOperationObjectEvent;

import data.dynamics.ConstructorExecution;
import data.dynamics.OperationExecution;
import data.dynamics.SClassInstance;
import data.dynamics.SMethodExecution;
import data.dynamics.STrace;
import data.dynamics.STraceEvent;
import data.dynamics.ThreadCreation;
import data.dynamics.ThreadTermination;
import data.dynamics.impl.DynamicsFactoryImpl;
import data.dynamics.impl.DynamicsPackageImpl;
import data.impl.DataFactoryImpl;
import data.impl.DataPackageImpl;
import data.statics.SClass;
import data.synchronisation.MonitorEvent;
import data.synchronisation.MonitorMethod;
import data.synchronisation.MonitorObject;
import data.synchronisation.ThreadJoin;
import data.synchronisation.WaitNotifyEvent;
import data.synchronisation.WaitNotifyMethod;
import data.synchronisation.WaitNotifyObject;
import data.synchronisation.impl.SynchronisationFactoryImpl;
import data.synchronisation.impl.SynchronisationPackageImpl;

import importer.exception.KiekerImportException;
import importer.utils.KiekerSignatureUtils;

import models.data.DataModel;

import utils.StringUtils;

/**
 * A filter to import dynamic event traces into the 'Synchrovis' application for visualization. All incoming event traces on the input port will be forwarded to the
 * output port without any changes or filtering.
 * 
 * @author pdo
 * 
 */
@Plugin(
		name = "Kieker import filter",
		description = "A filter to import Kieker event traces into the 'Synchrovis' application for visualization. "
				+ "All incoming event traces on the input port will be forwarded to the output port without any changes or filtering.",
		outputPorts = {
			@OutputPort(name = KiekerImportFilter.OUTPUT_PORT_NAME,
					description = "Outputs all incoming event traces unchanged and unfiltered.",
					eventTypes = { TraceEventRecords.class }) })
public final class KiekerImportFilter extends AbstractFilterPlugin {

	// =========================== Attributes =================================

	/** The name of the input port of this Kieker import filter. */
	public static final String INPUT_PORT_NAME = "inputTraces";

	/** The name of the output port of this Kieker import filter. */
	public static final String OUTPUT_PORT_NAME = "outputTraces";

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(KiekerImportFilter.class);

	/** The {@link DataModel} instance to write all imported Kieker event traces into. */
	private final DataModel model;

	/** A list of all trace event records received on the input port and being parsed to {@link STrace} instances. */
	private final List<STrace> traces = new LinkedList<STrace>();

	/** A mapping of all {@link STraceEvent} instances to their source and target object ids created while importing the Kieker trace event records. */
	private final Map<STraceEvent, Pair<Integer, Integer>> operationExecutions = new HashMap<STraceEvent, Pair<Integer, Integer>>();

	/** A list of all {@link ThreadCreation} instances created while importing the Kieker trace event records. */
	private final List<ThreadCreation> threadCreations = new LinkedList<ThreadCreation>();

	/** A mapping of all {@link ThreadJoin} instances created while importing the Kieker trace event records to the id of their joined traces. */
	private final Map<ThreadJoin, Long> sthreadJoins = new HashMap<ThreadJoin, Long>();

	/** A mapping of all {@link MonitorEvent} instances created while importing the Kieker trace event records to the object if of their target monitor object. */
	private final Map<MonitorEvent, Integer> monitorEvents = new HashMap<MonitorEvent, Integer>();

	/**
	 * A mapping of all {@link WaitNotifyEvent} instances created while importing the Kieker trace event records to the object id of their target wait-/notify
	 * object.
	 */
	private final Map<WaitNotifyEvent, Integer> waitNotifyEvents = new HashMap<WaitNotifyEvent, Integer>();

	/** A mapping from Kieker object Ids to their corresponding {@link WaitNotifyObject} instances. */
	private final Map<Integer, WaitNotifyObject> idToWaitObject = new HashMap<Integer, WaitNotifyObject>(20);

	/** A mapping from Kieker object Ids to their corresponding {@link MonitorObject} instances. */
	private final Map<Integer, MonitorObject> idToMonitorObject = new HashMap<Integer, MonitorObject>(20);

	/** A mapping from Kieker object Ids to their corresponding {@link VisualizableElement} instances. */
	private final Map<Integer, SClassInstance> idToSObject = new HashMap<Integer, SClassInstance>(200);

	/**
	 * The current operation executions just imported containing the currently parsed trace elements along with the Kieker class signature of the current method
	 * invoking object.
	 */
	private final List<Pair<OperationExecution, String>> stack = new LinkedList<Pair<OperationExecution, String>>();

	/** The list of {@link OperationExecution} instances for which its executing object is to be set after all Kieker files have been imported. */
	private final List<Pair<STraceEvent, ConstructorExecution>> execObjects = new LinkedList<Pair<STraceEvent, ConstructorExecution>>();

	// ========================== Constructor =================================

	/**
	 * Construct a new {@link KiekerImportFilter} instance to import dynamic event traces created by the Kieker Monitoring & Analysis Framework.
	 * 
	 * @param configuration
	 *            The configuration of this kieker importer.
	 * @param dataModel
	 *            The {@link DataModel} instance to add all created {@link SThread} objects to.
	 * @throws IllegalArgumentException
	 *             If the data model instance passed in is {@code null}.
	 */
	public KiekerImportFilter(final Configuration configuration, final DataModel dataModel) throws IllegalArgumentException {

		super(configuration);

		if (dataModel == null) {
			throw new IllegalArgumentException("You cannot import kieker traces into a null model.");
		}

		this.model = dataModel;

		DataPackageImpl.init();
		DataFactoryImpl.init();
		DynamicsPackageImpl.init();
		DynamicsFactoryImpl.init();
		SynchronisationPackageImpl.init();
		SynchronisationFactoryImpl.init();
	}

	// ============================ Methods ===================================

	@Override
	public boolean init() {

		LOGGER.info("Importing Kieker traces...");

		// no need to reset data structure, since a new filter instance will be created on any new import job
		return true;
	}

	@Override
	public void terminate(final boolean error) {

		super.terminate(error);

		// post-processing
		LOGGER.info("Post-processing data...");
		this.setThreadJoinTargets();
		this.setThreadCreationTargets();
		this.setExecSources();
		this.setOperationObjects();
		this.setMonitorObjects();
		this.setWaitNotifyObjects();
		LOGGER.info("Kieker import complete.\n");
		this.model.setDynamicsLoaded();

		LOGGER.info("Dynamics loaded. Currently " + ManagementFactory.getMemoryMXBean().getHeapMemoryUsage().getUsed()
				+ " bytes of heap memory in use.\n");
	}

	@Override
	public Configuration getCurrentConfiguration() {

		return super.configuration;
	}

	/**
	 * Handles incoming {@link TraceEventRecords} instances and imports them into the {@link DataModel} instance passed to the constructor of this Kieker import
	 * filter. This method will be invoked automatically whenever a new trace event record is being forwarded to this filter.
	 * 
	 * @param record
	 *            The trace event record to import.
	 */
	@InputPort(name = INPUT_PORT_NAME, eventTypes = { TraceEventRecords.class },
			description = "Receives incoming trace event records and imports their events into the 'Synchrovis' application for visualization.")
	public void inputRecord(final TraceEventRecords record) {

		final STrace trace = this.createSThread(record);
		for (AbstractTraceEvent event : record.getTraceEvents()) {

			final AbstractTraceEvent[] events = record.getTraceEvents();
			if (event.equals(events[0]) || event.equals(events[events.length - 1])) {
				// handle first and last event
				if (this.handleOuterEvent(event, trace)) {
					continue;
				}
			}

			// import event
			this.importEvent(event, trace);
		}

		if (this.stack.size() != 0) {
			LOGGER.error("Corrupt Trace Event Record detected. Execution stack size does not match. Value is '"
					+ this.stack.size() + "' but has to be '0' at record end. Skipping its import.");
			return;
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Imported new STrace with values: " + trace.toString());
		}

		// add termination event
		final ThreadTermination term = DynamicsFactoryImpl.eINSTANCE.createThreadTermination();
		term.setEss(1);
		term.setOrderIndex(Integer.MAX_VALUE);
		term.setSource(null);
		term.setTin(record.getTraceEvents()[record.getTraceEvents().length - 1].getTimestamp());
		term.setTrace(trace);
		trace.getEvents().add(term);

		// add trace to data model
		this.traces.add(trace);
		this.model.addSTrace(trace);
		this.stack.clear();

		super.deliver(OUTPUT_PORT_NAME, record);
	}

	/**
	 * Imports the {@link AbstractTraceEvent} instance passed in into the {@link DataModel} object of this kieker importer ({@link #model}).
	 * 
	 * @param event
	 *            The abstract trace event to import.
	 * @param trace
	 *            The trace the the specified trace event belongs to.
	 */
	private void importEvent(final AbstractTraceEvent event, final STrace trace) {

		if (event instanceof JoinEvent) {
			this.importJoinEvent((JoinEvent) event, trace);
		} else if (event instanceof SplitEvent) {
			this.importSplitEvent((SplitEvent) event, trace);
		} else if (event instanceof MonitorEntryEvent) {
			this.importMonitorEntryEvent((MonitorEntryEvent) event, trace);
		} else if (event instanceof MonitorRequestEvent) {
			this.importMonitorRequestEvent((MonitorRequestEvent) event, trace);
		} else if (event instanceof MonitorExitEvent) {
			this.importMonitorExitEvent((MonitorExitEvent) event, trace);
		} else if (event instanceof MonitorNotifyAllEvent) {
			this.importMonitorNotifyAllEvent((MonitorNotifyAllEvent) event, trace);
		} else if (event instanceof MonitorNotifyEvent) {
			this.importMonitorNotifyEvent((MonitorNotifyEvent) event, trace);
		} else if (event instanceof MonitorWaitEvent) {
			this.importMonitorWaitEvent((MonitorWaitEvent) event, trace);
		} else if (event instanceof CallConstructorEvent) {
			this.importCallConstructorEvent((CallConstructorEvent) event, trace);
		} else if (event instanceof BeforeConstructorEvent) {
			this.importBeforeConstructorEvent((BeforeConstructorEvent) event, trace);
		} else if (event instanceof ConstructionEvent) {
			this.importConstructionEvent((ConstructionEvent) event);
		} else if (event instanceof AfterConstructorEvent || event instanceof AfterConstructorFailedEvent) {
			this.importAfterConstructorEvent((AbstractOperationEvent) event, trace);
		} else if (event instanceof CallOperationObjectEvent) {
			this.importCallOperationObjectEvent((CallOperationObjectEvent) event, trace);
		} else if (event instanceof BeforeOperationObjectEvent) {
			this.importBeforeOperationObjectEvent((BeforeOperationObjectEvent) event, trace);
		} else if (event instanceof AfterOperationEvent || event instanceof AfterOperationFailedEvent) {
			this.importAfterOperationEvent((AbstractOperationEvent) event, trace);
		} else if (!(event instanceof CallOperationEvent || event instanceof BeforeOperationEvent)) {
			LOGGER.warn("Unknown abstract trace event (" + event.getClass().getName() + ") detected. Its values are: " + event.toString());
		}
	}

	/**
	 * Creates a new {@link STrace} instance representing the given {@link AbstractTraceEvent} object addable to the {@link DataModel} instance passed to
	 * the constructor of this filter.
	 * 
	 * @param record
	 *            The record to create the new execution trace from.
	 * @return A new execution trace representing the given trace.
	 */
	private STrace createSThread(final TraceEventRecords record) {

		final STrace sTrace = DynamicsFactoryImpl.eINSTANCE.createSTrace();
		final Trace trace = record.getTrace();
		sTrace.setTraceId(trace.getTraceId());
		sTrace.setParentTraceId(trace.getParentTraceId());
		sTrace.setParentOrderIndex(trace.getParentOrderId());
		sTrace.setHostname(trace.getHostname());
		sTrace.setSessionId(trace.getSessionId());
		sTrace.setThreadId(trace.getThreadId());

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Created new STrace with values: " + sTrace.toString());
		}

		return sTrace;
	}

	/**
	 * Handles the first (respectively last) event in the trace currently being imported. It sets the resulting {@link SThread} object's operation signature, as well
	 * as its entry and exit timestamp (tin and tout).
	 * 
	 * @param event
	 *            The first, respectively last event of the trace currently being imported.
	 * @param trace
	 *            The {@link STrace} instance currently being created representing the current event trace being imported.
	 * @return {@code True}, if the event passed in belongs either to the application's main or thread's run method, {@code false} otherwise.
	 */
	private boolean handleOuterEvent(final AbstractTraceEvent event, final STrace trace) {

		if (event instanceof CallOperationEvent) {
			// first trace element
			final CallOperationEvent ev = (CallOperationEvent) event;
			trace.setOperationSignature(KiekerSignatureUtils.toSimplifiedSignature(ev.getOperationSignature()));
			trace.setClassSignature(StringUtils.cutFromFirst(ev.getClassSignature(), '$'));
			trace.setTin(ev.getTimestamp());
		} else if (event instanceof BeforeOperationEvent) {
			// first trace element
			final BeforeOperationEvent ev = (BeforeOperationEvent) event;
			trace.setOperationSignature(KiekerSignatureUtils.toSimplifiedSignature(ev.getOperationSignature()));
			trace.setClassSignature(StringUtils.cutFromFirst(ev.getClassSignature(), '$'));
			trace.setTin(ev.getTimestamp());
		} else if (event instanceof AfterOperationEvent) {
			// last trace element
			final AfterOperationEvent ev = (AfterOperationEvent) event;
			trace.setTout(ev.getTimestamp());
			trace.setOperationSignature(KiekerSignatureUtils.toSimplifiedSignature(ev.getOperationSignature()));
		} else {
			final String error = "Illegal event as first or last trace element detected: " + event.getClass().getName();
			LOGGER.error(error);
			throw new KiekerImportException(error);
		}

		return "public static void main(String[])".equals(trace.getOperationSignature()) || "public void run()".equals(trace.getOperationSignature());
	}

	/**
	 * Imports a {@link JoinEvent} instance by creating a corresponding {@link ThreadJoin} instance and adding it to the given {@link SThread} object.
	 * 
	 * @param event
	 *            The join event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the thread join object to.
	 */
	private void importJoinEvent(final JoinEvent event, final STrace trace) {

		final ThreadJoin exec = SynchronisationFactoryImpl.eINSTANCE.createThreadJoin();
		this.setCommonElementAttributes(event, trace, exec);
		final STrace temp = DynamicsFactoryImpl.eINSTANCE.createSTrace();
		temp.setTraceId(event.getJoinedTraceId());
		exec.setTarget(temp);
		this.sthreadJoins.put(exec, event.getJoinedTraceId());
		trace.getEvents().add(exec);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Imported new JoinEvent with values: " + exec.toString());
		}
	}

	/**
	 * Imports a {@link SplitEvent} instance by creating a corresponding {@link SThreadCreation} instance and adding it to the given {@link SThread} object.
	 * 
	 * @param event
	 *            The split event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the thread creation object to.
	 */
	private void importSplitEvent(final SplitEvent event, final STrace trace) {

		final ThreadCreation exec = DynamicsFactoryImpl.eINSTANCE.createThreadCreation();
		this.setCommonElementAttributes(event, trace, exec);
		this.threadCreations.add(exec);
		trace.getEvents().add(exec);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Imported new SplitEvent with values: " + exec.toString());
		}
	}

	/**
	 * Imports a {@link CallConstructorEvent} instance by creating a corresponding {@link ObjectConstruction} instance and adding it to the given {@link SThread}
	 * object.
	 * 
	 * @param event
	 *            The call constructor event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the imported object to.
	 */
	private void importCallConstructorEvent(final CallConstructorEvent event, final STrace trace) {

		final String calleeKiekerSignature;
		if (event.getClassSignature().contains("$$")) {
			calleeKiekerSignature = event.getCalleeClassSignature().substring(0, event.getCalleeClassSignature().indexOf("$$"));
		} else {
			calleeKiekerSignature = event.getCalleeClassSignature();
		}
		
		final SClass sClass = this.model.getClassByKiekerClassSignature(calleeKiekerSignature);
		if (sClass != null) {
			final ConstructorExecution exec = DynamicsFactoryImpl.eINSTANCE.createConstructorExecution();
			this.setCommonElementAttributes(event, trace, exec);
			exec.setSMethod(this.model.getMethodBySimplifiedSignature(sClass, KiekerSignatureUtils.toSimplifiedSignature(event.getCalleeOperationSignature())));
			if (exec.getSMethod() != null) {
				this.stack.add(0, Pair.of((OperationExecution) exec, calleeKiekerSignature));
				return;
			}
		}
		// discard event, if a class is instantiated, which is not defined in the visualized application itself or is anonymous inner class.

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Unable to find class or method matching Kieker class signature " + calleeKiekerSignature + " and operation signature "
					+ event.getCalleeOperationSignature() + ". Discarding object construction.");
		}
	}

	/**
	 * Imports a {@link BeforeConstructorEvent} instance by creating a corresponding {@link ObjectConstruction} instance and adding it to the given {@link SThread}
	 * object.
	 * 
	 * @param event
	 *            The before constructor event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the imported object to.
	 */
	private void importBeforeConstructorEvent(final BeforeConstructorEvent event, final STrace trace) {

		final String kiekerSignature;
		if (event.getClassSignature().contains("$$")) {
			kiekerSignature = event.getClassSignature().substring(0, event.getClassSignature().indexOf("$$"));
		} else {
			kiekerSignature = event.getClassSignature();
		}

		final SClass sClass = this.model.getClassByKiekerClassSignature(kiekerSignature);
		if (sClass != null && !kiekerSignature.equals(this.getCurrentKiekerSignature())) {
			// constructor call is in not monitored classes
			final ConstructorExecution exec = DynamicsFactoryImpl.eINSTANCE.createConstructorExecution();
			this.setCommonElementAttributes(event, trace, exec);
			exec.setSMethod(this.model.getMethodBySimplifiedSignature(sClass, KiekerSignatureUtils.toSimplifiedSignature(event.getOperationSignature())));
			if (exec.getSMethod() != null) {
				this.stack.add(0, Pair.of((OperationExecution) exec, kiekerSignature));
				return;
			}
			// discard event, if a class is instantiated, which is not defined in the visualized application itself or is anonymous inner class.

			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Unable to find class or method matching Kieker class signature " + kiekerSignature + " and operation signature "
						+ event.getOperationSignature() + ". Discarding this object construction.");
			}
		} else if (sClass == null) {
			LOGGER.warn("Found no class named " + kiekerSignature + " in the KDM file. Skipping event import.");
		}
	}

	/**
	 * Imports the given {@link ConstructionEvent} by adding the information about the constructed event to the constructor execution just being parsed.
	 * 
	 * @param event
	 *            The construction event to import.
	 */
	private void importConstructionEvent(final ConstructionEvent event) {

		final String kiekerSignature;
		if (event.getClassSignature().contains("$$")) {
			kiekerSignature = event.getClassSignature().substring(0, event.getClassSignature().indexOf("$$"));
		} else {
			kiekerSignature = event.getClassSignature();
		}
		
		final String sig = this.getCurrentKiekerSignature();
		if (sig != null && sig.equals(kiekerSignature)) {
			final OperationExecution exec = this.getCurrentParent();
			if (exec instanceof SMethodExecution) {
				throw new KiekerImportException(
						"Corrupt trace event record detected. Nesting structure is flawed. Found no matching CallConstructorEvent to ConstructorEvent"
								+ " with orderIndex = " + event.getOrderIndex() + ".");
			}
			final SClassInstance instance = DynamicsFactoryImpl.eINSTANCE.createSClassInstance();
			instance.setSClass(exec.getSMethod().getSClass());
			final ConstructorExecution constructor = (ConstructorExecution) exec;
			instance.setConstructorExecution(constructor);
			constructor.setConstruction(instance);
			instance.setKiekerObjectId(event.getObjectId());
			this.idToSObject.put(event.getObjectId(), instance);
		}
	}

	/**
	 * Imports the given {@link AfterConstructorEvent} or {@link AfterOperationFailedEvent} instance by finalizing the current {@link ObjectConstruction} object.
	 * 
	 * @param event
	 *            The after constructor (failed) event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the imported object to.
	 */
	private void importAfterConstructorEvent(final AbstractOperationEvent event, final STrace trace) {

		final String kiekerSignature;
		if (event.getClassSignature().contains("$$")) {
			kiekerSignature = event.getClassSignature().substring(0, event.getClassSignature().indexOf("$$"));
		} else {
			kiekerSignature = event.getClassSignature();
		}
		
		if (kiekerSignature.equals(this.getCurrentKiekerSignature())) {
			final OperationExecution exec = this.getCurrentParent();
			if (exec instanceof SMethodExecution) {
				throw new KiekerImportException(
						"Corrupt trace event record detected. Nesting structure is flawed. Found no matching CallConstructorEvent to AfterConstructorEvent "
								+ "with orderIndex = " + event.getOrderIndex() + ".");
			}
			exec.setTout(event.getTimestamp());
			trace.getEvents().add(exec);
			this.removeStackFirst(ConstructorExecution.class);

			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Imported new ObjectConstruction with values: " + ((ConstructorExecution) exec).toString());
			}
		} else if (this.model.getClassByKiekerClassSignature(event.getClassSignature()) == null) {
			LOGGER.warn("Found no class named " + event.getClassSignature() + " in the KDM file. Skipping event import.");
		}
	}

	/**
	 * Imports a {@link CallObjectOperationEvent} instance by creating a corresponding {@link SMethodExecution} instance and setting it as the value of
	 * {@code this.currentExecution}.
	 * 
	 * @param event
	 *            The call object operation event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the imported object to.
	 */
	private void importCallOperationObjectEvent(final CallOperationObjectEvent event, final STrace trace) {

		final SClass sClass = this.model.getClassByKiekerClassSignature(event.getCalleeClassSignature());
		if (sClass != null && !event.getCalleeOperationSignature().contains("access$")) {
			// call within visualized application
			final SMethodExecution exec = DynamicsFactoryImpl.eINSTANCE.createSMethodExecution();
			this.setCommonElementAttributes(event, trace, exec);
			exec.setSMethod(this.model.getMethodBySimplifiedSignature(sClass, KiekerSignatureUtils.toSimplifiedSignature(event.getCalleeOperationSignature())));
			if (event.getCalleeOperationSignature().contains("static ") || "static".equalsIgnoreCase(event.getCalleeOperationSignature())) {
				// target is class object
				exec.setTarget(sClass.getInstances().get(0));
				this.operationExecutions.put(exec, Pair.of(this.getCurrentExecutingObjectId(exec), (Integer) null));
			} else {
				this.operationExecutions.put(exec, Pair.of(event.getCallerObjectId(), event.getCalleeObjectId()));
			}

			if (exec.getSMethod() != null) {
				this.stack.add(0, Pair.of((OperationExecution) exec, event.getCalleeClassSignature()));
			}
		} else if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Unable to find class or method matching Kieker class signature " + event.getCalleeClassSignature() + " and operation signature "
					+ event.getCalleeOperationSignature() + ". Discarding call object operation.");
		}
	}

	/**
	 * Imports a {@link BeforeOperationObjectEvent} instance by creating a corresponding {@link SMethodExecution} instance and setting it as the value of
	 * {@code this.currentExecution}.
	 * 
	 * @param event
	 *            The before object operation event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the imported object to.
	 */
	private void importBeforeOperationObjectEvent(final BeforeOperationObjectEvent event, final STrace trace) {

		final SClass sClass = this.model.getClassByKiekerClassSignature(event.getClassSignature());
		if (sClass != null && !event.getClassSignature().equals(this.getCurrentKiekerSignature())) {
			// operation call is in not monitored classes
			final SMethodExecution exec = DynamicsFactoryImpl.eINSTANCE.createSMethodExecution();
			this.setCommonElementAttributes(event, trace, exec);
			exec.setSMethod(this.model.getMethodBySimplifiedSignature(sClass, KiekerSignatureUtils.toSimplifiedSignature(event.getOperationSignature())));
			if (event.getOperationSignature().contains("static ") || "static".equalsIgnoreCase(event.getOperationSignature())) {
				// target is class object
				exec.setTarget(sClass.getInstances().get(0));
				this.operationExecutions.put(exec, Pair.of(this.getCurrentExecutingObjectId(exec), (Integer) null));
			} else {
				this.operationExecutions.put(exec, Pair.of(this.getCurrentExecutingObjectId(exec), event.getObjectId()));
			}

			if (exec.getSMethod() != null) {
				this.stack.add(0, Pair.of((OperationExecution) exec, event.getClassSignature()));
			}
		} else if (sClass == null) {
			LOGGER.warn("Found no class named " + event.getClassSignature() + " in the KDM file. Skipping event import.");
		}
	}

	/**
	 * Imports the given {@link AfterOperationEvent} or {@link AfterOperationFailedEvent} instance by finalizing the current {@link SMethodExecution} object and
	 * adding it to the given {@link STrace} instance.
	 * 
	 * @param event
	 *            The after operation (failed) event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the imported object to.
	 */
	private void importAfterOperationEvent(final AbstractOperationEvent event, final STrace trace) {

		if (event.getClassSignature().equals(this.getCurrentKiekerSignature())) {
			final OperationExecution exec = this.getCurrentParent();
			if (exec instanceof ConstructorExecution) {
				throw new KiekerImportException(
						"Corrupt trace event record detected. Nesting structure is flawed. Found no matching CallObjectOperationEvent to AfterOperationEvent "
								+ "with orderIndex = " + event.getOrderIndex() + ".");
			}
			exec.setTout(event.getTimestamp());
			trace.getEvents().add(exec);
			this.removeStackFirst(OperationExecution.class);

			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Imported new SMethodExecution with values: " + ((SMethodExecution) exec).toString());
			}
		}
	}

	/**
	 * Imports the given {@link MonitorRequestEvent} instance by creating a new corresponding {@link MonitorEvent} object and adding it to the given {@link STrace}
	 * instance.
	 * 
	 * @param event
	 *            The monitor request event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the imported object to.
	 */
	private void importMonitorRequestEvent(final MonitorRequestEvent event, final STrace trace) {

		final MonitorEvent exec = SynchronisationFactoryImpl.eINSTANCE.createMonitorEvent();
		this.setCommonElementAttributes(event, trace, exec);
		exec.setType(MonitorMethod.REQUEST);
		this.monitorEvents.put(exec, event.getLockId());
		trace.getEvents().add(exec);
	}

	/**
	 * Imports the given {@link MonitorEntryEvent} instance by creating a new corresponding {@link MonitorEvent} object and adding it to the given {@link STrace}
	 * instance.
	 * 
	 * @param event
	 *            The monitor entry event to import.
	 * @param trace
	 *            The {@link SThread} instance to add the imported object to.
	 */
	private void importMonitorEntryEvent(final MonitorEntryEvent event, final STrace trace) {

		final MonitorEvent exec = SynchronisationFactoryImpl.eINSTANCE.createMonitorEvent();
		this.setCommonElementAttributes(event, trace, exec);
		exec.setType(MonitorMethod.ACQUIRE);
		this.monitorEvents.put(exec, event.getLockId());
		trace.getEvents().add(exec);
	}

	/**
	 * Imports the given {@link MonitorExitEvent} instance by creating a new corresponding {@link MonitorEvent} object and adding it to the given {@link STrace}
	 * instance.
	 * 
	 * @param event
	 *            The monitor exit event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the imported object to.
	 */
	private void importMonitorExitEvent(final MonitorExitEvent event, final STrace trace) {

		final MonitorEvent exec = SynchronisationFactoryImpl.eINSTANCE.createMonitorEvent();
		this.setCommonElementAttributes(event, trace, exec);
		exec.setType(MonitorMethod.RELEASE);
		this.monitorEvents.put(exec, event.getLockId());
		trace.getEvents().add(exec);
	}

	/**
	 * Imports the given {@link MonitorWaitEvent} instance by creating a new corresponding {@link WaitNotifyEvent} object and adding it to the given {@link STrace}
	 * instance.
	 * 
	 * @param event
	 *            The monitor wait event to import.
	 * @param trace
	 *            The {@link STrace} instance to add the imported object to.
	 */
	private void importMonitorWaitEvent(final MonitorWaitEvent event, final STrace trace) {

		final WaitNotifyEvent exec = SynchronisationFactoryImpl.eINSTANCE.createWaitNotifyEvent();
		this.setCommonElementAttributes(event, trace, exec);
		exec.setType(WaitNotifyMethod.WAIT);
		this.waitNotifyEvents.put(exec, event.getLockId());
		trace.getEvents().add(exec);
	}

	/**
	 * Imports the given {@link MonitorNotifyEvent} instance by creating a new corresponding {@link WaitNotifyEvent} object and adding it to the given
	 * {@link SThread} instance.
	 * 
	 * @param event
	 *            The monitor notify event to import.
	 * @param trace
	 *            The {@link SThread} instance to add the imported object to.
	 */
	private void importMonitorNotifyEvent(final MonitorNotifyEvent event, final STrace trace) {

		final WaitNotifyEvent exec = SynchronisationFactoryImpl.eINSTANCE.createWaitNotifyEvent();
		this.setCommonElementAttributes(event, trace, exec);
		exec.setType(WaitNotifyMethod.NOTIFY);
		this.waitNotifyEvents.put(exec, event.getLockId());
		trace.getEvents().add(exec);
	}

	/**
	 * Imports the given {@link MonitorNotifyAllEvent} instance by creating a new corresponding {@link WaitNotifyEvent} object and adding it to the given
	 * {@link SThread} instance.
	 * 
	 * @param event
	 *            The monitor notifyAll event to import.
	 * @param trace
	 *            The {@link SThread} instance to add the imported object to.
	 */
	private void importMonitorNotifyAllEvent(final MonitorNotifyAllEvent event, final STrace trace) {

		final WaitNotifyEvent exec = SynchronisationFactoryImpl.eINSTANCE.createWaitNotifyEvent();
		this.setCommonElementAttributes(event, trace, exec);
		exec.setType(WaitNotifyMethod.NOTIFYALL);
		this.waitNotifyEvents.put(exec, event.getLockId());
		trace.getEvents().add(exec);
	}

	/**
	 * Sets attributes all {@link TraceElement} instances commonly have. Specifically, this method sets the 'source', 'orderIndex', 'tin', 'tout', 'traceId',
	 * 'sThread' and 'ess' attributes of the given trace element, whose values are extracted from the given abstract trace event.
	 * 
	 * @param event
	 *            The Kieker {@link AbstractTraceEvent} representing the given trace element and containing the attribute values to set.
	 * @param trace
	 *            The trace currently being imported.
	 * @param element
	 *            The trace element to set its attribute values.
	 */
	private void setCommonElementAttributes(final AbstractTraceEvent event, final STrace trace, final STraceEvent element) {

		element.setOrderIndex(event.getOrderIndex());
		element.setTin(event.getTimestamp());
		element.setEss(this.stack.size() + 1);
		element.setParent(this.getCurrentParent());
		element.setTraceId(trace.getThreadId());
		element.setTrace(trace);

		this.operationExecutions.put(element, Pair.of(this.getCurrentExecutingObjectId(element), (Integer) null));
	}

	/**
	 * Sets the 'target' attribute of all {@link ThreadCreation} instances in {@code this.threadCreations}.
	 */
	private void setThreadCreationTargets() {

		for (STrace trace : this.traces) {
			for (ThreadCreation creation : this.threadCreations) {
				if (trace.getParentOrderIndex() == creation.getOrderIndex() && trace.getParentTraceId() == creation.getTrace().getTraceId()) {
					creation.setTarget(trace);
					continue;
				}
			}
		}
	}

	/**
	 * Sets the 'target' attribute of all {@link ThreadJoin} instances in {@code this.sthreadJoins}.
	 */
	private void setThreadJoinTargets() {

		for (STrace trace : this.traces) {
			for (Entry<ThreadJoin, Long> join : this.sthreadJoins.entrySet()) {
				if (join.getValue() == trace.getTraceId()) {
					join.getKey().setTarget(trace);
					System.out.println("here");
					continue;
				}
			}
		}
	}

	/**
	 * Sets the correct source and target objects for each {@link OperationExecution} in {@link #operationExecutions}.
	 */
	private void setOperationObjects() {

		for (Entry<STraceEvent, Pair<Integer, Integer>> entry : this.operationExecutions.entrySet()) {

			// set source object
			if (entry.getKey().getSource() == null) {
				final SClassInstance source = (SClassInstance) this.idToSObject.get(entry.getValue().getLeft());
				entry.getKey().setSource(source);
			}

			// set target object
			if (entry.getKey() instanceof SMethodExecution) {
				final SMethodExecution exec = (SMethodExecution) entry.getKey();
				if (exec.getTarget() == null) {
					final SClassInstance target = (SClassInstance) this.idToSObject.get(entry.getValue().getRight());
					if (target == null) {
						final String error = "Object with id " + entry.getValue().getRight()
								+ " is target of an execution, but it has not been created within the Kieker traces.";
						LOGGER.error(error);
						throw new KiekerImportException(error);
					}
					exec.setTarget(target);
				}
			}
		}
	}

	/**
	 * Sets the 'source' attribute of all {@link ThreadJoin} instances in {@code this.execObjects}.
	 */
	private void setExecSources() {

		for (Pair<STraceEvent, ConstructorExecution> entry : this.execObjects) {
			entry.getLeft().setSource(entry.getRight().getConstruction());

			if (entry.getLeft().getSource() == null) {
				final String error = "Found no source object with id = " + entry.getRight() + " for event with order index = " + entry.getLeft().getOrderIndex()
						+ " in the Kieker files.";
				LOGGER.error(error);
				throw new KiekerImportException(error);
			}
		}
	}

	/**
	 * Sets the target attribute of each {@link MonitorEvent} instance. To be invoked after import.
	 */
	private void setMonitorObjects() {

		for (Entry<MonitorEvent, Integer> entry : this.monitorEvents.entrySet()) {

			MonitorObject monitor = this.idToMonitorObject.get(entry.getValue());
			if (monitor == null) {
				monitor = SynchronisationFactoryImpl.eINSTANCE.createMonitorObject();
				monitor.setKiekerObjectId(entry.getValue());
				this.idToMonitorObject.put(monitor.getKiekerObjectId(), monitor);

				// set creator
				MonitorEvent first = null;
				for (Entry<MonitorEvent, Integer> entryC : this.monitorEvents.entrySet()) {
					if (entryC.getValue() == monitor.getKiekerObjectId() && (first == null || entryC.getKey().getTin() < first.getTin())) {
						first = entryC.getKey();
					}
				}

				monitor.setCreator(first);
				monitor.setRealInstance((SClassInstance) this.idToSObject.get(monitor.getKiekerObjectId())); // null, if not existing
			}
			entry.getKey().setTarget(monitor);

			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Imported new Monitor event with values: " + entry.getKey().toString());
			}
		}
	}

	/**
	 * Sets the target attribute of each {@link WaitNotifyEvent} instance. To be invoked after import.
	 */
	private void setWaitNotifyObjects() {

		for (Entry<WaitNotifyEvent, Integer> entry : this.waitNotifyEvents.entrySet()) {

			WaitNotifyObject wait = this.idToWaitObject.get(entry.getValue());
			if (wait == null) {
				wait = SynchronisationFactoryImpl.eINSTANCE.createWaitNotifyObject();
				wait.setKiekerObjectId(entry.getValue());
				this.idToWaitObject.put(wait.getKiekerObjectId(), wait);

				// set creator
				WaitNotifyEvent first = null;
				for (Entry<WaitNotifyEvent, Integer> entryC : this.waitNotifyEvents.entrySet()) {
					if (entryC.getValue() == wait.getKiekerObjectId() && (first == null || entryC.getKey().getTin() < first.getTin())) {
						first = entryC.getKey();
					}
				}

				wait.setCreator(first);
				wait.setRealInstance((SClassInstance) this.idToSObject.get(wait.getKiekerObjectId())); // null, if not existing
			}
			entry.getKey().setTarget(wait);

			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Imported new wait-/notify event with values: " + entry.getKey().toString());
			}
		}
	}

	/**
	 * Removes the first element from {@link #stack}, which is of the specified class.
	 * 
	 * @param clazz
	 *            The class of the stack element to be removed.
	 */
	private void removeStackFirst(final Class<? extends OperationExecution> clazz) {

		final ListIterator<Pair<OperationExecution, String>> iter = this.stack.listIterator();

		while (iter.hasNext()) {
			if (clazz.isInstance(iter.next().getLeft())) {
				iter.remove();
				return;
			}
		}
		final String error = "Kieker trace files are corrupt. Found no " + clazz.getName() + " in the current execution stack.";
		LOGGER.error(error);
		throw new KiekerImportException(error);
	}

	/**
	 * Get the kieker object id of {@link VisualizableElement} instance, which currently holds the control flow regarding the currently imported trace event. This
	 * method will return {@code null}, if the processes or threads main method is the executing instance.
	 * 
	 * @param event
	 *            The event to get its executing object.
	 * @return The id of the object with the current control flow.
	 */
	private Integer getCurrentExecutingObjectId(final STraceEvent event) {

		final Integer retval;

		if (this.stack.isEmpty()) {
			retval = null;
		} else {
			final OperationExecution currentParent = this.stack.get(0).getLeft();
			if (currentParent instanceof ConstructorExecution) {
				final ConstructorExecution cons = (ConstructorExecution) currentParent;
				if (cons.getConstruction() == null) {
					retval = null; // source object will be set after import
					this.execObjects.add(Pair.of(event, (ConstructorExecution) currentParent));
				} else {
					retval = ((ConstructorExecution) currentParent).getConstruction().getKiekerObjectId();
				}
			} else {
				retval = this.operationExecutions.get(currentParent).getRight();
			}
		}
		return retval;
	}

	/**
	 * Get the current parent {@link OperationExecution} instance regarding any event currently being imported.
	 * 
	 * @return The current parent trace element.
	 */
	private OperationExecution getCurrentParent() {

		if (this.stack.isEmpty()) {
			return null;
		}
		return this.stack.get(0).getLeft();
	}

	/**
	 * Get the Kieker class signature of the {@link OperationExecution} just being imported.
	 * 
	 * @return The Kieker class signature.
	 */
	private String getCurrentKiekerSignature() {

		if (this.stack.isEmpty()) {
			return null;
		}
		return this.stack.get(0).getRight();
	}

}
