/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package importer.utils;

import java.util.ArrayList;
import java.util.List;

import utils.StringUtils;

/**
 * A utility class offering commonly used methods to manipulate the method signature String contained in Kieker operation records.
 * 
 * @author pdo
 * 
 */
public final class KiekerSignatureUtils {

	// ========================== Constructor =================================

	/**
	 * Private constructor. This class should not be instantiated since it contains static methods only.
	 */
	private KiekerSignatureUtils() {}

	// ============================ Methods ===================================

	/**
	 * Convert the Kieker method signature string into a general simplified signature, i.e. the usual method signature with all package information removed. Example:
	 * {@code toSimplifiedSignature( "public static native java.lang.String ops.utils.Util.op(long)" )} returns {@code "public static native String op(long)" }.
	 * 
	 * @param signature
	 *            The Kieker method signature to simplify.
	 * @return The simplified signature extracted from the given Kieker signature string.
	 * @throws IllegalArgumentException
	 *             If the input string is {@code null}.
	 */
	public static String toSimplifiedSignature(final String signature) throws IllegalArgumentException {

		if (signature == null) {
			throw new IllegalArgumentException("Input String is null.");
		}

		String retval = signature;
		if (!"static".equals(signature)) {

			// handle <init> and <clinit>
			final String substring = signature.substring(signature.lastIndexOf('.'));
			if (signature.endsWith("<init>")) {
				retval = "<init>";
			} else if (substring.endsWith("<clinit>")) {
				retval = "static <clinit>";
			} else {

				final StringBuffer buffer = new StringBuffer();
				for (String modifier : extractModifiers(signature)) {
					buffer.append(modifier);
					buffer.append(' ');
				}

				final String returnType = extractReturnType(signature);
				if (!returnType.isEmpty()) {
					buffer.append(extractReturnType(signature));
					buffer.append(' ');
				}

				buffer.append(extractMethodName(signature));
				buffer.append('(');
				final String[] params = extractParameterTypes(signature);
				for (int i = 0; i < params.length; i++) {
					buffer.append(params[i]);
					if (i < params.length - 1) {
						buffer.append(", ");
					}
				}
				buffer.append(')');
				retval = buffer.toString();
			}
		}

		return retval;
	}

	/**
	 * Extract the simple method name from a Kieker method signature string. Example:
	 * {@code extractMethodName( "public static native java.lang.String ops.utils.Util.op(long)" )} returns {@code "op" }.
	 * 
	 * @param signature
	 *            The signature string to extract the simple method name from.
	 * @return The simple method name.
	 * @throws IllegalArgumentException
	 *             If the input string is {@code null}.
	 */
	private static String extractMethodName(final String signature) throws IllegalArgumentException {

		if (signature == null) {
			throw new IllegalArgumentException("Input String is null.");
		}

		if ("static".equals(signature)) {
			// static block executions, considered as method as well
			return signature;
		}

		final int opBracketIndex = signature.indexOf('(');
		final String substring = signature.substring(0, opBracketIndex);
		final int indexLastDot = substring.lastIndexOf('.');
		return substring.substring(indexLastDot + 1);
	}

	/**
	 * Extract the simple return type from a Kieker method signature string. Example:
	 * {@code extractReturnType( "public static native java.lang.String ops.utils.Util.op(long)" )} returns {@code "String" }.
	 * 
	 * @param signature
	 *            The signature string to extract the return type from.
	 * @return The simple return type.
	 * @throws IllegalArgumentException
	 *             If the input string is {@code null}.
	 */
	private static String extractReturnType(final String signature) throws IllegalArgumentException {

		if (signature == null) {
			throw new IllegalArgumentException("Input String is null.");
		}

		if ("static".equals(signature)) {
			// static block executions, considered as method as well
			return "void";
		}

		String retval = "";
		final int opBracketIndex = signature.indexOf('(');
		final String[] components = signature.substring(0, opBracketIndex).split("\\s+");
		if (components.length > 1) {
			retval = StringUtils.cutToLast(components[components.length - 2], '.').trim();

			// consider constructors: they do not have a return type
			if ("public".equals(retval) || "private".equals(retval) || "protected".equals(retval)) {
				retval = "";
			}
		}

		return retval;
	}

	/**
	 * Extract the package name from a Kieker method signature string. Example:
	 * {@code extractPackageName( "public static native java.lang.String ops.utils.Util.op(long)" )} returns {@code "ops.utils" }.
	 * 
	 * @param signature
	 *            The signature string to extract the package name from.
	 * @return The package name.
	 * @throws IllegalArgumentException
	 *             If the input string is {@code null}.
	 */
	public static String extractPackageName(final String signature) throws IllegalArgumentException {

		if (signature == null) {
			throw new IllegalArgumentException("Input String is null.");
		}

		if ("static".equals(signature)) {
			// static block executions, considered as method as well
			return "";
		}

		int lastBracketIndex = signature.lastIndexOf('(');
		if (lastBracketIndex == -1) {
			// method is <init> or <clinit>
			lastBracketIndex = signature.length() - 1;
		}

		final int indexLastDot = signature.lastIndexOf('.', lastBracketIndex);
		final int secLastDotIndex = signature.lastIndexOf('.', indexLastDot - 1);
		final int indexSpaceChar = signature.lastIndexOf(' ', secLastDotIndex);
		return signature.substring(indexSpaceChar + 1, secLastDotIndex);
	}

	/**
	 * Extract the simple class name from a Kieker method signature string. Example:
	 * {@code extractClassName( "public static native java.lang.String ops.utils.Util.op(long)" )} returns {@code "Util" }.
	 * 
	 * @param signature
	 *            The signature string to extract the simple class name from.
	 * @return The simple class name.
	 * @throws IllegalArgumentException
	 *             If the input string is {@code null}.
	 */
	public static String extractClassName(final String signature) throws IllegalArgumentException {

		if (signature == null) {
			throw new IllegalArgumentException("Input String is null.");
		}

		if ("static".equals(signature)) {
			// static block executions, considered as method as well
			return "";
		}

		int lastBracketIndex = signature.lastIndexOf('(');
		if (lastBracketIndex == -1) {
			// method is <init> or <clinit>
			lastBracketIndex = signature.length() - 1;
		}

		final int indexLastDot = signature.lastIndexOf('.', lastBracketIndex);
		final int secLastDotIndex = signature.lastIndexOf('.', indexLastDot - 1);
		return signature.substring(secLastDotIndex + 1, indexLastDot);
	}

	/**
	 * Extract the simple class name from a Kieker method signature string. Example:
	 * {@code extractClassName( "public static native java.lang.String ops.utils.Util.op(long)" )} will return {@code "Util" }.
	 * 
	 * @param signature
	 *            The signature string to extract the simple class name from.
	 * @return The simple class name.
	 * @throws IllegalArgumentException
	 *             If the input string is {@code null}.
	 */
	private static String[] extractParameterTypes(final String signature) throws IllegalArgumentException {

		if (signature == null) {
			throw new IllegalArgumentException("Input String is null.");
		}

		if ("static".equals(signature)) {
			// static block executions, considered as method as well
			return new String[0];
		}

		final String[] params = signature.substring(signature.indexOf('(') + 1, signature.indexOf(')')).split(",");
		final String[] retval = new String[params.length];
		for (int i = 0; i < params.length; i++) {
			params[i] = params[i].trim();
			final String paramType = params[i].substring(params[i].lastIndexOf('.') + 1);
			retval[i] = paramType;
		}
		return retval;
	}

	/**
	 * Extracts the method modifiers from a Kieker method signature string in proper order. Example:
	 * {@code extractModifiers( "public static native java.lang.String ops.utils.Util.op(long)" )} will return {@code "["public", "static", "native"]" }.
	 * 
	 * @param signature
	 *            The signature to extract its method modifiers from.
	 * @return The method modifiers in proper order.
	 * @throws IllegalArgumentException
	 *             If the input string is {@code null}.
	 */
	private static String[] extractModifiers(final String signature) throws IllegalArgumentException {

		if (signature == null) {
			throw new IllegalArgumentException("Input String is null.");
		}

		if ("static".equals(signature)) {
			// static block executions, considered as method as well
			return new String[0];
		}

		final String[] subStrings = signature.substring(0, signature.indexOf('.')).split("\\s+");
		final List<String> modifiers = new ArrayList<String>(subStrings.length);
		if (subStrings.length > 1) {
			for (int i = 0; i < subStrings.length - 2; i++) {
				modifiers.add(subStrings[i]);
			}
		} else if (subStrings.length == 1) {
			// method has no return type -> is constructor
			modifiers.add(subStrings[0]);
		}
		return modifiers.toArray(new String[0]);
	}

}
