/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package logic;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

import configuration.Configuration;

import data.dynamics.OperationExecution;
import data.dynamics.STrace;
import data.dynamics.STraceEvent;
import data.dynamics.ThreadCreation;
import data.statics.SClass;
import data.statics.SPackage;
import data.synchronisation.MonitorEvent;
import data.synchronisation.MonitorMethod;
import data.synchronisation.WaitNotifyEvent;
import data.synchronisation.WaitNotifyMethod;

import models.city.CityModel;
import models.data.DataModelUtils;

import observer.AbstractObservable;
import observer.IObserver;
import observer.ObservableAction;
import visual.gui.CityControlBar;
import visual.gui.CityControlBar.ControlBarButton;

/**
 * A city controller to control the city metaphor style visualization on screen.
 * 
 * @author pdo
 * 
 */
public final class CityController implements IObserver {

	// =========================== Attributes =================================

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(CityController.class);

	/** The {@link CityModel} instance currently being controlled by this city controller. */
	private CityModel model;

	/** The {@link CityControlBar} instance currently being controlled by this city controller. */
	private CityControlBar controlBar;

	/** A list of all imported {@link STraceEvent} instances to handle by this city controller. */
	private List<STraceEvent> allEvents = new LinkedList<STraceEvent>();

	/** A list of all thread events currently visualized sorted by the trace they belong to. */
	private List<List<STraceEvent>> visibleEvents = new ArrayList<List<STraceEvent>>(10);

	/** A list of all thread events currently hidden sorted by the trace they belong to. */
	private List<List<STraceEvent>> hiddenEvents = new ArrayList<List<STraceEvent>>(10);

	/** A list of all thread events already visualized up to the pointer {@link #iterator} sorted by the trace they belong to. */
	private List<List<STraceEvent>> statusEvents = new ArrayList<List<STraceEvent>>(10);

	/** A list iterator to iterate over {@link #allEvents}. */
	private ListIterator<STraceEvent> iterator;

	/** The event most recently added to the visualization. */
	private STraceEvent currentEvent;

	/** The current time index used for concrete interval step operations. */
	private long timeIndex;

	/** A flag indicating whether a replay is currently running. */
	private volatile boolean replayRunning = false;

	/** A flag indicating whether the threads' and processes' actual execution paths are supposed to be displayed. */
	private volatile boolean showExecutionPath = Configuration.getInstance().isShowExecPathEnabled();

	/**
	 * The time interval to move forwards or backwards in milliseconds when pressing {@link ControlBarButton#NEXT_TIME} or {@link ControlBarButton#PREVIOUS_TIME}.
	 */
	private volatile int intervalStep = Configuration.getInstance().getIntervalStep() * 1000;

	/** The time to wait when replaying until the next thread event is supposed to be visualized in milliseconds. */
	private volatile int replaySpeed = Configuration.getInstance().getReplaySpeed();

	/** The {@link RelationsManager} to display application relations. */
	private RelationsManager relationsController;

	/** The {@link ReplayExecutor} performing the replay. */
	private ReplayExecutor executor = new ReplayExecutor();

	// ========================== Constructor =================================

	/**
	 * Protected constructor. This class is not intended for instantiation from outside of this package.
	 */
	protected CityController() {

		this.executor.start();
	}

	// ============================ Methods ===================================

	/**
	 * Initializes this city controller with the specified scene model to be controlled. Both this method and {@link #init(CityControlBar)} have to be invoked in
	 * order to initialize this city controller.
	 * 
	 * @param sceneModel
	 *            The scene model to be controlled by this city controller.
	 * @throws IllegalArgumentException
	 *             If the scene model passed in is {@code null}.
	 */
	public void init(final CityModel sceneModel) throws IllegalArgumentException {

		if (sceneModel == null) {
			throw new IllegalArgumentException("You cannot control a null scene model.");
		}

		this.model = sceneModel;
	}

	/**
	 * Initializes this city controller with the specified city control bar to be controlled. Both this method and {@link #init(CityModel)} have to be invoked in
	 * order to initialize this city controller.
	 * 
	 * @param cityControlBar
	 *            The city control bar to be controlled by this city controller.
	 * @throws IllegalArgumentException
	 *             If the control bar passed in is {@code null}.
	 */
	public void init(final CityControlBar cityControlBar) throws IllegalArgumentException {

		if (cityControlBar == null) {
			throw new IllegalArgumentException("You cannot control a null city control bar.");
		}

		this.controlBar = cityControlBar;
	}

	@Override
	@SuppressWarnings("unchecked")
	public void update(final AbstractObservable observable, final ObservableAction action, final Object object) {

		switch (action) {
		case STATICS_LOADED:
			this.onStaticsLoaded((SPackage) object);
			break;
		case DYNAMICS_LOADED:
			this.onDynamicsLoaded((List<STrace>) object);
			break;
		case ALL_CLEARED:
			this.onAllCleared();
			break;
		default:
			LOGGER.warn("Unsupported observable action received: " + action.name() + ".");
		}
	}

	/**
	 * The action to be performed, when a notification with {@link ObservableAction#STATICS_LOADED} has been received. This method simply invokes
	 * {@link CityModel#setRootPackage(SPackage)} on the currently controlled scene model.
	 * 
	 * @param rootPack
	 *            The root package of the application to visualize.
	 */
	private void onStaticsLoaded(final SPackage rootPack) {

		this.relationsController = new RelationsManager(this.model, rootPack);
		this.model.setRootPackage(rootPack);
	}

	/**
	 * The action to be performed, when a notification with {@link ObservableAction#DYNAMICS_LOADED} has been received. This method invokes
	 * {@link CityModel#addThreadElement(SThread)} on the currently controlled scene model for each thread instance in the given list in proper order, adds all
	 * thread events to {@link #allEvents}, and sorts the list using a {@link STraceEventComparator} instance.
	 * 
	 * @param traces
	 *            The list of traces to be added to the scene model. All processes have to be at the beginning of the list in no specified order.
	 */
	private synchronized void onDynamicsLoaded(final List<STrace> traces) {

		// add all thread events
		this.allEvents = new LinkedList<STraceEvent>();
		for (STrace trace : traces) {
			if (trace.getParentOrderIndex() < 0) {
				// thread represents a process
				this.model.addThreadObject(trace);
				if (trace.getExecOrder() == 0) {
					this.timeIndex = trace.getTin();
				}
			}
			this.visibleEvents.add(new ArrayList<STraceEvent>(trace.getEvents().size()));
			this.statusEvents.add(new ArrayList<STraceEvent>(trace.getEvents().size()));
			this.hiddenEvents.add(new ArrayList<STraceEvent>(trace.getEvents().size()));
			this.allEvents.addAll(trace.getEvents());
		}

		// prepare iterator and replay executor
		if (!traces.isEmpty()) {
			Logic.getInstance().getTreeViewController().setContent(traces);
			DataModelUtils.orderByTin(this.allEvents);
			this.iterator = this.allEvents.listIterator();
			this.replayRunning = false;
			this.updateButtons();

			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Imported " + traces + " threads with a total of " + this.allEvents.size() + " thread events.");
			}
		}
	}

	/**
	 * The action to be performed, when a notification with {@link ObservableAction#ALL_CLEARED} has been received. This method simply invokes
	 * {@link CityModel#clearAll()} on the currently controlled scene model.
	 */
	private synchronized void onAllCleared() {

		this.removeVisibleRelations();
		this.allEvents.clear();
		this.iterator = null;
		this.currentEvent = null;
		this.timeIndex = 0;
		this.visibleEvents.clear();
		this.statusEvents.clear();
		this.model.clearAll();

		// deactivate all buttons
		for (ControlBarButton button : ControlBarButton.values()) {
			this.controlBar.setButtonEnabled(button, false);
		}

		Logic.getInstance().getMenuBarController().setShowGlobalRelationsItemEnabled(false);
		Logic.getInstance().getTreeViewController().removeAllContent();
	}

	/**
	 * Rebuilds the entire {@link CityControlBar} currently controlled by this controller. This method has no effect, if this controller currently does not control
	 * any control bar.
	 */
	public void rebuild() {

		if (this.controlBar == null) {
			LOGGER.warn("Control bar is null in method 'rebuild'.");
		} else {
			this.controlBar.rebuild();
		}
	}

	/**
	 * Displays or hides the actual execution paths of all threads or processes in the final visualization.
	 * 
	 * @param enabled
	 *            {@code True}, to activate the visualization of the actual execution path, {@code} false to deactivate.
	 */
	public void setShowExecutionPath(final boolean enabled) {

		if (this.showExecutionPath != enabled) {
			for (List<STraceEvent> events : this.statusEvents) {
				// enable execution path
				if (!events.isEmpty() && enabled) {
					final List<STraceEvent> visibles = this.visibleEvents.get(events.get(0).getTrace().getExecOrder());
					for (int i = 0; i < events.size() - 1; i++) {
						this.model.addThreadEvent(events.get(i), true);
						visibles.add(events.get(i));
					}
				} else if (!events.isEmpty()) {
					// disable execution path
					final List<STraceEvent> visibles = this.visibleEvents.get(events.get(0).getTrace().getExecOrder());
					for (int i = 0; i < events.size() - 1; i++) {
						this.model.removeTraceEvent(events.get(i), true);
						visibles.remove(events.get(i));
					}
				}
			}
			this.showExecutionPath = enabled;
		}
	}

	/**
	 * Sets the speed of the replay. I.e. the number of milliseconds until the next trace event will be displayed.
	 * 
	 * @param speed
	 *            The replay speed in milliseconds to set.
	 * @throws IllegalArgumentException
	 *             If the speed passed in is non-positive.
	 */
	public void setReplaySpeed(final int speed) throws IllegalArgumentException {

		if (speed <= 0) {
			throw new IllegalArgumentException("You cannot set a non-positive speed.");
		}
		this.replaySpeed = speed;

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully set replay speed to " + speed + ".");
		}
	}

	/**
	 * Sets the time interval to move forwards or backwards, when pressing {@link ControlBarButton#NEXT_TIME} or {@link ControlBarButton#PREVIOUS_TIME},
	 * respectively.
	 * 
	 * @param interval
	 *            The time interval in milliseconds to set.
	 * @throws IllegalArgumentException
	 *             If the interval passed in is non-positive.
	 */
	public void setIntervalStep(final int interval) throws IllegalArgumentException {

		if (interval <= 0) {
			throw new IllegalArgumentException("You cannot set a non-positive interval step.");
		}
		this.intervalStep = interval * 1000;

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully set interval step to " + interval + ".");
		}
	}

	/**
	 * Moves the current {@link #iterator} pointer to the specified trace event and updates the visualization accordingly.
	 * 
	 * @param traceEvent
	 *            The trace event to jump to.
	 * @throws IllegalArgumentException
	 *             If the trace event passed in is {@code null}.
	 */
	public synchronized void jumpTo(final STraceEvent traceEvent) throws IllegalArgumentException {

		if (traceEvent == null) {
			throw new IllegalArgumentException("You cannot jump to a null thread event.");
		}

		// jump to specified thread event
		while (this.currentEvent == null || this.currentEvent.getExecOrder() < traceEvent.getExecOrder()) {
			this.oneStepForwards();
		}
		while (this.currentEvent.getExecOrder() > traceEvent.getExecOrder()) {
			this.oneStepBackwards();
		}
	}

	/**
	 * The action to be performed when {@link ControlBarButton#NEXT_EVENT} has been pressed. It moves the current {@link #iterator} pointer one event forwards (if
	 * existing). The Gui will be updated automatically afterwards ({@link #updateGui(int)});
	 */
	public synchronized void oneStepForwards() {

		final STraceEvent next = this.getNext();
		if (next != null) {
			this.visualizeEvent(next);
		}
	}

	/**
	 * The action to be performed when {@link ControlBarButton#PREVIOUS_EVENT} has been pressed. It moves the current {@link #iterator} pointer one event backwards
	 * (if existing). The Gui will be updated automatically afterwards ({@link #updateGui(int)});
	 */
	public synchronized void oneStepBackwards() {

		final STraceEvent prev = this.getPrevious();
		if (prev != null) {
			this.unvisualizeEvent(prev);
		}
	}

	/**
	 * The action to be performed when {@link ControlBarButton#NEXT_TIME} has been pressed. It moves the current {@link #iterator} pointer one interval (i.e.
	 * {@link #intervalStep} milliseconds) forwards. The Gui will be updated automatically afterwards ( {@link #updateGui(int)});
	 */
	public synchronized void oneIntervalForwards() {

		final long tin = this.timeIndex;
		while (this.iterator.hasNext()) {
			final STraceEvent next = this.iterator.next();
			this.iterator.previous();

			if (tin + this.intervalStep >= next.getTin()) {
				this.oneStepForwards();
			} else {
				break;
			}
		}

		this.timeIndex = tin + this.intervalStep;
	}

	/**
	 * The action to be performed when {@link ControlBarButton#PREVIOUS_TIME} has been pressed. It moves the current {@link #iterator} pointer one interval (i.e.
	 * {@link #intervalStep} milliseconds) backwards. The Gui will be updated automatically afterwards ({@link #updateGui(int)});
	 */
	public synchronized void oneIntervalBackwards() {

		final long tin = this.timeIndex;
		while (this.iterator.hasPrevious()) {
			final STraceEvent prev = this.iterator.previous();
			this.iterator.next();

			if (tin - this.intervalStep <= prev.getTin()) {
				this.oneStepBackwards();
			} else {
				break;
			}
		}

		this.timeIndex = tin - this.intervalStep;
	}

	/**
	 * The action to be performed when {@link ControlBarButton#REPLAY_PAUSE} has been pressed. It starts (or resumes) the replay at the current position defined by
	 * {@link #currentEvent}.
	 */
	public synchronized void startReplay() {

		if (this.iterator.hasNext()) {
			this.replayRunning = true;
			synchronized (this.executor) {
				this.executor.notify();
				this.updateButtons();
			}
		}
	}

	/**
	 * The action to be performed when {@link ControlBarButton#REPLAY_PAUSE} has been pressed. It pauses the current replay, i.e. suspends the {@link #executor}. The
	 * Buttons will be updated automatically afterwards ({@link #updateButtons()}).
	 */
	public synchronized void pauseReplay() {

		this.replayRunning = false;
		this.updateButtons();
	}

	/**
	 * Returns the previous {@link STraceEvent} relative to the current trace pointer ({@link #iterator}) and updates all pointer related values.
	 * 
	 * @return The previous {@link STraceEvent} relative to the current trace pointer.
	 */
	private STraceEvent getPrevious() {

		STraceEvent prev = null;
		if (this.iterator.hasPrevious()) {
			prev = this.iterator.previous();
			this.timeIndex = prev.getTin();
		}
		this.currentEvent = prev;
		return prev;
	}

	/**
	 * Returns the next {@link STraceEvent} relative to the current trace pointer ({@link #iterator}) and updates all pointer related values.
	 * 
	 * @return The next {@link STraceEvent} relative to the current trace pointer.
	 */
	private STraceEvent getNext() {

		STraceEvent next = null;
		if (this.iterator.hasNext()) {
			next = this.iterator.next();
			this.timeIndex = next.getTin();
		}
		this.currentEvent = next;
		return next;
	}

	/**
	 * Adds the specified {@link STraceEvent} to the city model, i.e. visualizes it on screen.
	 * 
	 * @param event
	 *            The event to visualize.
	 */
	private void visualizeEvent(final STraceEvent event) {

		final List<STraceEvent> visibles = this.visibleEvents.get(event.getTrace().getExecOrder());
		final List<STraceEvent> stati = this.statusEvents.get(event.getTrace().getExecOrder());

		if (!this.showExecutionPath) {
			
			for (STraceEvent prev : visibles) {
				this.model.removeTraceEvent(prev, true);
			}
			visibles.clear();
			
			for (int i = 0; i < this.visibleEvents.size(); i++) {
				final List<STraceEvent> visiblez = this.visibleEvents.get(i);
				for (STraceEvent visible : visiblez) { // contains only one element at max
					if (visible instanceof OperationExecution && ((OperationExecution) visible).getTout() < event.getTin()) {
						if (!this.hiddenEvents.get(i).contains(visible)) {
							this.model.removeTraceEvent(visible, true);
							this.hiddenEvents.get(i).add(visible);
						}
					} else if (visible instanceof ThreadCreation) {
						if (!this.hiddenEvents.get(i).contains(visible)) {
							this.model.removeTraceEvent(visible, true);
							this.hiddenEvents.get(i).add(visible);
						}
					} else if (visible instanceof MonitorEvent && ((MonitorEvent) visible).getType() != MonitorMethod.REQUEST) {
						if (!this.hiddenEvents.get(i).contains(visible)) {
							this.model.removeTraceEvent(visible, true);
							this.hiddenEvents.get(i).add(visible);
						}
					} else if (visible instanceof WaitNotifyEvent && ((WaitNotifyEvent) visible).getType() != WaitNotifyMethod.WAIT) {
						if (!this.hiddenEvents.get(i).contains(visible)) {
							this.model.removeTraceEvent(visible, true);
							this.hiddenEvents.get(i).add(visible);
						}
					}
				}
			}
		}

		visibles.add(event);
		stati.add(event);
		this.model.addThreadEvent(event, false);
		this.updateGui();
	}

	/**
	 * Removes the specified {@link STraceEvent} from the city model, i.e. removes it from the visualization on screen.
	 * 
	 * @param event
	 *            The event to remove from the current scene.
	 */
	private void unvisualizeEvent(final STraceEvent event) {

		final List<STraceEvent> visibles = this.visibleEvents.get(event.getTrace().getExecOrder());
		final List<STraceEvent> stati = this.statusEvents.get(event.getTrace().getExecOrder());

//		if (!this.showExecutionPath) {
//			final List<STraceEvent> hidden = new LinkedList<STraceEvent>();
//			for (int i = 0; i < this.hiddenEvents.size(); i++) {
//				final List<STraceEvent> visiblez = this.hiddenEvents.get(i);
//				for (STraceEvent visible : visiblez) { // contains only one element at max
//					if (visible instanceof OperationExecution && ((OperationExecution) visible).getTout() >= event.getTin()) {
//						this.model.addThreadEvent(visible, true);
//						hidden.add(visible);
//					} else if (visible instanceof ThreadCreation) {
//						this.model.addThreadEvent(visible, true);
//						hidden.add(visible);
//					} else if (visible instanceof MonitorEvent && ((MonitorEvent) visible).getType() != MonitorMethod.REQUEST) {
//						this.model.addThreadEvent(visible, true);
//						hidden.add(visible);
//					} else if (visible instanceof WaitNotifyEvent && ((WaitNotifyEvent) visible).getType() != WaitNotifyMethod.WAIT) {
//						this.model.addThreadEvent(visible, true);
//						hidden.add(visible);
//					}
//				}
//				for (STraceEvent hidd : hidden) {
//					visiblez.remove(hidd);
//				}
//				hidden.clear();
//			}
//		}

		visibles.remove(visibles.size() - 1);
		stati.remove(stati.size() - 1);
		this.model.removeTraceEvent(event, false);

		if (!this.showExecutionPath && !stati.isEmpty()) {
			final STraceEvent last = stati.get(stati.size() - 1);
			this.model.addThreadEvent(last, true);
			visibles.add(last);
		}

		this.updateGui();
	}

	/**
	 * Updates all {@link Gui} components to the current {@link #iterator} status by setting their content or highlight to {@link #currentEvent}.
	 */
	private void updateGui() {

		if (this.iterator.hasPrevious()) {
			this.currentEvent = this.iterator.previous();
			Logic.getInstance().getTreeViewController().selectNode(this.currentEvent);
			this.iterator.next();
		} else {
			Logic.getInstance().getTreeViewController().selectNode(null);
		}
		this.removeVisibleRelations();
		this.updateButtons();
	}

	/**
	 * Updates the buttons of the associated {@link CityControlBar}. I.e. it sets each of them enabled or disabled depending of the availability of their related
	 * actions.
	 */
	private void updateButtons() {

		this.controlBar.setButtonEnabled(ControlBarButton.PREVIOUS_EVENT, !this.replayRunning && this.iterator.hasPrevious());
		this.controlBar.setButtonEnabled(ControlBarButton.NEXT_EVENT, !this.replayRunning && this.iterator.hasNext());
		this.controlBar.setButtonEnabled(ControlBarButton.PREVIOUS_TIME, !this.replayRunning && this.iterator.hasPrevious());
		this.controlBar.setButtonEnabled(ControlBarButton.NEXT_TIME, !this.replayRunning && this.iterator.hasNext());
		this.controlBar.setButtonEnabled(ControlBarButton.REPLAY_START, !this.replayRunning && this.iterator.hasNext());
		this.controlBar.setButtonEnabled(ControlBarButton.REPLAY_PAUSE, this.replayRunning);
	}

	/**
	 * Returns the event most recently added to the visualization or {@code null}, if no such event exists.
	 * 
	 * @return The event most recently added to the visualization.
	 */
	public STraceEvent getCurrentEvent() {
		return this.currentEvent;
	}

	// ======================== Delegating Methods ============================

	/**
	 * Adds all incoming and outgoing {@link Relation} instances to the scene model currently being controlled. This will make them visible on screen. This method
	 * has no effect, if no static data has been imported yet.
	 * 
	 * @param dataObject
	 *            The data object to show all its outgoing relations on screen.
	 * @throws IllegalArgumentException
	 *             If the sClass object passed in is {@code null}.
	 */
	public void showLocalRelations(final SClass dataObject) throws IllegalArgumentException {

		if (this.relationsController == null) {
			LOGGER.warn("Relations controller is null in method 'showLocalRelations'.");
		} else {
			this.relationsController.showLocalRelations(dataObject);
		}
	}

	/**
	 * Adds all {@link Relation} instances available to the scene model currently being controlled. This will make them visible on screen. If no statics data has
	 * been imported yet, this method will have no effect.
	 * 
	 */
	public void showGlobalRelations() {

		if (this.relationsController == null) {
			LOGGER.warn("Relations controller is null in method 'showGlobalRelations'.");
		} else {
			this.relationsController.showGlobalRelations();
		}
	}

	/**
	 * Removes all currently visible relations (i.e. all elements in {@link #visibleRelations}) from the visualization. This method has no effect, if no static data
	 * has been imported yet.
	 */
	public void removeVisibleRelations() {

		if (this.relationsController == null) {
			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Relations controller is null in method 'removeVisibleRelations'.");
			}
		} else {
			this.relationsController.removeVisibleRelations();
		}
	}

	/**
	 * Returns the current status of the relations manager. The pair's left component indicates, whether global or local relations are currently being displayed. If
	 * {@code false}, the {@link SClass} instance whose local relations are currently displayed is stored in the pair's right component ({@code null} otherwise, i.e.
	 * if no relations are currently visible at all). This method will return {@code null}, if no static data has been imported yet.
	 * 
	 * @return The current status of the relations manager or {@code null}, if no static data has been imported yet.
	 */
	public Pair<Boolean, SClass> getRelationsStatus() {

		if (this.relationsController == null) {
			LOGGER.warn("Relations controller is null in method 'getRelationsStatus'.");
			return null;
		} else {
			return this.relationsController.getStatus();
		}
	}

	/**
	 * Updates the relation menu items in menu bar -> 'Relations' according to the currently selected class and relation visualization status ({@link #getStatus()}).
	 * If no statics data has been imported yet, this method will have no effect.
	 */
	public void updateRelationButtons() {

		if (this.relationsController == null) {
			LOGGER.warn("Relations controller is null in method 'showGlobalRelations'.");
		} else {
			this.relationsController.updateButtons();
		}
	}

	// ========================== Inner Classes ===============================

	/**
	 * A replay executor realizing and performing the automatic replay functionality of Synchrovis in a separate thread.
	 * 
	 * @author pdo
	 * 
	 */
	private final class ReplayExecutor extends Thread {

		/**
		 * Constructs a new {@link ReplayExecutor} instance.
		 */
		public ReplayExecutor() {
			super();
		}

		@Override
		public void run() {

			while (true) {
				try {
					while (!(CityController.this.replayRunning)) {
						synchronized (this) {
							this.wait();
						}
					}

					synchronized (CityController.this) {
						CityController.this.oneStepForwards();
						if (!CityController.this.iterator.hasNext()) {
							CityController.this.replayRunning = false;
							CityController.this.updateButtons();
						}
					}
					Thread.sleep(CityController.this.replaySpeed);
				} catch (InterruptedException e) {
					if (CityController.LOGGER.isDebugEnabled()) {
						LOGGER.debug("The replay executor has been interrupted and terminated.", e);
					}
					return;
				}
			}
		}
	}
}
