/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package logic;

import java.awt.EventQueue;
import java.util.ArrayList;
import java.util.List;

import javax.swing.table.DefaultTableModel;

import configuration.Language;

import data.VisualizableElement;
import data.dynamics.ConstructorExecution;
import data.dynamics.SClassInstance;
import data.dynamics.SMethodExecution;
import data.dynamics.STrace;
import data.dynamics.STraceEvent;
import data.dynamics.ThreadCreation;
import data.dynamics.ThreadTermination;
import data.statics.Relation;
import data.statics.RelationType;
import data.statics.SPackage;
import data.synchronisation.MonitorEvent;
import data.synchronisation.MonitorObject;
import data.synchronisation.SynchronisationObject;
import data.synchronisation.WaitNotifyEvent;
import data.synchronisation.WaitNotifyObject;

import models.data.DataModelUtils;

/**
 * A properties panel controller used to add or remove information to a given {@link PropertiesPanel}.
 * 
 * @author pdo
 * 
 */
public final class PropertiesController {

	// =========================== Attributes =================================

	/** A lock used to guarantee mutual exclusiveness of all {@code showProperty(...)} methods. */
	private final Object lock = new Object();

	/** The default table model to which information will be added to. */
	private DefaultTableModel tableModel;

	/** A list of property keys of the values just being displayed. */
	private List<String> keys = new ArrayList<String>();

	/** The {@link VisualizableElement} whose properties are currently displayed. */
	private VisualizableElement displayed;

	// ========================== Constructor =================================

	/**
	 * Protected constructor. This class is not intended for instantiation from outside of this package.
	 */
	protected PropertiesController() {}

	// ============================ Methods ===================================

	/**
	 * Initializes this {@link PropertiesController} with the given {@link DefaultTableModel} instance of the {@link PropertiesPanel} to be controlled.
	 * 
	 * @param defaultTableModel
	 *            The default table model to which information will be added to. New rows may be created or existing ones may be deleted when executing a method
	 *            in this class.
	 * @throws IllegalArgumentException
	 *             If the table model passed in is {@code null}.
	 */
	public void init(final DefaultTableModel defaultTableModel) throws IllegalArgumentException {

		if (defaultTableModel == null) {
			throw new IllegalArgumentException("You cannot pass in a null table model.");
		}

		this.tableModel = defaultTableModel;
	}

	/**
	 * Rebuilds the entire properties panel being controlled by this properties controller.
	 */
	public void rebuild() {
		EventQueue.invokeLater(new RebuildExecutor());
	}

	/**
	 * Removes all content currently added to this table. This {@link #tableModel} will be empty of rows when this method returns.
	 */
	public void removeAllContent() {
		EventQueue.invokeLater(new ContentRemovalExecutor());
	}

	/**
	 * Displays properties of the given {@link VisualizableElement} on the properties panel.
	 * 
	 * @param element
	 *            The entity to display its properties.
	 * @throws IllegalArgumentException
	 *             If the entity passed in is {@code null} or no instance of {@link Street}, {@link Cable}, {@link RegularFloor}, {@link AdministrationFloor}, or
	 *             {@link RegularDistrict}.
	 */
	public void showProperties(final VisualizableElement element) throws IllegalArgumentException {

		if (element == null) {
			throw new IllegalArgumentException("You cannot pass in a null entity.");
		}

		EventQueue.invokeLater(new PropertiesExecutor(element));
	}

	/**
	 * Displays properties of the given {@link STrace} on the properties panel.
	 * 
	 * @param thread
	 *            The thread to show its properties.
	 */
	private void showSTraceProperties(final STrace thread) {

		this.clear();
		if (thread.getParentOrderIndex() < 0) {
			this.showProperty("properties.class", Language.getInstance().get("properties.adminFloor.thread.process.class"));
		} else {
			this.showProperty("properties.class", Language.getInstance().get("properties.adminFloor.thread.thread.class"));
		}
		this.showProperty("properties.adminFloor.thread.threadId", Long.toString(thread.getThreadId()));
		this.showProperty("properties.adminFloor.thread.traceId", Long.toString(thread.getTraceId()));
		this.showProperty("properties.adminFloor.thread.parentTraceId", Long.toString(thread.getParentTraceId()));
		this.showProperty("properties.adminFloor.thread.hostname", thread.getHostname());
		this.showProperty("properties.adminFloor.thread.sessionId", thread.getSessionId());
		this.showProperty("properties.adminFloor.thread.numEvents", Long.toString(thread.getEvents().size()));

		this.showProperty("properties.adminFloor.thread.OperationSignature", thread.getOperationSignature());
		this.showProperty("properties.adminFloor.thread.tin", Long.toString(thread.getTin()));
		this.showProperty("properties.adminFloor.thread.tout", Long.toString(thread.getTout()));

	}

	/**
	 * Displays properties of the given {@link STraceEvent} on the properties panel.
	 * 
	 * @param event
	 *            The thread event to show its properties.
	 */
	private void showTraceEventProperties(final STraceEvent event) {

		// set table content
		if (event instanceof SMethodExecution) {
			this.showSMethodExecutionProperties((SMethodExecution) event);
		} else if (event instanceof ConstructorExecution) {
			this.showConstructorExecutionProperties((ConstructorExecution) event);
		} else if (event instanceof MonitorEvent) {
			this.showMonitorEventProperties((MonitorEvent) event);
		} else if (event instanceof WaitNotifyEvent) {
			this.showWaitNotifyEventProperties((WaitNotifyEvent) event);
		} else if (event instanceof ThreadCreation) {
			this.showThreadCreationProperties((ThreadCreation) event);
		} else if (event instanceof ThreadTermination) {
			this.showThreadTerminationProperties((ThreadTermination) event);
		}
	}

	/**
	 * Displays properties of the given {@link SMethodExecution} on the properties panel.
	 * 
	 * @param exec
	 *            The method execution to show its properties.
	 */
	private void showSMethodExecutionProperties(final SMethodExecution exec) {

		this.clear();
		this.showProperty("properties.class", Language.getInstance().get("properties.street.method.class"));
		if (exec.getSMethod() != null) {
			if (exec.getSMethod().getKdmSignature() != null) {
				this.showProperty("properties.street.methodKDMSignature", exec.getSMethod().getKdmSignature());
			} else {
				this.showProperty("properties.street.methodKDMSignature", exec.getSMethod().getSimpleSignature());
			}
		}
		if (exec.getParent() == null) {
			this.showProperty("properties.street.parent.kdmSignature", exec.getTrace().getOperationSignature());
		} else if (exec.getParent().getSMethod() != null) {
			if (exec.getParent().getSMethod().getKdmSignature() != null) {
				this.showProperty("properties.street.parent.kdmSignature", exec.getParent().getSMethod().getKdmSignature());
			} else {
				this.showProperty("properties.street.parent.kdmSignature", exec.getParent().getSMethod().getSimpleSignature());
			}
		}
		if (exec.getSource() != null) {
			if (exec.getSource().getSClass().getKdmSignature() != null) {
				this.showProperty("properties.street.sourceClass", exec.getSource().getSClass().getKdmSignature());
			} else {
				this.showProperty("properties.street.sourceClass", exec.getSource().getSClass().getName());
			}
			this.showProperty("properties.street.sourceKiekerObjectId", this.toKiekerIdString(exec.getSource().getKiekerObjectId()));
		}
		if (exec.getTarget().getKiekerObjectId() != -1) { // is static method
			this.showProperty("properties.street.method.targetKiekerObjectId", this.toKiekerIdString(((SMethodExecution) exec).getTarget().getKiekerObjectId()));
		}
		this.showProperty("properties.street.tin", Long.toString(exec.getTin()));
		this.showProperty("properties.street.tout", Long.toString(exec.getTout()));
		this.showProperty("properties.street.execTime", Long.toString(exec.getTout() - exec.getTin()));
		this.showProperty("properties.street.execOrder", Integer.toString(exec.getExecOrder()));
		this.showProperty("properties.street.traceId", Long.toString(exec.getTraceId()));
		this.showProperty("properties.street.threadId", Long.toString(exec.getTrace().getThreadId()));
		this.showProperty("properties.street.orderIndex", Integer.toString(exec.getOrderIndex()));
		this.showProperty("properties.street.ess", Integer.toString(exec.getEss()));
	}

	/**
	 * Displays properties of the given {@link ConstructorExecution} on the properties panel.
	 * 
	 * @param exec
	 *            The operation execution to show its properties.
	 */
	private void showConstructorExecutionProperties(final ConstructorExecution exec) {

		this.clear();

		this.showProperty("properties.class", Language.getInstance().get("properties.street.constructor.class"));
		if (exec.getSMethod() != null) {

			if (exec.getSMethod().getKdmSignature() != null) {
				this.showProperty("properties.street.methodKDMSignature", exec.getSMethod().getKdmSignature());
			} else {
				this.showProperty("properties.street.methodKDMSignature", exec.getSMethod().getSimpleSignature());
			}
		}
		if (exec.getParent() == null) {
			this.showProperty("properties.street.parent.kdmSignature", exec.getTrace().getOperationSignature());
		} else if (exec.getParent().getSMethod() != null) {
			if (exec.getParent().getSMethod().getKdmSignature() != null) {
				this.showProperty("properties.street.parent.kdmSignature", exec.getParent().getSMethod().getKdmSignature());
			} else {
				this.showProperty("properties.street.parent.kdmSignature", exec.getParent().getSMethod().getSimpleSignature());
			}
		}
		if (exec.getSource() != null) {
			if (exec.getSource().getSClass().getKdmSignature() != null) {
				this.showProperty("properties.street.sourceClass", exec.getSource().getSClass().getKdmSignature());
			} else {
				this.showProperty("properties.street.sourceClass", exec.getSource().getSClass().getName());
			}
			this.showProperty("properties.street.sourceKiekerObjectId", this.toKiekerIdString(exec.getSource().getKiekerObjectId()));
		}

		this.showProperty("properties.street.constructor.targetKiekerObjectId",
				this.toKiekerIdString(((ConstructorExecution) exec).getConstruction().getKiekerObjectId()));
		this.showProperty("properties.street.tin", Long.toString(exec.getTin()));
		this.showProperty("properties.street.tout", Long.toString(exec.getTout()));
		this.showProperty("properties.street.execTime", Long.toString(exec.getTout() - exec.getTin()));
		this.showProperty("properties.street.execOrder", Integer.toString(exec.getExecOrder()));
		this.showProperty("properties.street.traceId", Long.toString(exec.getTraceId()));
		this.showProperty("properties.street.threadId", Long.toString(exec.getTrace().getThreadId()));
		this.showProperty("properties.street.orderIndex", Integer.toString(exec.getOrderIndex()));
		this.showProperty("properties.street.ess", Integer.toString(exec.getEss()));
	}

	/**
	 * Displays properties of the given {@link SynchronizationObject} on the properties panel.
	 * 
	 * @param syncObject
	 *            The synchronization object to show its properties.
	 */
	private void showSyncObjectProperties(final SynchronisationObject syncObject) {

		this.clear();
		if (syncObject instanceof MonitorObject) {
			this.showProperty("properties.class", Language.getInstance().get("properties.adminFloor.syncObject.monitor.class"));
		} else if (syncObject instanceof WaitNotifyObject) {
			this.showProperty("properties.class", Language.getInstance().get("properties.adminFloor.syncObject.waitNotify.class"));
		}

		this.showProperty("properties.adminFloor.syncObject.kiekerObjectId", Long.toString(syncObject.getKiekerObjectId()));
		if (syncObject.getRealInstance() != null) {
			if (syncObject.getRealInstance().getSClass().getKdmSignature() != null) {
				this.showProperty("properties.adminFloor.syncObject.KDMSignature",
						this.toKdmSignatureString(syncObject.getRealInstance().getSClass().getKdmSignature()));
			} else {
				this.showProperty("properties.adminFloor.syncObject.name",
						this.toKdmSignatureString(syncObject.getRealInstance().getSClass().getName()));
			}
		}

		// object specific content
		if (syncObject instanceof MonitorObject) {
			final MonitorObject monitor = (MonitorObject) syncObject;
			this.showProperty("properties.adminFloor.syncObject.tin", Long.toString(monitor.getCreator().getTin()));
		} else if (syncObject instanceof WaitNotifyObject) {
			final WaitNotifyObject waitNotify = (WaitNotifyObject) syncObject;
			this.showProperty("properties.adminFloor.syncObject.tin", Long.toString(waitNotify.getCreator().getTin()));
		}
	}

	/**
	 * Displays properties of the given {@link ThreadCreation} on the properties panel.
	 * 
	 * @param exec
	 *            The thread creation event to show its properties.
	 */
	private void showThreadCreationProperties(final ThreadCreation exec) {

		this.clear();
		this.showProperty("properties.class", Language.getInstance().get("properties.street.threadCreation.class"));
		if (exec.getSource() != null) {
			this.showProperty("properties.street.sourceKiekerObjectId", this.toKiekerIdString(exec.getSource().getKiekerObjectId()));
		}
		this.showProperty("properties.street.createdThreadId", Long.toString(exec.getTarget().getThreadId()));
		this.showProperty("properties.street.createdTraceId", Long.toString(exec.getTarget().getTraceId()));
		this.showProperty("properties.street.tin", Long.toString(exec.getTin()));
		this.showProperty("properties.street.execOrder", Integer.toString(exec.getExecOrder()));
		this.showProperty("properties.street.traceId", Long.toString(exec.getTraceId()));
		this.showProperty("properties.street.threadId", Long.toString(exec.getTrace().getThreadId()));
		this.showProperty("properties.street.orderIndex", Integer.toString(exec.getOrderIndex()));
		this.showProperty("properties.street.ess", Integer.toString(exec.getEss()));
	}

	/**
	 * Displays properties of the given {@link ThreadTermination} on the properties panel.
	 * 
	 * @param exec
	 *            The thread termination event to show its properties.
	 */
	private void showThreadTerminationProperties(final ThreadTermination exec) {

		this.clear();
		this.showProperty("properties.class", Language.getInstance().get("properties.street.threadTermination.class"));
		this.showProperty("properties.street.tin", Long.toString(exec.getTin()));
		this.showProperty("properties.street.execOrder", Integer.toString(exec.getExecOrder()));
		this.showProperty("properties.street.traceId", Long.toString(exec.getTraceId()));
		this.showProperty("properties.street.threadId", Long.toString(exec.getTrace().getThreadId()));
		this.showProperty("properties.street.ess", Integer.toString(exec.getEss()));
	}

	/**
	 * Displays properties of the given {@link WaitNotifyEvent} on the properties panel.
	 * 
	 * @param exec
	 *            The wait-/notify event to show its properties.
	 */
	private void showWaitNotifyEventProperties(final WaitNotifyEvent exec) {

		this.clear();
		this.showProperty("properties.class", Language.getInstance().get("properties.street.waitNotify.class"));
		if (exec.getSource() != null) {
			this.showProperty("properties.street.sourceKiekerObjectId", this.toKiekerIdString(exec.getSource().getKiekerObjectId()));
		}
		this.showProperty("properties.street.targetKiekerObjectId", this.toKiekerIdString(exec.getTarget().getKiekerObjectId()));
		this.showProperty("properties.street.type", exec.getType().getName().toLowerCase());
		this.showProperty("properties.street.texec", Long.toString(exec.getTin()));
		this.showProperty("properties.street.execOrder", Integer.toString(exec.getExecOrder()));
		this.showProperty("properties.street.traceId", Long.toString(exec.getTraceId()));
		this.showProperty("properties.street.threadId", Long.toString(exec.getTrace().getThreadId()));
		this.showProperty("properties.street.orderIndex", Integer.toString(exec.getOrderIndex()));
		this.showProperty("properties.street.ess", Integer.toString(exec.getEss()));
	}

	/**
	 * Displays properties of the given {@link MonitorEvent} on the properties panel.
	 * 
	 * @param exec
	 *            The monitor event to show its properties.
	 */
	private void showMonitorEventProperties(final MonitorEvent exec) {

		this.clear();
		this.showProperty("properties.class", Language.getInstance().get("properties.street.monitor.class"));
		if (exec.getSource() != null) {
			this.showProperty("properties.street.sourceKiekerObjectId", this.toKiekerIdString(exec.getSource().getKiekerObjectId()));
		}
		this.showProperty("properties.street.targetKiekerObjectId", this.toKiekerIdString(exec.getTarget().getKiekerObjectId()));
		this.showProperty("properties.street.type", exec.getType().getName().toLowerCase());
		this.showProperty("properties.street.texec", Long.toString(exec.getTin()));
		this.showProperty("properties.street.execOrder", Integer.toString(exec.getExecOrder()));
		this.showProperty("properties.street.traceId", Long.toString(exec.getTraceId()));
		this.showProperty("properties.street.threadId", Long.toString(exec.getTrace().getThreadId()));
		this.showProperty("properties.street.orderIndex", Integer.toString(exec.getOrderIndex()));
		this.showProperty("properties.street.ess", Integer.toString(exec.getEss()));
	}

	/**
	 * Displays properties of the given {@link SClassInstance} on the properties panel.
	 * 
	 * @param appObject
	 *            The application object to show its properties.
	 */
	private void showRegFloorProperties(final SClassInstance appObject) {

		this.clear();
		if (appObject.equals(appObject.getSClass().getInstances().get(0))) {
			this.showProperty("properties.class", Language.getInstance().get("properties.regFloor.class.classObject"));
		} else {
			this.showProperty("properties.class", Language.getInstance().get("properties.regFloor.class.classInstance"));
			this.showProperty("properties.regFloor.kiekerObjectId", this.toKiekerIdString(appObject.getKiekerObjectId()));
			this.showProperty("properties.regFloor.tin", Long.toString(((SClassInstance) appObject).getConstructorExecution().getTin()));
		}
		this.showProperty("properties.regFloor.class.name", appObject.getSClass().getName());
		if (appObject.getSClass().getKdmSignature() != null) {
			this.showProperty("properties.regFloor.class.signature", appObject.getSClass().getKdmSignature());
		}
		this.showProperty("properties.regFloor.class.package", appObject.getSClass().getSPackage().getName());
		this.showProperty("properties.regFloor.class.KDMId", appObject.getSClass().getKdmId());
		this.showProperty("properties.regFloor.class.numAttributes", Integer.toString(appObject.getSClass().getNumAttributes()));
		this.showProperty("properties.regFloor.class.numMethods", Integer.toString(appObject.getSClass().getSMethods().size()));
		this.showProperty("properties.regFloor.class.numRelations.outgoing", Integer.toString(DataModelUtils.getOutgoingRelations(appObject.getSClass()).size()));
		this.showProperty("properties.regFloor.class.numRelations.incoming", Integer.toString(DataModelUtils.getIncomingRelations(appObject.getSClass()).size()));
	}

	/**
	 * Displays properties of the given {@link Relation} on the properties panel.
	 * 
	 * @param relation
	 *            The relation to show its properties.
	 */
	private void showRelationProperties(final Relation relation) {

		this.clear();

		// set table content
		this.showProperty("properties.class", Language.getInstance().get("properties.cable.class"));
		if (!"<n/e>".equalsIgnoreCase(relation.getName())) {
			this.showProperty("properties.cable.name", relation.getName());
		}
		this.showProperty("properties.cable.KDMId", relation.getKdmId());
		if (relation.getType() == RelationType.ASSOCIATION && relation.getKdmSignature() != null) {
			this.showProperty("properties.cable.name", relation.getKdmSignature());
		}
		if (relation.getSource().getKdmSignature() != null && relation.getSource().getKdmSignature() != null) {
			this.showProperty("properties.cable.sourceKdmSignature", relation.getSource().getKdmSignature());
			this.showProperty("properties.cable.targetKdmSignature", relation.getTarget().getKdmSignature());
		} else {
			this.showProperty("properties.cable.sourceKdmSignature", relation.getSource().getName());
			this.showProperty("properties.cable.targetKdmSignature", relation.getTarget().getName());
		}
		this.showProperty("properties.cable.type", Language.getInstance().get("properties.cable.type." + relation.getType().getName().toLowerCase()));
	}

	/**
	 * Displays properties of the given {@link SPackage} on the properties panel.
	 * 
	 * @param pack
	 *            The package to show its properties.
	 */
	private void showRegDistrictProperties(final SPackage pack) {

		this.clear();
		this.showProperty("properties.class", Language.getInstance().get("properties.district.class"));
		this.showProperty("properties.district.name", pack.getName());
		this.showProperty("properties.district.level", Integer.toString(pack.getLevel()));
		this.showProperty("properties.district.numSubPackages", Integer.toString(pack.getSubPackages().size()));
		this.showProperty("properties.district.numSubClasses", Integer.toString(pack.getSClasses().size()));
	}

	/**
	 * Appends the specified key-value pair to the end of the properties panel, i.e. displays them in a separate row.
	 * 
	 * @param key
	 *            The properties key of the value to display.
	 * @param value
	 *            The value to display.
	 */
	private void showProperty(final String key, final String value) {

		this.tableModel.addRow(new String[] { Language.getInstance().get(key), value, });
		this.keys.add(key);
	}

	/**
	 * Removes all content currently added to this table. This {@link #tableModel} will be empty of rows when this method returns.
	 */
	private void clear() {

		for (int i = 0; i < this.keys.size(); i++) {
			this.tableModel.removeRow(0);
		}
		this.keys.clear();
	}

	/**
	 * Returns the properties value string to use when a Kieker Id is supposed to be displayed. It returns {@code <n/a>}, if the specified Kieker-Id is negative or
	 * (otherwise) the id as a string itself.
	 * 
	 * @param kiekerId
	 *            The Kieker-Id to return its string to display.
	 * @return The string to be displayed instead of the raw id itself.
	 */
	private String toKiekerIdString(final int kiekerId) {

		if (kiekerId < 0) {
			return "<n/a>";
		}
		return Integer.toString(kiekerId);
	}

	/**
	 * Returns the properties value string to use when a KDM signature is supposed to be displayed. It returns {@code <n/a>}, if the specified signature is null or
	 * empty and (otherwise) the unchanged signature itself.
	 * 
	 * @param kdmSignature
	 *            The KDM signature to return its string to display.
	 * @return The string to be displayed instead of the raw KDM signature itself.
	 */
	private String toKdmSignatureString(final String kdmSignature) {

		if (kdmSignature == null || "".equals(kdmSignature)) {
			return "<n/a>";
		}
		return kdmSignature;
	}

	/**
	 * A properties executor to display a specified visualizable element's properties in the properties panel.
	 * 
	 * @author pdo
	 * 
	 */
	private final class PropertiesExecutor implements Runnable {

		/** The entity to display its properties. */
		private final VisualizableElement element;

		/**
		 * Creates a new {@link PropertiesController} instance displaying the specified visualizable element's properties in the properties panel.
		 * 
		 * @param elem
		 *            The entity to display its properties.
		 */
		private PropertiesExecutor(final VisualizableElement elem) {

			this.element = elem;
		}

		@Override
		public void run() {
			synchronized (PropertiesController.this.lock) {
				if (this.element instanceof Relation) {
					PropertiesController.this.showRelationProperties((Relation) this.element);
				} else if (this.element instanceof STraceEvent) {
					PropertiesController.this.showTraceEventProperties((STraceEvent) this.element);
				} else if (this.element instanceof SClassInstance) {
					PropertiesController.this.showRegFloorProperties((SClassInstance) this.element);
				} else if (this.element instanceof SynchronisationObject) {
					PropertiesController.this.showSyncObjectProperties((SynchronisationObject) this.element);
				} else if (this.element instanceof STrace) {
					PropertiesController.this.showSTraceProperties((STrace) this.element);
				} else if (this.element instanceof SPackage) {
					PropertiesController.this.showRegDistrictProperties((SPackage) this.element);
				} else {
					throw new IllegalArgumentException("Unsupported visual entity type passed in: " + this.element.getClass().getName() + ".");
				}

				PropertiesController.this.displayed = this.element;
			}
		}
	}

	/**
	 * A rebuild executor rebuilding the entire properties panel controlled by this properties controller on demand.
	 * 
	 * @author pdo
	 * 
	 */
	private final class RebuildExecutor implements Runnable {

		/**
		 * Creates a new {@link RebuildExecutor} instance.
		 */
		private RebuildExecutor() {}

		@Override
		public void run() {
			synchronized (PropertiesController.this.lock) {
				if (PropertiesController.this.tableModel != null) {

					// rebuild column identifiers
					PropertiesController.this.tableModel.setColumnIdentifiers(new String[] { Language.getInstance().get("properties.identifier.column1"),
						Language.getInstance().get("properties.identifier.column2"), });

					// rebuild row identifiers
					if (PropertiesController.this.displayed != null) {
						PropertiesController.this.showProperties(PropertiesController.this.displayed);
					}
				}
			}
		}
	}

	/**
	 * A content removal executor removing all properties currently displayed on the properties panel currently controlled by the properties controller on demand.
	 * 
	 * @author pdo
	 * 
	 */
	private final class ContentRemovalExecutor implements Runnable {

		/**
		 * Creates a new {@link ContentRemovalExecutor} instance.
		 */
		private ContentRemovalExecutor() {}

		@Override
		public void run() {
			synchronized (PropertiesController.this.lock) {
				PropertiesController.this.clear();
				PropertiesController.this.displayed = null;
			}
		}
	}

}
