/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package logic;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import org.apache.log4j.Logger;

import configuration.Configuration;

import data.dynamics.OperationExecution;
import data.dynamics.STrace;
import data.dynamics.STraceEvent;
import data.dynamics.ThreadCreation;
import data.dynamics.ThreadTermination;
import data.synchronisation.MonitorEvent;
import data.synchronisation.ThreadJoin;
import data.synchronisation.WaitNotifyEvent;

import utils.NodeObject;
import utils.comparators.JTreeNodeByTinComparator;
import utils.comparators.JTreeNodeByTraceComparator;

/**
 * A trace tree view controller used to add {@link STraceEvent} representing tree nodes to a {@link TraceTreeView} and handle node selections.
 * 
 * @author pdo
 * 
 */
public final class TraceTreeViewController {

	// =========================== Attributes =================================

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(TraceTreeViewController.class);

	/** The underlying {@link JTree} instance of the {@link TraceTreeView} currently being controlled by this trace tree view controller. */
	private JTree tree;

	/** The underlying root {@link DefaultMutableTreeNode} instance of the {@link TraceTreeView} currently being controlled by this trace tree view controller. */
	private DefaultMutableTreeNode root;

	/** The tree model of the underlying tree in the trace tree view being controlled by this controller. */
	private DefaultTreeModel model;

	/** A list of all tree nodes currently added to the controlled tree. */
	private List<DefaultMutableTreeNode> nodes = new LinkedList<DefaultMutableTreeNode>();

	/** A list of all thread representing tree nodes. */
	private List<DefaultMutableTreeNode> threadnodes = new ArrayList<DefaultMutableTreeNode>(0); // prevents null pointer exception when nothing is loaded

	/** A mapping from each {@link STraceEvent} displayed in this tree to their representing tree nodes. */
	private Map<STraceEvent, DefaultMutableTreeNode> eventToNode;

	/** The current tree structure as chosen by the user. */
	private TreeStructure currentStructure = Configuration.getInstance().getTraceViewerLayout();

	/** The {@link STraceEvent} currently being selected in the trace view. */
	private STraceEvent selected;

	// ========================== Constructor =================================

	/**
	 * Protected constructor. This class is not intended for instantiation from outside of this package.
	 */
	protected TraceTreeViewController() {}

	// ============================ Methods ===================================

	/**
	 * Initializes this trace tree view controller with the given tree and root tree node. This tree view controller controls the trace tree view with the given
	 * tree as underlying data structure when this method returns.
	 * 
	 * @param jtree
	 *            The {@link JTree} of the trace tree view to be controlled.
	 * @param treeModel
	 *            The tree model object of the trace tree view's tree to be controlled.
	 * @throws IllegalArgumentException
	 *             If either the given tree or its model is {@code null}.
	 */
	public void init(final JTree jtree, final DefaultTreeModel treeModel) throws IllegalArgumentException {

		if (jtree == null) {
			throw new IllegalArgumentException("You cannot pass in a null tree.");
		}
		if (treeModel == null) {
			throw new IllegalArgumentException("You cannot pass in a null tree model.");
		}

		this.root = (DefaultMutableTreeNode) treeModel.getRoot();
		this.model = treeModel;
		this.tree = jtree;

		jtree.setExpandsSelectedPaths(true);
		jtree.setRootVisible(false);
		jtree.setEditable(false);
	}

	/**
	 * Sets the current {@link TreeStructure} of the controlled trace tree view. This might cause the entire tree to be rebuild.
	 * 
	 * @param structure
	 *            The new structure of the tree controlled by this trace tree view controller.
	 * @throws IllegalArgumentException
	 *             If the tree structure passed in is {@code null}.
	 */
	public void setTreeStructure(final TreeStructure structure) throws IllegalArgumentException {

		if (structure == null) {
			throw new IllegalArgumentException("The tree structure cannot be null.");
		}

		if (this.currentStructure != structure) {
			this.removeAllNodes();
			this.currentStructure = structure;
			this.buildTree();

			// reselect node
			if (this.selected != null) {
				this.selectNode(this.selected);
			}
		}
	}

	/**
	 * Removes all nodes from the default tree model ({@link #model}) currently being controlled by this trace tree view controller.
	 */
	private void removeAllNodes() {

		while (this.root.getChildCount() > 0) {
			this.model.removeNodeFromParent((DefaultMutableTreeNode) this.root.getFirstChild());
		}

		for (DefaultMutableTreeNode node : this.nodes) {
			while (node.getChildCount() > 0) {
				this.model.removeNodeFromParent((DefaultMutableTreeNode) node.getFirstChild());
			}
		}
		
		this.selected = null;
	}

	/**
	 * Selects the node representing the given {@link STraceEvent} in the trace tree view.
	 * 
	 * @param event
	 *            The event to select its representing node or {@code null} to unselect the node currently being selected.
	 * @throws IllegalArgumentException
	 *             If there is no node representing the given trace event in this tree.
	 */
	public void selectNode(final STraceEvent event) throws IllegalArgumentException {

		if (event == null) {
			this.tree.clearSelection();
			this.selected = null;
		} else {
			final DefaultMutableTreeNode node = this.eventToNode.get(event);
			if (node != null) {
				final TreePath path = new TreePath(node.getPath());
				this.tree.setSelectionPath(path);
				this.tree.expandPath(path);
				this.tree.validate();
				this.selected = event;
			}
		}
	}

	/**
	 * Handles a {@link TreeSelectionEvent} in the underlying tree in the controlled trace tree view.
	 */
	public void handleTreeSelection() {

		// request focus on city canvas
		Logic.getInstance().getCanvasController().requestFocus();

		// get selected node
		final DefaultMutableTreeNode selectedNode = (DefaultMutableTreeNode) this.tree.getLastSelectedPathComponent();
		if (selectedNode != null) {

			// display properties and jump to method execution in trace
			final NodeObject nodeObject = (NodeObject) selectedNode.getUserObject();
			if (nodeObject.getTraceEvent() == null) {
				Logic.getInstance().getPropertiesController().showProperties(nodeObject.getTrace());
			} else {
				Logic.getInstance().getPropertiesController().showProperties(nodeObject.getTraceEvent());
				Logic.getInstance().getCityController().jumpTo(nodeObject.getTraceEvent());
				this.selected = nodeObject.getTraceEvent();
			}
		}
	}

	/**
	 * Sets the content of the controlled trace tree view's underlying tree. It will create a new node for each trace event in the specified threads and add them to
	 * the tree according to the layout defined in {@link #currentStructure}.
	 * 
	 * @param traces
	 *            The list of traces to be displayed in the tree.
	 * @throws IllegalArgumentException
	 *             If the traces list passed in or any thread contained in it is {@code null},
	 */
	public void setContent(final List<STrace> traces) throws IllegalArgumentException {

		if (traces == null) {
			throw new IllegalArgumentException("You cannot set a null content.");
		}

		// prepare data structure
		this.prepareDataStructure(traces);
		this.selected = null;

		// create nodes
		this.threadnodes = new ArrayList<DefaultMutableTreeNode>(traces.size());
		for (STrace thread : traces) {
			if (thread == null) {
				throw new IllegalArgumentException("You cannot display a null node.");
			}
			// threads are sorted according to their tin attribute value
			this.threadnodes.add(new DefaultMutableTreeNode(new NodeObject("Thread " + thread.getThreadId(), thread, null))); // null identifies thread nodes

			for (STraceEvent event : thread.getEvents()) {
				// create and add node
				final DefaultMutableTreeNode node = this.createNode(thread, event);
				this.nodes.add(node);
				this.eventToNode.put(event, node);
			}
		}

		// build tree
		this.buildTree();
	}

	/**
	 * Removes all nodes from the trace tree view currently controlled by this controller.
	 */
	public void removeAllContent() {

		this.setContent(new LinkedList<STrace>());
	}

	/**
	 * Prepares the data structure of this controller in order to set the contents of the controlled tree ({@link #setContent(List)}). If will remove all currently
	 * displayed tree nodes from the tree and clear {@link #eventToNode} and {@link #nodes}.
	 * 
	 * @param traces
	 *            The list of traces to be set as the new content of the trace tree view currently being controlled by this controller.
	 */
	private void prepareDataStructure(final List<STrace> traces) {

		// remove all previous nodes
		this.removeAllNodes();

		// determine number of thread events to display
		int count = 0;
		for (STrace thread : traces) {
			count += thread.getEvents().size();
		}

		// prepare attributes
		this.eventToNode = new HashMap<STraceEvent, DefaultMutableTreeNode>((int) (count * 4 / 3)); // ideal hash map size
		this.nodes = new LinkedList<DefaultMutableTreeNode>();
	}

	/**
	 * Creates the tree node representing the specified trace event.
	 * 
	 * @param trace
	 *            The trace the specified thread event belongs to.
	 * @param event
	 *            The trace event to create its representing tree node
	 * @return The tree node representing the specified thread event.
	 */
	private DefaultMutableTreeNode createNode(final STrace trace, final STraceEvent event) {

		final DefaultMutableTreeNode node;
		if (event instanceof ThreadCreation) {
			node = new DefaultMutableTreeNode(new NodeObject("<new Thread>", trace, event));
		} else if (event instanceof ThreadJoin) {
			node = new DefaultMutableTreeNode(new NodeObject("<Thread-Join>", trace, event));
		} else if (event instanceof ThreadTermination) {
			node = new DefaultMutableTreeNode(new NodeObject("<Thread Termination>", trace, event));
		} else if (event instanceof MonitorEvent) {
			node = new DefaultMutableTreeNode(new NodeObject("<" + ((MonitorEvent) event).getType().getName().toLowerCase() + ">", trace, event));
		} else if (event instanceof WaitNotifyEvent) {
			node = new DefaultMutableTreeNode(new NodeObject("<" + ((WaitNotifyEvent) event).getType().getName().toLowerCase() + ">", trace, event));
		} else if (event instanceof OperationExecution) {
			final OperationExecution exec = (OperationExecution) event;
			if (exec.getSMethod() == null) {
				node = new DefaultMutableTreeNode(new NodeObject("<n/a>", trace, event));
			} else {
				node = new DefaultMutableTreeNode(new NodeObject(exec.getSMethod().getSimpleSignature(), trace, event));
			}
		} else {
			final String error = "Unsupported thread event detected: " + event.getClass().getName() + ".";
			LOGGER.error(error);
			throw new IllegalStateException(error);
		}

		return node;
	}

	/**
	 * Builds the tree of the trace tree view currently being controlled. This method will call {@link #buildFlatTree()} or {@link #buildNestedTree()} depending on
	 * the current value of {@link #currentStructure}. This method has no effect, if this controller currently does not control any trace viewer.
	 */
	private void buildTree() {

		if (this.tree != null) {
			// controller controls some trace tree viewer
			switch (this.currentStructure) {
			case CHRONOLOGICAL:
				this.buildFlatTree();
				break;
			case HIERARCHICAL:
				this.buildNestedTree();
				break;
			default:
				LOGGER.warn("Unsupported tree structure detected: " + this.currentStructure.name() + ".");
			}

			// expand entire tree
			this.tree.expandPath(new TreePath(this.root.getPath()));
			for (int i = 0; i < this.tree.getRowCount(); i++) {
				this.tree.expandRow(i);
			}

			this.tree.validate();
		}
	}

	/**
	 * Builds the tree of the trace tree viewer according to the {@link TreeStructure#CHRONOLOGICAL} style.
	 */
	private void buildFlatTree() {

		// remove thread nodes
		final List<DefaultMutableTreeNode> threadNodes = new LinkedList<DefaultMutableTreeNode>();
		for (DefaultMutableTreeNode node : this.nodes) {
			if (((NodeObject) node.getUserObject()).getTraceEvent() == null) {
				// node is a thread node
				threadNodes.add(node);
			}
		}
		this.nodes.removeAll(threadNodes);

		// display thread event nodes
		Collections.sort(this.nodes, new JTreeNodeByTinComparator());
		final boolean isAppended = !this.nodes.isEmpty() && ((NodeObject) this.nodes.get(0).getUserObject()).toString().contains(" <");
		for (DefaultMutableTreeNode node : this.nodes) {
			final NodeObject nodeObject = (NodeObject) node.getUserObject();
			if (!isAppended) {
				nodeObject.setName(nodeObject.toString() + "     <" + nodeObject.getTrace().getThreadId() + ">");
			}
			this.model.insertNodeInto(node, this.root, this.root.getChildCount());
		}
	}

	/**
	 * Builds the tree of the trace tree viewer according to the {@link TreeStructure#HIERARCHICAL} style.
	 */
	private void buildNestedTree() {

		Collections.sort(this.nodes, new JTreeNodeByTraceComparator());
		final List<DefaultMutableTreeNode> traceNodes = new LinkedList<DefaultMutableTreeNode>();

		for (DefaultMutableTreeNode node : this.threadnodes) {
			this.model.insertNodeInto(node, this.root, this.root.getChildCount());
		}

		final boolean isAppended = !this.nodes.isEmpty() && ((NodeObject) this.nodes.get(0).getUserObject()).toString().contains(" <");
		for (DefaultMutableTreeNode node : this.nodes) {
			final NodeObject object = (NodeObject) node.getUserObject();
			if (isAppended) {
				object.setName(object.toString().substring(0, object.toString().indexOf("     <")));
			}

			final STraceEvent parent = object.getTraceEvent().getParent();
			if (parent == null) {
				final DefaultMutableTreeNode threadNode = this.threadnodes.get(object.getTrace().getExecOrder());
				this.model.insertNodeInto(node, threadNode, threadNode.getChildCount());
			} else {
				final DefaultMutableTreeNode parentNode = this.eventToNode.get(parent);
				this.model.insertNodeInto(node, parentNode, parentNode.getChildCount());
			}
		}

		this.nodes.addAll(traceNodes);
	}

	/**
	 * Returns the {@link STraceEvent} instance whose node is currently selected in the trace tree view or {@code null}, if no node is selected.
	 * 
	 * @return The trace event instance whose node is currently selected in the trace tree view.
	 */
	public STraceEvent getSelected() {
		return this.selected;
	}

	// ========================== Inner Classes ===============================

	/**
	 * An enumeration of all node layouts currently supported in a {@link TraceTreeView}.
	 * 
	 * @author pdo
	 * 
	 */
	public static enum TreeStructure {

		/** All displayed events will be ordered by their method entry timestamp (tin) and set as direct children of the tree's root node. */
		CHRONOLOGICAL,

		/**
		 * All displayed events will be arranged according their nesting structure in the call graph of each process or thread. I.e. if thread event n is the direct
		 * parent of thread event m, then n's node will be the direct parent of m's node in the trace view graph. Each first event of a trace will be the direct
		 * child of the tree's root node. The single traces are ordered by their trace IDs in descending order.
		 */
		HIERARCHICAL,
	}

}
