/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package logic.executors;

import java.io.File;

import javax.swing.JOptionPane;

import org.apache.log4j.Logger;

import kieker.analysis.AnalysisController;
import kieker.analysis.exception.AnalysisConfigurationException;
import kieker.analysis.plugin.filter.flow.EventRecordTraceReconstructionFilter;
import kieker.analysis.plugin.reader.filesystem.FSReader;
import kieker.common.configuration.Configuration;

import configuration.Language;

import importer.KiekerImportFilter;

import logic.Logic;

/**
 * A kieker import executor performing the import of Kieker files into the Synchrovis application in a separate thread.
 * 
 * @author pdo
 * 
 */
public final class KiekerImportExecutor extends Thread {
	
	// =========================== Attributes =================================

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(KiekerImportExecutor.class);

	/** The {@link KiekerImportFilter} performing the import. */
	private final KiekerImportFilter filter;

	/** The folder containing the Kieker traces to be imported. */
	private final File file;

	// ========================== Constructor =================================
	
	/**
	 * Constructs a new {@link KiekerImportExecutor} performing the import of all Kieker files in the given folder using the specified kieker import filter in a
	 * separate thread.
	 * 
	 * @param kiekerImportFilter
	 *            The Kieker import filter to use when performing the import.
	 * @param file
	 *            The folder containing all Kieker trace files to import.
	 * @throws IllegalArgumentException
	 *             If the Kieker import filter is {@code null} or the file passed in is either {@code null}, does not exist on disc, or does not represent a folder.
	 */
	public KiekerImportExecutor(final KiekerImportFilter kiekerImportFilter, final File file) throws IllegalArgumentException {

		super();

		if (kiekerImportFilter == null) {
			throw new IllegalArgumentException("Kieker traces cannot be imported with a null filter.");
		}
		if (file == null || !file.isDirectory()) {
			throw new IllegalArgumentException("The file passed in is either null, does not exist on disc, or does not represent a folder.");
		}

		this.filter = kiekerImportFilter;
		this.file = file;
	}

	// ============================ Methods ===================================
	
	@Override
	public void run() {
		try {
			Logic.getInstance().getMenuBarController().setDynamicsItemEnabled(false);
			// create new file system Reader
			final Configuration confFSReader = new Configuration();
			confFSReader.setProperty(FSReader.CONFIG_PROPERTY_NAME_INPUTDIRS, this.file.getAbsolutePath());
			final FSReader reader = new FSReader(confFSReader);

			// create new event record trace reconstruction filter
			final Configuration confEventTraceReconstructionFilter = new Configuration();
			final EventRecordTraceReconstructionFilter eventTraceReconstructionFilter = new EventRecordTraceReconstructionFilter(
					confEventTraceReconstructionFilter);

			// create new analysis controller
			final AnalysisController analysisController = new AnalysisController("");

			// register all filters and file system reader
			analysisController.registerReader(reader);
			analysisController.registerFilter(eventTraceReconstructionFilter);
			analysisController.registerFilter(this.filter);

			// connect everything
			analysisController.connect(reader, FSReader.OUTPUT_PORT_NAME_RECORDS, eventTraceReconstructionFilter,
					EventRecordTraceReconstructionFilter.INPUT_PORT_NAME_TRACE_RECORDS);
			analysisController.connect(eventTraceReconstructionFilter, EventRecordTraceReconstructionFilter.OUTPUT_PORT_NAME_TRACE_VALID,
					this.filter, KiekerImportFilter.INPUT_PORT_NAME);

			// run import
			analysisController.run();
			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Successfully imported dynamics from file " + this.file.getAbsolutePath() + ".");
			}
		} catch (AnalysisConfigurationException | IllegalStateException e) {
			LOGGER.error("Failed to import Kieker traces.", e);
			JOptionPane.showMessageDialog(null, Language.getInstance().get("kieker.error.message"), Language.getInstance().get("kieker.error.title"),
					JOptionPane.ERROR_MESSAGE);
		}
	}

}
