/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package main;

import org.apache.log4j.Logger;

import importer.kdm.AbstractKDMImporter;
import importer.kdm.KDMImporter;

import logic.Logic;

import models.city.CityModel;
import models.data.DataModel;
import models.metaphor.MetaphorImporter;
import models.metaphor.MetaphorModel;
import models.view.ViewImporter;
import models.view.ViewModel;

import observer.ObserverManager;

import visual.gui.Gui;

/**
 * The main class of Synchrovis. It contains the {@link #main(String[])} method to launch this application.
 * 
 * @author pdo
 * 
 */
public final class Main {

	// =========================== Attributes =================================

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(Main.class);

	// ========================== Constructor =================================

	/**
	 * This class is not intended for instantiation since it contains static methods only.
	 */
	private Main() {}

	// ============================ Methods ===================================

	/**
	 * The entry point of Synchrovis. It launches the application.
	 * 
	 * @param args
	 *            Ignored.
	 */
	public static void main(final String[] args) {

		LOGGER.info("Application started.\n");
		
		// create model instances
		final DataModel dataModel = new DataModel();
		final CityModel sceneModel = new CityModel();
		final MetaphorModel metaphorModel = new MetaphorModel();
		final ViewModel viewModel = new ViewModel();

		// create data model importers
		final AbstractKDMImporter kdmImporter = new KDMImporter(dataModel);
		Logic.getInstance().getSavesManager().init(dataModel);

		// create and initialize model to model importers
		final MetaphorImporter metaphorImporter = new MetaphorImporter(metaphorModel);
		final ViewImporter viewImporter = new ViewImporter(viewModel);
		Logic.getInstance().getCityController().init(sceneModel);

		// connect by observer pattern
		ObserverManager.getInstance().connect(dataModel, sceneModel, metaphorModel, metaphorImporter, viewImporter);

		// create GUI
		new Gui(kdmImporter, dataModel, viewModel);
	}

}
