/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package models.city;

import org.apache.log4j.Logger;

import data.dynamics.ConstructorExecution;
import data.dynamics.SClassInstance;
import data.dynamics.STrace;
import data.dynamics.STraceEvent;
import data.dynamics.ThreadCreation;
import data.statics.Relation;
import data.statics.SPackage;
import data.synchronisation.MonitorEvent;
import data.synchronisation.MonitorMethod;
import data.synchronisation.MonitorObject;
import data.synchronisation.WaitNotifyEvent;
import data.synchronisation.WaitNotifyObject;

import metaphor.entities.AdministrationFloor;
import models.data.DataModelUtils;

import observer.AbstractObservable;
import observer.ObservableAction;
import utils.LoggerUtils;

/**
 * The data structure containing all data objects currently visible on screen.
 * 
 * @author pdo
 * 
 */
public class CityModel extends AbstractObservable {

	// ============================= Attributes ===============================

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(CityModel.class);

	// ============================= Constructor ==============================

	/**
	 * Default constructor. It creates a new {@link CityModel} instance.
	 */
	public CityModel() {

		super();
	}

	// =============================== Methods ================================

	/**
	 * Removes all objects in this {@link CityModel} instance. This scene model will be empty after this call returns.
	 */
	public void clearAll() {

		super.notifyObservers(ObservableAction.ALL_CLEARED, null);
	}

	/**
	 * Sets the common root package in this {@link CityModel} instance. This operation will add all contained sub-packages to this scene model automatically.
	 * 
	 * @param root
	 *            The common root package of all packages to add to this scene model.
	 * @throws IllegalArgumentException
	 *             If the given package is {@code null}.
	 */
	public void setRootPackage(final SPackage root) throws IllegalArgumentException {

		if (root == null) {
			throw new IllegalArgumentException("You cannot set a null package as root package.");
		}

		// initialize data structure
		for (SPackage sPack : root.getSubPackages()) {
			this.initClassInstancesList(sPack);
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully set package with values: " + LoggerUtils.toLogString(root) + " as root package.");
		}
		super.notifyObservers(ObservableAction.STATICS_LOADED, root);
	}

	/**
	 * initializes the list of {@link SClassInstance} objects of each class contained in the given package.
	 * 
	 * @param sPack
	 *            The package to initialize all class instances lists in.
	 */
	private void initClassInstancesList(final SPackage sPack) {

		// recursive call for each sub-package
		for (SPackage subPack : sPack.getSubPackages()) {
			this.initClassInstancesList(subPack);
		}
	}

	/**
	 * Adds the given {@link SClassInstance} instance to this {@link CityModel} instance. This method has no effect, if the given object already exists in this
	 * scene model.
	 * 
	 * @param instance
	 *            The dynamic class instance to add to this scene model.
	 */
	private void addClassInstance(final SClassInstance instance) {

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully added new dynamic class instance with values: " + LoggerUtils.toLogString(instance) + " to this scene model.");
		}
		super.notifyObservers(ObservableAction.CLASS_INSTANCE_ADDED, instance);
	}

	/**
	 * Adds the given {@link STraceEvent} instance to this {@link CityModel} instance. This method has no effect, if the given object already exists in this scene
	 * model.
	 * 
	 * @param event
	 *            The thread event to add to this scene model
	 * @param invocationOnly
	 *            {@code True} to visualize the invocation of the specified thread event only, {@code false} otherwise.
	 * @throws IllegalArgumentException
	 *             If the given event is {@code null}.
	 */
	public void addThreadEvent(final STraceEvent event, final boolean invocationOnly) throws IllegalArgumentException {

		if (event == null) {
			throw new IllegalArgumentException("You cannot add a null event.");
		}

		// add constructed or related objects
		if (!invocationOnly) {
			if (event instanceof MonitorEvent) {
				final MonitorEvent monitor = (MonitorEvent) event;
				if (monitor.getTarget().getCreator().equals(monitor)) {
					this.addMonitorObject(monitor.getTarget());
				}

				// handle posters
				final AdministrationFloor floor = (AdministrationFloor) monitor.getTarget().getMetaphorObject();
				if (monitor.getType() == MonitorMethod.RELEASE) {
					DataModelUtils.removeLock(monitor.getTrace(), monitor.getTarget());
					if (!DataModelUtils.hasLock(monitor.getTrace(), monitor.getTarget())) {
						floor.getPoster().setColor(null);
						this.applyMonitorStatusChanged(floor);
					}
				} else if (monitor.getType() == MonitorMethod.ACQUIRE) {
					final AdministrationFloor thread = (AdministrationFloor) event.getTrace().getMetaphorObject();
					if (!DataModelUtils.hasLock(monitor.getTrace(), monitor.getTarget())) {
						floor.getPoster().setColor(thread.getPoster().getColor());
						this.applyMonitorStatusChanged(floor);
					}
					monitor.getTrace().getLocks().add(monitor);
				}
			} else if (event instanceof WaitNotifyEvent) {
				final WaitNotifyEvent wait = (WaitNotifyEvent) event;
				if (wait.getTarget().getCreator().equals(wait)) {
					this.addWaitNotifyObject(wait.getTarget());
				}
			} else if (event instanceof ThreadCreation) {
				this.addThreadObject(((ThreadCreation) event).getTarget());				
			} else if (event instanceof ConstructorExecution) {
				this.addClassInstance(((ConstructorExecution) event).getConstruction());
			}
		}

		super.notifyObservers(ObservableAction.THREAD_EVENT_ADDED, event);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully added new method instance with values: " + LoggerUtils.toLogString(event) + " to this scene model.");
		}
	}

	/**
	 * Removes the given {@link STraceEvent} instance from this {@link CityModel} instance. This method has no effect, if the given object is not contained in this
	 * model instance.
	 * 
	 * @param event
	 *            The thread event to remove from this scene model.
	 * @param invocationOnly
	 *            {@code False} to remove any objects 'created' by the specified thread event as well, {@code true} otherwise.
	 * @throws IllegalArgumentException
	 *             If the given event is {@code null}.
	 */
	public void removeTraceEvent(final STraceEvent event, final boolean invocationOnly) throws IllegalArgumentException {

		if (event == null) {
			throw new IllegalArgumentException("You cannot remove a null event.");
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully removed traceElement with values: " + LoggerUtils.toLogString(event) + " from this scene model");
		}
		super.notifyObservers(ObservableAction.THREAD_EVENT_REMOVED, event);

		// also remove constructed objects, if requested
		if (!invocationOnly) {
			if (event instanceof ConstructorExecution) {
				this.removeClassInstance(((ConstructorExecution) event).getConstruction());
			} else if (event instanceof MonitorEvent) {
				final MonitorEvent monitor = (MonitorEvent) event;
				if (monitor.getTarget().getCreator().equals(event)) {
					this.removeMonitorObject(((MonitorEvent) event).getTarget());
				}

				// handle posters
				final AdministrationFloor floor = (AdministrationFloor) monitor.getTarget().getMetaphorObject();
				if (monitor.getType() == MonitorMethod.ACQUIRE) {
					DataModelUtils.removeLock(monitor.getTrace(), monitor.getTarget());
					if (!DataModelUtils.hasLock(monitor.getTrace(), monitor.getTarget())) {
						floor.getPoster().setColor(null);
						this.applyMonitorStatusChanged(floor);
					}
				} else if (monitor.getType() == MonitorMethod.RELEASE) {
					if (!DataModelUtils.hasLock(monitor.getTrace(), monitor.getTarget())) {
						final AdministrationFloor thread = (AdministrationFloor) event.getTrace().getMetaphorObject();
						floor.getPoster().setColor(thread.getPoster().getColor());
						this.applyMonitorStatusChanged(floor);
					}
					monitor.getTrace().getLocks().add(monitor);
				}
			} else if (event instanceof WaitNotifyEvent && ((WaitNotifyEvent) event).getTarget().getCreator().equals(event)) {
				this.removeWaitNotifyObject(((WaitNotifyEvent) event).getTarget());
			} else if (event instanceof ThreadCreation) {
				this.removeThreadObject(((ThreadCreation) event).getTarget());
			}
		}
	}

	/**
	 * Adds a new {@link SThread} instance to this {@link CityModel} instance. This method has no effect, if the given object is already contained in this
	 * scene model.
	 * 
	 * @param thread
	 *            The thread to add to this scene model.
	 */
	public void addThreadObject(final STrace thread) {

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully added new thread object with values: " + LoggerUtils.toLogString(thread) + " to this scene model.");
		}
		super.notifyObservers(ObservableAction.THREAD_INSTANCE_ADDED, thread);
	}

	/**
	 * Adds a new {@link Relation} instance to this {@link CityModel} instance. This method has no effect, if the given object is already contained in this
	 * scene model.
	 * 
	 * @param relation
	 *            The relation to add to this scene model.
	 * @throws IllegalArgumentException
	 *             If the given relation is {@code null}.
	 */
	public void addRelation(final Relation relation) throws IllegalArgumentException {

		if (relation == null) {
			throw new IllegalArgumentException("You cannot add a null relation.");
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully added new relation object with values: " + LoggerUtils.toLogString(relation) + " to this scene model.");
		}
		super.notifyObservers(ObservableAction.RELATION_ADDED, relation);
	}

	/**
	 * Removes the given {@link Relation} instance from this {@link CityModel} instance. This method has no effect, if the given object is not contained
	 * in this model instance.
	 * 
	 * @param relation
	 *            The relation object to remove from this scene model.
	 * @throws IllegalArgumentException
	 *             If the given instance is {@code null}.
	 */
	public void removeRelation(final Relation relation) throws IllegalArgumentException {

		if (relation == null) {
			throw new IllegalArgumentException("You cannot remove a null relation.");
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully removed relation instance with values: " + LoggerUtils.toLogString(relation) + " from this scene model");
		}
		super.notifyObservers(ObservableAction.RELATION_REMOVED, relation);
	}

	/**
	 * Removes the given {@link SClassInstance} instance from this {@link CityModel} instance. This method has no effect, if the given object is not contained
	 * in this model instance.
	 * 
	 * @param instance
	 *            The class object to remove from this scene model.
	 * @throws IllegalArgumentException
	 *             If the given instance is {@code null}.
	 */
	private void removeClassInstance(final SClassInstance instance) throws IllegalArgumentException {

		if (instance == null) {
			throw new IllegalArgumentException("You cannot remove a null instance.");
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully removed class instance with values: " + LoggerUtils.toLogString(instance) + " from this scene model");
		}
		super.notifyObservers(ObservableAction.CLASS_INSTANCE_REMOVED, instance);
	}

	/**
	 * Adds a new {@link MonitorObject} instance to this {@link CityModel} instance. This method has no effect, if the given object is already contained in this
	 * scene model.
	 * 
	 * @param monitor
	 *            The monitor object to add to this scene model.
	 */
	private void addMonitorObject(final MonitorObject monitor) {

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully added new monitor object with values: " + LoggerUtils.toLogString(monitor) + " to this scene model.");
		}
		super.notifyObservers(ObservableAction.MONITOR_OBJECT_ADDED, monitor);
	}

	/**
	 * Removes a {@link MonitorObject} instance from this {@link CityModel} instance. This method has no effect, if the given object is not contained in this
	 * scene model.
	 * 
	 * @param monitor
	 *            The monitor object to remove from this scene model.
	 */
	private void removeMonitorObject(final MonitorObject monitor) {

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully removed monitor object with values: " + LoggerUtils.toLogString(monitor) + " from this scene model.");
		}
		super.notifyObservers(ObservableAction.MONITOR_OBJECT_REMOVED, monitor);
	}

	/**
	 * Applies the monitor status change by notifying all obersers.
	 * 
	 * @param floor
	 *            The {@link AdministrationFloor} representing the monitor whose status has changed.
	 */
	private void applyMonitorStatusChanged(final AdministrationFloor floor) {

		super.notifyObservers(ObservableAction.MONITOR_STATUS_CHANGED, floor);
	}

	/**
	 * Adds a new {@link WaitNotifyObject} instance to this {@link CityModel} instance. This method has no effect, if the given object is already contained in this
	 * scene model.
	 * 
	 * @param waitNotifyObj
	 *            -/notify object to add to this scene model.
	 */
	private void addWaitNotifyObject(final WaitNotifyObject waitNotifyObj) {

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully added new wait-/notify object with values: " + LoggerUtils.toLogString(waitNotifyObj) + " to this scene model.");
		}
		super.notifyObservers(ObservableAction.WAIT_NOTIFY_OBJECT_ADDED, waitNotifyObj);
	}

	/**
	 * Removes a {@link WaitNotifyObject} instance from this {@link CityModel} instance. This method has no effect, if the given object is not contained in this
	 * scene model.
	 * 
	 * @param waitNotifyObj
	 *            The wait-/notify object to remove from this scene model.
	 */
	private void removeWaitNotifyObject(final WaitNotifyObject waitNotifyObj) {

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully removed wait-/notify object with values: " + LoggerUtils.toLogString(waitNotifyObj) + " from this scene model.");
		}
		super.notifyObservers(ObservableAction.MONITOR_OBJECT_REMOVED, waitNotifyObj);
	}

	/**
	 * Removes a new {@link SThread} instance from this {@link CityModel} instance. This method has no effect, if the given object is not contained in this scene
	 * model.
	 * 
	 * @param thread
	 *            The thread to remove from this scene model.
	 */
	private void removeThreadObject(final STrace thread) {

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully removed thread object with values: " + LoggerUtils.toLogString(thread) + " from this scene model.");
		}
		super.notifyObservers(ObservableAction.THREAD_INSTANCE_REMOVED, thread);
	}

}
