/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package models.data;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import data.VisualizableElement;
import data.dynamics.SClassInstance;
import data.dynamics.STrace;
import data.dynamics.STraceEvent;
import data.dynamics.ThreadCreation;
import data.dynamics.ThreadTermination;
import data.dynamics.impl.DynamicsFactoryImpl;
import data.statics.Relation;
import data.statics.RelationType;
import data.statics.SClass;
import data.statics.SMethod;
import data.statics.SPackage;
import data.synchronisation.MonitorEvent;
import data.synchronisation.MonitorObject;
import data.synchronisation.SynchronisationObject;
import data.synchronisation.ThreadJoin;
import data.synchronisation.WaitNotifyEvent;

import utils.comparators.SThreadEventComparator;
import utils.comparators.STraceByThreadIdComparator;

/**
 * A utility class offering general methods to extract, manipulate, or compute certain information from data model related class instances.
 * 
 * @author pdo
 * 
 */
public final class DataModelUtils {

	/** A mapping from {@link VisualizableElement} instances to their metaphor objects used by all removeMapping and restoreMapping methods */
	private static Map<VisualizableElement, Object> map;

	// ========================== Constructor =================================

	/**
	 * Private constructor. This class is not intended for instantiation, since it contains static methods only.
	 */
	private DataModelUtils() {}

	// ============================ Methods ===================================

	/**
	 * Computes the length of the longest package path relative to the given package. More formally, this method will return the highest integer {@code n} with the
	 * following properties:
	 * (1) The {@code n}-value of the given package is {@code 0}.
	 * (2) If a package has a {@code n}-value of {@code m}, then all direct sub-packages have an {@code n}-value of {@code m+1}.
	 * 
	 * @param root
	 *            The common (not necessarily direct) super-package of all packages to consider.
	 * @return The highest nesting level.
	 * @throws IllegalArgumentException
	 *             If the package passed in is {@code null}.
	 */
	public static int maxLevel(final SPackage root) throws IllegalArgumentException {

		if (root == null) {
			throw new IllegalArgumentException("The package passed in is null.");
		}

		int maxLevel = root.getLevel();
		for (SPackage child : root.getSubPackages()) {
			// recursive call for each sub-package
			maxLevel = Math.max(maxLevel, maxLevel(child));
		}
		return maxLevel;
	}

	/**
	 * Returns a list of all outgoing {@link Relation} objects of the specified {@link SClass} instance.
	 * 
	 * @param sClass
	 *            The class to get all its outgoing relations.
	 * @return All outgoing relations of the given class.
	 * @throws IllegalArgumentException
	 *             If the class passed in is {@code null}.
	 */
	public static List<Relation> getOutgoingRelations(final SClass sClass) throws IllegalArgumentException {

		if (sClass == null) {
			throw new IllegalArgumentException("You cannot get outgoing relations from a null class.");
		}

		final List<Relation> outgoing = new LinkedList<Relation>();
		for (Relation relation : sClass.getRelations()) {
			if (relation.getSource().equals(sClass)) {
				outgoing.add(relation);
			}
		}
		return outgoing;
	}

	/**
	 * Returns a list of all incoming {@link Relation} objects of the specified {@link SClass} instance.
	 * 
	 * @param sClass
	 *            The class to get all its incoming relations.
	 * @return All incoming relations of the given class.
	 * @throws IllegalArgumentException
	 *             If the class passed in is {@code null}.
	 */
	public static List<Relation> getIncomingRelations(final SClass sClass) throws IllegalArgumentException {

		if (sClass == null) {
			throw new IllegalArgumentException("You cannot get incoming relations from a null class.");
		}

		final List<Relation> incoming = new LinkedList<Relation>();
		for (Relation relation : sClass.getRelations()) {
			if (relation.getTarget().equals(sClass)) {
				incoming.add(relation);
			}
		}
		return incoming;
	}

	/**
	 * Orders the specified list if {@link STraceEvent} objects by their execution entry timestamp (tin, primary criterion), by their trace Id (secondary), and by
	 * their order index (tertiary). Their execution order attribute will be set according their position in the resulting list.
	 * 
	 * @param events
	 *            The list of events to order.
	 * @throws IllegalArgumentException
	 *             If the list passed in is {@code null}.
	 */
	public static void orderByTin(final List<STraceEvent> events) throws IllegalArgumentException {

		if (events == null) {
			throw new IllegalArgumentException("You cannot order a null list.");
		}

		Collections.sort(events, new SThreadEventComparator());

		int index = 0;
		for (STraceEvent event : events) {
			event.setExecOrder(index++);
		}
	}

	/**
	 * Removes all mappings from the {@link SPackage} and list of {@link STrace} instances passed in, i.e. sets their metaphor object reference to {@code null} after
	 * saving it. May be restored by {@link #restoreMappings(SPackage, List)}.
	 * 
	 * @param pack
	 *            The package to remove all its mappings.
	 * @param threads
	 *            The list of threads to remove all its mappings.
	 * @throws IllegalArgumentException
	 *             If an argument passed in is {@code null}.
	 */
	public static void removeMappings(final SPackage pack, final List<STrace> threads) throws IllegalArgumentException {

		if (pack == null) {
			throw new IllegalArgumentException("You cannot remove the mappings of a null package.");
		}

		if (pack.getSuperPackage() == null) {
			map = new HashMap<VisualizableElement, Object>(500);
		}

		Object metaphor;

		metaphor = pack.getMetaphorObject();
		pack.setMetaphorObject(null);
		map.put(pack, metaphor);
		for (SPackage child : pack.getSubPackages()) {
			removeMappings(child, new LinkedList<STrace>());
		}

		for (SClass child : pack.getSClasses()) {
			metaphor = child.getMetaphorObject();
			child.setMetaphorObject(null);
			map.put(child, metaphor);
			for (SClassInstance instance : child.getInstances()) {
				metaphor = instance.getMetaphorObject();
				instance.setMetaphorObject(null);
				map.put(instance, metaphor);
			}
			for (Relation relation : child.getRelations()) {
				metaphor = relation.getMetaphorObject();
				relation.setMetaphorObject(null);
				map.put(relation, metaphor);
			}
			for (SMethod method : child.getSMethods()) {
				metaphor = method.getMetaphorObject();
				method.setMetaphorObject(null);
				map.put(method, metaphor);
			}
		}

		removeMappings(threads);
	}

	/**
	 * Removes all mappings from the list of {@link SThread} instances passed in, i.e. sets their metaphor object reference to {@code null} after saving it. May be
	 * restored by {@link #restoreMappings(List)}.
	 * 
	 * @param threads
	 *            The list of threads to remove all its mappings.
	 * @throws IllegalArgumentException
	 *             If the list of threads passed in is {@code null}.
	 */
	private static void removeMappings(final List<STrace> threads) throws IllegalArgumentException {

		if (threads == null) {
			throw new IllegalArgumentException("You cannot remove the mappings of null threads.");
		}

		Object metaphor;

		for (STrace thread : threads) {
			metaphor = thread.getMetaphorObject();
			thread.setMetaphorObject(null);
			map.put(thread, metaphor);
			for (STraceEvent event : thread.getEvents()) {
				metaphor = event.getMetaphorObject();
				event.setMetaphorObject(null);
				map.put(event, metaphor);
				if (event instanceof MonitorEvent) {
					final SynchronisationObject sync = ((MonitorEvent) event).getTarget();
					metaphor = sync.getMetaphorObject();
					if (metaphor != null) {
						map.put(sync, metaphor);
						System.out.println("restore");
						sync.setMetaphorObject(null);
					}
				} else if (event instanceof WaitNotifyEvent) {
					final SynchronisationObject sync = ((WaitNotifyEvent) event).getTarget();
					metaphor = sync.getMetaphorObject();
					if (metaphor != null) {
						map.put(sync, sync.getMetaphorObject());
						sync.setMetaphorObject(null);
					}
				}
			}
		}
	}

	/**
	 * Restores the mappings previously removed by {@link #removeMappings(SPackage, List)}, i.e. sets their metaphor objects to their original values.
	 * 
	 * @param pack
	 *            The package to restore its metaphor object reference.
	 * @param threads
	 *            The list of threads to restore its members metaphor object references.
	 * @throws IllegalArgumentException
	 *             If an argument passed in is {@code null}.
	 */
	public static void restoreMappings(final SPackage pack, final List<STrace> threads) throws IllegalArgumentException {

		if (pack == null) {
			throw new IllegalArgumentException("You cannot restore the mappings of a null package.");
		}

		pack.setMetaphorObject(map.get(pack));
		for (SPackage child : pack.getSubPackages()) {
			restoreMappings(child, new LinkedList<STrace>());
		}

		for (SClass child : pack.getSClasses()) {
			child.setMetaphorObject(map.get(child));
			for (SClassInstance instance : child.getInstances()) {
				instance.setMetaphorObject(map.get(instance));
			}
			for (Relation relation : child.getRelations()) {
				relation.setMetaphorObject(map.get(relation));
			}
			for (SMethod method : child.getSMethods()) {
				method.setMetaphorObject(map.get(method));
			}
		}

		restoreMappings(threads);

	}

	/**
	 * Restores the mappings previously removed by {@link #removeMappings(List)}, i.e. sets their metaphor objects to their original values.
	 * 
	 * @param threads
	 *            The list of threads to restore its members metaphor object references.
	 * @throws IllegalArgumentException
	 *             If an the list of threads passed in is {@code null}.
	 */
	private static void restoreMappings(final List<STrace> threads) throws IllegalArgumentException {

		if (threads == null) {
			throw new IllegalArgumentException("You cannot restore the mappings of null threads.");
		}

		Object metaphor;
		for (STrace thread : threads) {
			thread.setMetaphorObject(map.get(thread));
			for (STraceEvent event : thread.getEvents()) {
				event.setMetaphorObject(map.get(event));
				if (event instanceof MonitorEvent) {
					final SynchronisationObject sync = ((MonitorEvent) event).getTarget();
					metaphor = map.get(sync);
					if (metaphor != null) {
						sync.setMetaphorObject(metaphor);
					}
				} else if (event instanceof WaitNotifyEvent) {
					final SynchronisationObject sync = ((WaitNotifyEvent) event).getTarget();
					metaphor = map.get(sync);
					if (metaphor != null) {
						sync.setMetaphorObject(metaphor);
					}
				}
			}
		}
	}

	/**
	 * Determines whether the given {@link STrace} currently holds the lock on the specified {@link MonitorObject} instance (i.e. has it acquired).
	 * 
	 * @param trace
	 *            The trace to determine whether it holds the lock.
	 * @param monitor
	 *            The monitor to test.
	 * @return {@code True}, if the specified trace currently holds the lock, {@code false} otherwise.
	 * @throws IllegalArgumentException
	 *             If either the trace or monitor instance passed in is {@code null}.
	 */
	public static boolean hasLock(final STrace trace, final MonitorObject monitor) throws IllegalArgumentException {

		if (trace == null) {
			throw new IllegalArgumentException("A null trace cannot hold monitor locks.");
		}
		if (monitor == null) {
			throw new IllegalArgumentException("A trace cannot hold a null lock.");
		}

		for (MonitorEvent event : trace.getLocks()) {
			if (event.getTarget().equals(monitor)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Removes the lock on the specified {@link MonitorObject} instance from the given trace.
	 * 
	 * @param trace
	 *            The trace to remove its lock on the specified monitor object.
	 * @param monitor
	 *            The monitor to remove the lock.
	 * @throws IllegalArgumentException
	 *             If either the trace or monitor instance passed in is {@code null}.
	 */
	public static void removeLock(final STrace trace, final MonitorObject monitor) throws IllegalArgumentException {

		if (trace == null) {
			throw new IllegalArgumentException("A null trace cannot hold monitor locks.");
		}
		if (monitor == null) {
			throw new IllegalArgumentException("A null monitor does not offer locks.");
		}

		MonitorEvent toRemove = null;
		for (MonitorEvent event : trace.getLocks()) {
			if (event.getTarget().equals(monitor)) {
				toRemove = event;
			}
		}
		if (toRemove != null) {
			trace.getLocks().remove(toRemove);
		}
	}

	/**
	 * Adds all methods (except constructors) of the inherited classes to the classes in the given package and all its subpackages.
	 * 
	 * @param pack
	 *            The package in which all classes are contained.
	 * @throws IllegalArgumentException
	 *             If the package passed in is {@code null}.
	 */
	public static void addInheritedMethods(final SPackage pack) throws IllegalArgumentException {

		if (pack == null) {
			throw new IllegalArgumentException("This method does not accept null packages.");
		}

		// recursive call for each child
		for (SPackage child : pack.getSubPackages()) {
			addInheritedMethods(child);
		}

		for (SClass clazz : pack.getSClasses()) {
			addInheritedmethods(clazz);
		}
	}

	/**
	 * Adds all inherited methods to the given {@link SClass} instance.
	 * 
	 * @param clazz
	 *            The class to add all its inherited methods to its list of methods.
	 */
	private static void addInheritedmethods(final SClass clazz) {

		for (Relation relation : clazz.getRelations()) {
			if (relation.getType() == RelationType.EXTENSION && relation.getSource() == clazz) {
				addInheritedmethods(relation.getTarget());
				for (SMethod method : relation.getTarget().getSMethods()) {
					if (!(method.getName().equals(relation.getTarget().getName()) || clazz.getSMethods().contains(method))) {
						// method is no constructor
						clazz.getSMethods().add(method);
					}
				}
			}
		}
	}

	/**
	 * Combines the given {@link STrace} instances by the {@code threadId} attribute. For each unique attribute value a new trace object will be created and all
	 * events of traces with the same thread id will be added to it in descending chronological order.
	 * 
	 * @param traces
	 *            The list of traces to combine
	 * @return The list of new traces.
	 * @throws IllegalArgumentException
	 *             If the list of traces passed in is {@code null}.
	 */
	public static List<STrace> combineByThreadId(final List<STrace> traces) throws IllegalArgumentException {

		if (traces == null) {
			throw new IllegalArgumentException("You cannot combine null traces.");
		}

		// maps the original trace objects to the new combined ones
		final Map<STrace, STrace> oldToNew = new HashMap<STrace, STrace>(100);
		final List<STrace> newTraces = new LinkedList<STrace>();

		// sort traces
		Collections.sort(traces, new STraceByThreadIdComparator());

		Long currentThreadId = null;
		STrace currentTrace = null;
		for (STrace trace : traces) {
			if (currentThreadId == null || currentThreadId != trace.getThreadId()) {

				// create new trace instance: note that the orderIndex is not not unique or ordering any more
				currentTrace = DynamicsFactoryImpl.eINSTANCE.createSTrace();
				currentTrace.setClassSignature(trace.getClassSignature());
				currentTrace.setHostname(trace.getHostname());
				currentTrace.setOperationSignature(trace.getOperationSignature());
				currentTrace.setParentOrderIndex(trace.getParentOrderIndex());
				currentTrace.setSessionId(trace.getSessionId());
				currentTrace.setThreadId(trace.getThreadId());
				currentTrace.setTin(trace.getTin());

				currentThreadId = trace.getThreadId();
			}

			currentTrace.setTout(trace.getTout());
			oldToNew.put(trace, currentTrace);

			// remove termination event
			trace.getEvents().remove(trace.getEvents().size() - 1);

			// add all trace elements
			currentTrace.getEvents().addAll(trace.getEvents());

			if (!newTraces.contains(currentTrace)) {
				newTraces.add(currentTrace);
			}
		}

		// add termination event to each new trace
		for (STrace newTrace : newTraces) {
			final ThreadTermination termination = DynamicsFactoryImpl.eINSTANCE.createThreadTermination();
			termination.setEss(1);
			termination.setOrderIndex(Integer.MAX_VALUE);
			termination.setTin(newTrace.getTout());
			termination.setTrace(newTrace);
			newTrace.getEvents().add(newTrace.getEvents().size(), termination);
		}

		// update references
		for (STrace trace : traces) {
			final STrace newTrace = oldToNew.get(trace);
			for (STraceEvent event : trace.getEvents()) {
				event.setTrace(newTrace);
				if (event instanceof ThreadCreation) {
					final ThreadCreation creation = (ThreadCreation) event;
					creation.setTarget(oldToNew.get(creation.getTarget()));
				} else if (event instanceof ThreadJoin) {
					final ThreadJoin join = (ThreadJoin) event;
					join.setTarget(oldToNew.get(join.getTarget()));
				}
			}
		}

		return newTraces;
	}

}
