/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package models.metaphor;

import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

import data.VisualizableElement;
import data.dynamics.ConstructorExecution;
import data.dynamics.SClassInstance;
import data.dynamics.SMethodExecution;
import data.dynamics.STrace;
import data.dynamics.STraceEvent;
import data.dynamics.ThreadCreation;
import data.dynamics.ThreadTermination;
import data.statics.Relation;
import data.statics.SClass;
import data.statics.SPackage;
import data.synchronisation.MonitorEvent;
import data.synchronisation.MonitorMethod;
import data.synchronisation.MonitorObject;
import data.synchronisation.ThreadJoin;
import data.synchronisation.WaitNotifyEvent;
import data.synchronisation.WaitNotifyMethod;

import metaphor.entities.AdministrationBase;
import metaphor.entities.AdministrationBuilding;
import metaphor.entities.AdministrationDistrict;
import metaphor.entities.AdministrationFloor;
import metaphor.entities.Cable;
import metaphor.entities.Floor;
import metaphor.entities.RegularBase;
import metaphor.entities.RegularBuilding;
import metaphor.entities.RegularDistrict;
import metaphor.entities.RegularFloor;
import metaphor.entities.Status;
import metaphor.entities.Street;
import metaphor.entities.impl.EntitiesFactoryImpl;
import metaphor.entities.impl.EntitiesPackageImpl;
import metaphor.types.VisualEntity;
import models.metaphor.applier.ColorApplier;
import models.metaphor.applier.DimensionApplier;
import models.metaphor.applier.PositionApplier;

import observer.AbstractObservable;
import observer.IObserver;
import observer.ObservableAction;
import utils.LoggerUtils;
import visual.metaphor.SymbolSign.Symbol;

/**
 * A metaphor importer to import all visualizable data objects in a given {@link CityModel} instance to a new {@link MetaphorModel} object.
 * 
 * @author pdo
 * 
 */
public final class MetaphorImporter implements IObserver {

	// ============================= Attributes ===============================

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(MetaphorModel.class);

	/**
	 * The {@link AdministrationBase} instance all synchronization objects (i.e. instances of {@link SThread}, {@link MonitorObject}, and {@link WaitNotifyObject})
	 * are commonly placed on in the final visualization.
	 */
	private AdministrationBase adminBase;

	/** A list of all process representing {@link AdministrationFloor} instances already imported. */
	private final List<AdministrationFloor> processes = new LinkedList<AdministrationFloor>();

	/** The {@link MetaphorModel} instance to add all imported data to. */
	private final MetaphorModel metapher;

	/** A {@link DimensionApplier} to automatically set the dimensions of the imported metaphor objects. */
	private final DimensionApplier measuresApplier = new DimensionApplier();

	/** A {@link ColorApplier} to automatically set the color of the imported metaphor objects. */
	private final ColorApplier colorApplier = new ColorApplier();

	/** A {@link PositionApplier} to automatically set the position of the imported metaphor objects. */
	private final PositionApplier positionApplier = new PositionApplier();

	// ============================= Constructor ==============================

	/**
	 * Constructs a new {@link MetaphorImporter} instance with the given {@link MetaphorModel} object to import all data into.
	 * 
	 * @param metapherModel
	 *            The metaphor instance to add all imported data to.
	 * @throws IllegalArgumentException
	 *             If the given metaphor model is {@code null}.
	 */
	public MetaphorImporter(final MetaphorModel metapherModel) throws IllegalArgumentException {

		if (metapherModel == null) {
			throw new IllegalArgumentException("You cannot pass in a null metaphor model.");
		}

		this.metapher = metapherModel;

		EntitiesPackageImpl.init();
		EntitiesFactoryImpl.init();
	}

	// =========================== Interface Methods ==========================

	@Override
	public void update(final AbstractObservable observable, final ObservableAction action, final Object object) {

		switch (action) {
		case STATICS_LOADED:
			this.onStaticsLoaded((SPackage) object);
			break;
		case ALL_CLEARED:
			this.onAllCleared();
			break;
		case CLASS_INSTANCE_ADDED:
			this.onClassInstanceAdded((SClassInstance) object);
			break;
		case CLASS_INSTANCE_REMOVED:
			this.onClassInstanceRemoved((SClassInstance) object);
			break;
		case THREAD_EVENT_ADDED:
			this.onThreadEventAdded((STraceEvent) object);
			break;
		case THREAD_EVENT_REMOVED:
			this.onThreadEventRemoved((STraceEvent) object);
			break;
		case RELATION_ADDED:
			this.onRelationInstanceAdded((Relation) object);
			break;
		case RELATION_REMOVED:
			this.onRelationInstanceRemoved((Relation) object);
			break;
		case MONITOR_OBJECT_ADDED:
		case WAIT_NOTIFY_OBJECT_ADDED:
		case THREAD_INSTANCE_ADDED:
			this.onSynchronizationObjectAdded((VisualizableElement) object);
			break;
		case MONITOR_OBJECT_REMOVED:
		case WAIT_NOTIFY_OBJECT_REMOVED:
		case THREAD_INSTANCE_REMOVED:
			this.onSynchronizationObjectRemoved((VisualizableElement) object);
			break;
		case MONITOR_STATUS_CHANGED:
			this.metapher.onMonitorStatusChanged((AdministrationFloor) object);
			break;
		default:
			LOGGER.warn("ObservableAction '" + action + "' not supported.");
		}
	}

	// =============================== Methods ================================

	/**
	 * The method being automatically invoked when the {@link #update(AbstractObservable, ObservableAction, Object)} method has been called with the 'action'
	 * parameter set as {@link ObservableAction#STATICS_LOADED}. It imports the given {@link SPackage} by creating a representing {@link RegularBase} instance. All
	 * contained sub-packages and {@link SClass} objects will be added to this base using {@link RegularDistrict} and {@link RegularBuilding} objects.
	 * A (mostly) hard-coded {@link AdministrationBase} with its {@link AdministrationDistrict} and three {@link AdministrationBuilding} objects will be created
	 * and added to this metaphor model thereafter.
	 * 
	 * @param root
	 *            The root package to be imported.
	 */
	private void onStaticsLoaded(final SPackage root) {

		// create new regular base instance and build package structure
		final RegularBase regBase = this.buildStaticStructure(root);

		// set attributes
		this.colorApplier.applyColor(regBase);
		this.measuresApplier.applyDimensions(regBase);
		this.positionApplier.applyPosition(regBase);

		// import static class instances
		this.buildGroundFloors(root);

		// create pre-defined administration base and district
		final AdministrationBase adBase = EntitiesFactoryImpl.eINSTANCE.createAdministrationBase();
		final AdministrationDistrict district = EntitiesFactoryImpl.eINSTANCE.createAdministrationDistrict();
		adBase.setDistrict(district);
		this.adminBase = adBase;

		// create and add administration buildings
		final AdministrationBuilding threadBuilding = EntitiesFactoryImpl.eINSTANCE.createAdministrationBuilding();
		final AdministrationBuilding monitorBuilding = EntitiesFactoryImpl.eINSTANCE.createAdministrationBuilding();
		final AdministrationBuilding waitNotifyBuilding = EntitiesFactoryImpl.eINSTANCE.createAdministrationBuilding();

		district.getBuildings().add(Symbol.THREAD.ordinal(), threadBuilding);
		district.getBuildings().add(Symbol.WAIT_NOTIFY.ordinal(), waitNotifyBuilding);
		district.getBuildings().add(Symbol.MONITOR.ordinal(), monitorBuilding);

		threadBuilding.setDistrict(district);
		monitorBuilding.setDistrict(district);
		waitNotifyBuilding.setDistrict(district);

		// set attributes
		this.colorApplier.applyColor(adBase, regBase);
		this.measuresApplier.applyDimensions(adBase);
		this.positionApplier.applyPosition(adBase, regBase);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully created and imported administration base with values: " + LoggerUtils.toLogString(adBase) + ".");
			LOGGER.debug("Successfully created and imported administration district with values: " + LoggerUtils.toLogString(district) + ".");
			LOGGER.debug("Successfully created and imported thread building with values: " + LoggerUtils.toLogString(threadBuilding) + ".");
			LOGGER.debug("Successfully created and imported monitor building with values: " + LoggerUtils.toLogString(threadBuilding) + ".");
			LOGGER.debug("Successfully created and imported wait-/notify building with values: " + LoggerUtils.toLogString(waitNotifyBuilding) + ".");
		}
		this.metapher.setStaticsLoaded((RegularBase) root.getMetaphorObject());
		this.metapher.addAdministrationBase(this.adminBase);
	}

	/**
	 * Creates a new {@link RegularFloor} instance for each {@link RegularBuilding} object contained in the specified package and adds them to the visualization.
	 * These floors are the ground floors of each building, i.e. represent the class object of each class itself.
	 * 
	 * @param pack
	 *            The package containing all buildings for which a ground floor is to be created.
	 */
	private void buildGroundFloors(final SPackage pack) {

		for (SPackage child : pack.getSubPackages()) {
			this.buildGroundFloors(child);
		}

		for (SClass clazz : pack.getSClasses()) {
			this.buildGroundFloor(clazz.getInstances().get(0));
		}
	}

	/**
	 * Creates a new {@link RegularFloor} instance representing the specified {@link SClassObject} and adds it as a floor to the building representing its class.
	 * 
	 * @param sClassObject
	 *            The class object to create its representing (ground) floor.
	 */
	private void buildGroundFloor(final SClassInstance sClassObject) {

		// create new floor instance
		final RegularFloor floor = EntitiesFactoryImpl.eINSTANCE.createRegularFloor();
		this.map(floor, sClassObject);
		final RegularBuilding building = (RegularBuilding) sClassObject.getSClass().getMetaphorObject();
		floor.setBuilding(building);

		// set attributes
		this.measuresApplier.applyDimensions(floor);
		this.positionApplier.applyPosition(floor);
		this.colorApplier.applyColor(floor);

		building.getFloors().add(building.getFloors().size(), floor);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully created and imported new regular floor with values: " + LoggerUtils.toLogString(floor) + ".");
		}
	}

	/**
	 * Builds the static city structure. It creates a new base for the given {@link SPackage} instance and buildings and districts for all its contained members.
	 * 
	 * @param root
	 *            The root package of the application to visualize.
	 * @return The {@link RegularBase} instance representing the given root package.
	 */
	private RegularBase buildStaticStructure(final SPackage root) {

		// create new regular base instance
		final RegularBase base = EntitiesFactoryImpl.eINSTANCE.createRegularBase();
		this.map(base, root);
		this.measuresApplier.applyDimensions(base);

		// build package structure
		for (SPackage pack : root.getSubPackages()) {
			final RegularDistrict subDistrict = this.buildPackageStructure(pack);
			base.getDistricts().add(subDistrict);
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully created and imported regular base with values: " + LoggerUtils.toLogString(base) + ".");
		}
		return (RegularBase) root.getMetaphorObject();
	}

	/**
	 * Builds the static package structure. It creates a new {@link RegularDistrict} for the given {@link SPackage} instance and buildings for all its contained
	 * {@link SClass} instance members.
	 * 
	 * @param pack
	 *            The package to be imported.
	 * @return The {@link RegularDistrict} instance representing the given package.
	 */
	private RegularDistrict buildPackageStructure(final SPackage pack) {

		// create new regular district instance
		final RegularDistrict district = EntitiesFactoryImpl.eINSTANCE.createRegularDistrict();
		this.map(district, pack);
		this.measuresApplier.applyDimensions(district);

		// build class structure
		for (SClass sClass : pack.getSClasses()) {
			final RegularBuilding building = this.buildBuilding(sClass);
			district.getBuildungs().add(building);
			building.setDistrict(district);
		}

		// recursive call for each sub-package
		for (SPackage subPackage : pack.getSubPackages()) {
			final RegularDistrict subDistrict = this.buildPackageStructure(subPackage);
			district.getChildDistricts().add(subDistrict);
			subDistrict.setParentDistrict(district);
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully imported new regular district with values: " + LoggerUtils.toLogString(district) + ".");
		}
		return (RegularDistrict) pack.getMetaphorObject();
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#ALL_CLEARED} has been received. It removes all static and dynamics
	 * elements from the metaphor model. It will be totally empty when this method returns.
	 */
	private void onAllCleared() {

		this.processes.clear();
		this.colorApplier.updateAttributes();
		this.measuresApplier.updateAttributes();
		this.positionApplier.updateAttributes();

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully cleared both statics and dynamics.");
		}

		this.metapher.clearAll();
	}

	/**
	 * Creates a new {@link RegularBuilding} instance representing the specified {@link SClass} object.
	 * 
	 * @param sClass
	 *            The class object to create its corresponding regular building.
	 * @return The regular building instance representing the specified sClass.
	 */
	private RegularBuilding buildBuilding(final SClass sClass) {

		// create new regular building instance
		final RegularBuilding building = EntitiesFactoryImpl.eINSTANCE.createRegularBuilding();
		this.map(building, sClass);

		final RegularDistrict district = (RegularDistrict) sClass.getSPackage().getMetaphorObject();
		district.getBuildungs().add(district.getBuildungs().size(), building);
		building.setDistrict(district);

		this.measuresApplier.applyDimensions(building);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully created and imported new regular building with values: " + LoggerUtils.toLogString(building) + ".");
		}
		return (RegularBuilding) sClass.getMetaphorObject();
	}

	/**
	 * The action to be performed when a change notification has been received with {@link ObservableAction#CLASS_INSTANCE_ADDED}. It creates a new
	 * {@link RegularFloor} instance representing the specified {@link SClassInstance} object (if not already existing) and adds it to the metaphor model.
	 * 
	 * @param instance
	 *            The class instance to be imported into the metaphor model.
	 */
	private void onClassInstanceAdded(final SClassInstance instance) {

		// create new floor instance
		final RegularFloor floor = EntitiesFactoryImpl.eINSTANCE.createRegularFloor();
		this.map(floor, instance);

		final RegularBuilding building = (RegularBuilding) instance.getSClass().getMetaphorObject();
		floor.setBuilding(building);

		// set attributes
		this.colorApplier.applyColor(floor);
		this.measuresApplier.applyDimensions(floor);
		this.positionApplier.applyPosition(floor);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully created and imported new regular floor with values: " + LoggerUtils.toLogString(floor) + ".");
		}

		// add as floor
		final List<RegularFloor> floors = floor.getBuilding().getFloors();
		floors.add(floors.size(), floor);

		this.metapher.addFloor((RegularFloor) instance.getMetaphorObject());
	}

	/**
	 * The action to be performed when a change notification has been received with {@link ObservableAction#MONITOR_OBJECT_REMOVED} or
	 * {@link ObservableAction#WAIT_NOTIFY_OBJECT_REMOVED}. It removes the {@link SClassInstance}'s metaphor object from the metaphor model.
	 * 
	 * @param instance
	 *            The class instance element to remove its metaphor object from the metaphor model.
	 */
	private void onClassInstanceRemoved(final SClassInstance instance) {

		// remove from building's floor list
		final RegularFloor floor = (RegularFloor) instance.getMetaphorObject();
		final List<RegularFloor> floors = floor.getBuilding().getFloors();
		floors.remove(floor);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully removed regular floor with values: " + LoggerUtils.toLogString(floor) + ".");
		}
		this.metapher.removeFloor(floor);
	}

	/**
	 * The action to be performed when a change notification has been received with {@link ObservableAction#THREAD_EVENT_ADDED}. It creates a new {@link Street}
	 * instance representing the specified {@link STraceEvent} object (if not already existing) and adds it to the metaphor model.
	 * 
	 * @param event
	 *            The thread event to be imported to the metaphor model.
	 */
	private void onThreadEventAdded(final STraceEvent event) {

		if (event instanceof ThreadTermination) {
			this.metapher.addThreadTermination((AdministrationFloor) event.getTrace().getMetaphorObject());
		} else {
			// create new street instance
			final Street street = EntitiesFactoryImpl.eINSTANCE.createStreet();
			street.setStatus(Status.OPENED);
			this.map(street, event);

			// set source floor
			final Floor source;
			if (event.getSource() == null) {
				// event originates directly from run() or main(String[] args) method
				source = (Floor) event.getTrace().getMetaphorObject();
			} else {
				source = (Floor) event.getSource().getMetaphorObject();
			}

			// set target floor
			final Floor target;
			if (event instanceof ConstructorExecution) {
				target = (Floor) ((ConstructorExecution) event).getConstruction().getSClass().getInstances().get(0).getMetaphorObject();
			} else if (event instanceof SMethodExecution) {
				target = (Floor) ((SMethodExecution) event).getTarget().getMetaphorObject();
			} else if (event instanceof ThreadCreation) {
				target = this.processes.get(0);
			} else if (event instanceof ThreadJoin) {
				target = (Floor) ((ThreadJoin) event).getTarget().getMetaphorObject();
				street.setStatus(Status.CLOSED);
			} else if (event instanceof MonitorEvent) {
				final MonitorEvent monitorEvent = (MonitorEvent) event;
				target = (AdministrationFloor) monitorEvent.getTarget().getMetaphorObject();
				if (monitorEvent.getType() == MonitorMethod.REQUEST) {
					street.setStatus(Status.CLOSED);
				}
			} else {
				final WaitNotifyEvent waitNotifyEvent = (WaitNotifyEvent) event;
				target = (Floor) waitNotifyEvent.getTarget().getMetaphorObject();
				if (waitNotifyEvent.getType() == WaitNotifyMethod.WAIT) {
					street.setStatus(Status.CLOSED);
				}
			}

			street.setSource(source);
			street.setTarget(target);

			// set attributes
			this.colorApplier.applyColor(street);
			this.positionApplier.applyPosition(street);

			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Successfully created and imported new street with values: " + LoggerUtils.toLogString(street) + ".");
			}

			street.getSource().getStreets().add(street);
			this.metapher.addStreet(street);
		}
	}

	/**
	 * The action to be performed when a change notification has been received with {@link ObservableAction#MONITOR_OBJECT_REMOVED} or
	 * {@link ObservableAction#WAIT_NOTIFY_OBJECT_REMOVED}. It removes the {@link SThreadEvent}'s metaphor object from the metaphor model.
	 * 
	 * @param event
	 *            The thread event element to remove its metaphor object from the metaphor model.
	 */
	private void onThreadEventRemoved(final STraceEvent event) {

		if (event instanceof ThreadTermination) {
			this.metapher.removeThreadTermination((AdministrationFloor) event.getTrace().getMetaphorObject());
		} else {
			final Street street = (Street) event.getMetaphorObject();
			street.getSource().getStreets().remove(street);

			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Successfully removed street with values: " + LoggerUtils.toLogString(street) + ".");
			}
			this.metapher.removeStreet(street);
		}
	}

	/**
	 * The action to be performed when a change notification has been received with {@link ObservableAction#RELATION_ADDED}. It creates a new {@link Cable} instance
	 * representing the specified {@link Relation} object (if not already existing) and adds it to the metaphor model.
	 * 
	 * @param relation
	 *            The relation instance to be imported into the metaphor model.
	 */
	private void onRelationInstanceAdded(final Relation relation) {

		// create new cable instance
		final Cable cable = EntitiesFactoryImpl.eINSTANCE.createCable();
		this.map(cable, relation);

		// set source and target
		cable.setSource((RegularBuilding) relation.getSource().getMetaphorObject());
		cable.setTarget((RegularBuilding) relation.getTarget().getMetaphorObject());
		relation.getSource().getRelations().add(relation);
		if (!relation.getSource().equals(relation.getTarget())) {
			relation.getTarget().getRelations().add(relation);
		}

		// do not apply position since every cable will be translated to a bendable arrow anyways
		this.colorApplier.applyColor(cable);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully created and imported new cable with values: " + LoggerUtils.toLogString(cable) + ".");
		}
		this.metapher.addCable((Cable) relation.getMetaphorObject());
	}

	/**
	 * The action to be performed when a change notification has been received with {@link ObservableAction#MONITOR_OBJECT_REMOVED} or
	 * {@link ObservableAction#WAIT_NOTIFY_OBJECT_REMOVED}. It removes the {@link Relation}'s metaphor object from the metaphor model.
	 * 
	 * @param relation
	 *            The relation element to remove its metaphor object from the metaphor model.
	 */
	private void onRelationInstanceRemoved(final Relation relation) {

		final Cable cable = (Cable) relation.getMetaphorObject();
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully removed cable with values: " + LoggerUtils.toLogString(cable) + ".");
		}
		this.metapher.removeCable(cable);
	}

	/**
	 * The action to be performed when a change notification has been received with {@link ObservableAction#WAIT_NOTIFY_OBJECT_ADDED} or
	 * {@link ObservableAction#MONITOR_OBJECT_ADDED}. It creates a new {@link AdministrationFloor} instance representing the specified {@link MonitorObject} or
	 * {@link WaitNotifyObject} object (if not already existing) and adds it to the metaphor model.
	 * 
	 * @param object
	 *            The monitor or wait-/notify instance to be imported into the metaphor model.
	 */
	private void onSynchronizationObjectAdded(final VisualizableElement object) {

		// create new floor instance
		final AdministrationFloor floor = EntitiesFactoryImpl.eINSTANCE.createAdministrationFloor();
		floor.setPoster(EntitiesFactoryImpl.eINSTANCE.createPoster());
		this.map(floor, object);

		// add object to corresponding administration building
		final List<AdministrationBuilding> buildings = this.adminBase.getDistrict().getBuildings();
		final AdministrationBuilding building;

		if (object instanceof STrace) {
			final STrace trace = (STrace) object;
			floor.getPoster().setColor(this.colorApplier.getThreadColor(trace.getExecOrder()));
			building = buildings.get(Symbol.THREAD.ordinal());
			if (((STrace) object).getParentOrderIndex() < 0) {
				this.processes.add(floor);
			}
		} else if (object instanceof MonitorObject) {
			building = buildings.get(Symbol.MONITOR.ordinal());
		} else {
			building = buildings.get(Symbol.WAIT_NOTIFY.ordinal());
		}

		floor.setBuilding(building);
		building.getFloors().add(building.getFloors().size(), floor);

		// set attributes
		this.colorApplier.applyColor(floor);
		this.measuresApplier.applyDimensions(floor);
		this.positionApplier.applyPosition(floor);

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Successfully created and imported new administration floor with values: " + LoggerUtils.toLogString(floor) + ".");
		}
		this.metapher.addAdministrationFloor((AdministrationFloor) object.getMetaphorObject());
	}

	/**
	 * The action to be performed when a change notification has been received with {@link ObservableAction#MONITOR_OBJECT_REMOVED} or
	 * {@link ObservableAction#WAIT_NOTIFY_OBJECT_REMOVED}. It removes the {@link VisualizableElement}'s metaphor object from the metaphor model.
	 * 
	 * @param object
	 *            The visualizable element to remove its metaphor object from the metaphor model.
	 */
	private void onSynchronizationObjectRemoved(final VisualizableElement object) {

		if (object != null) {
			final AdministrationFloor floor = (AdministrationFloor) object.getMetaphorObject();
			floor.getBuilding().getFloors().remove(floor);
			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Successfully removed administration floor with values: " + LoggerUtils.toLogString(floor) + ".");
			}
			this.metapher.removeAdministrationFloor(floor);
		}
	}

	/**
	 * Maps the given objects to each other. It sets the given {@code entity} as the {@code elements} metaphor object and the {@code element} as {@code entity}'s
	 * data object.
	 * 
	 * @param entity
	 *            The visual entity to map to the given {@link VisualizableElement} object.
	 * @param element
	 *            The visualizable element to map to the given {@link VisualEntity} instance.
	 */
	private void map(final VisualEntity entity, final VisualizableElement element) {

		entity.setDataObject(element);
		element.setMetaphorObject(entity);
	}

}
