/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package models.metaphor;

import metaphor.entities.AdministrationBase;
import metaphor.entities.AdministrationFloor;
import metaphor.entities.Cable;
import metaphor.entities.RegularBase;
import metaphor.entities.RegularFloor;
import metaphor.entities.Street;
import observer.AbstractObservable;
import observer.ObservableAction;

/**
 * The metaphor model of the Synchrovis application containing all metaphor elements currently being displayed on screen.
 * 
 * @author pdo
 * 
 */
public final class MetaphorModel extends AbstractObservable {

	// ========================== Constructor =================================

	/**
	 * Creates a new {@link MetaphorModel} instance.
	 */
	public MetaphorModel() {

		super();
	}

	// ============================ Methods ===================================

	/**
	 * Adds a new {@link RegularBase} to this metaphor model.
	 * 
	 * @param base
	 *            The regular base to add.
	 */
	public void setStaticsLoaded(final RegularBase base) {

		super.notifyObservers(ObservableAction.STATICS_LOADED, base);
	}

	/**
	 * Adds a new {@link AdministrationBase} to this metaphor model.
	 * 
	 * @param base
	 *            The administration base to add.
	 */
	public void addAdministrationBase(final AdministrationBase base) {

		super.notifyObservers(ObservableAction.DYNAMICS_LOADED, base);
	}

	/**
	 * Clears all elements in this metaphor model.
	 */
	public void clearAll() {

		super.notifyObservers(ObservableAction.ALL_CLEARED, null);
	}

	/**
	 * Adds a new {@link RegularFloor} to this metaphor model.
	 * 
	 * @param floor
	 *            The regular floor to add.
	 */
	public void addFloor(final RegularFloor floor) {

		super.notifyObservers(ObservableAction.FLOOR_ADDED, floor);
	}

	/**
	 * Removes a {@link RegularFloor} from this metaphor model.
	 * 
	 * @param floor
	 *            The regular floor to remove.
	 */
	public void removeFloor(final RegularFloor floor) {

		super.notifyObservers(ObservableAction.FLOOR_REMOVED, floor);
	}

	/**
	 * Adds a new {@link Cable} to this metaphor model.
	 * 
	 * @param cable
	 *            The cable to add.
	 */
	public void addCable(final Cable cable) {

		super.notifyObservers(ObservableAction.CABLE_ADDED, cable);
	}

	/**
	 * Removes a {@link Cable} from this metaphor model.
	 * 
	 * @param cable
	 *            The cable to remove.
	 */
	public void removeCable(final Cable cable) {

		super.notifyObservers(ObservableAction.CABLE_REMOVED, cable);
	}

	/**
	 * Adds a new {@link Street} to this metaphor model.
	 * 
	 * @param street
	 *            The street to add.
	 */
	public void addStreet(final Street street) {

		super.notifyObservers(ObservableAction.STREET_ADDED, street);
	}

	/**
	 * Removes a {@link Street} from this metaphor model.
	 * 
	 * @param street
	 *            The street to remove.
	 */
	public void removeStreet(final Street street) {

		super.notifyObservers(ObservableAction.STREET_REMOVED, street);
	}

	/**
	 * Adds a new {@link AdministrationFloor} to this metaphor model.
	 * 
	 * @param floor
	 *            The administration floor to add.
	 */
	public void addAdministrationFloor(final AdministrationFloor floor) {

		super.notifyObservers(ObservableAction.ADMINISTRATION_FLOOR_ADDED, floor);
	}

	/**
	 * Removes a {@link AdministrationFloor} from this metaphor model.
	 * 
	 * @param floor
	 *            The administration floor to remove.
	 */
	public void removeAdministrationFloor(final AdministrationFloor floor) {

		super.notifyObservers(ObservableAction.ADMINISTRATION_FLOOR_REMOVED, floor);
	}

	/**
	 * To be invoked when a thread has terminated, i.e. a trace has reached its end.
	 * 
	 * @param floor
	 *            The administration floor representing the terminated trace.
	 */
	public void addThreadTermination(final AdministrationFloor floor) {

		super.notifyObservers(ObservableAction.THREAD_TERMINATION_ADDED, floor);
	}

	/**
	 * To be invoked when a thread termination has to be undone.
	 * 
	 * @param floor
	 *            The administration floor representing the terminated trace.
	 */
	public void removeThreadTermination(final AdministrationFloor floor) {

		super.notifyObservers(ObservableAction.THREAD_TERMINATION_REMOVED, floor);
	}

	/**
	 * To be invoked when a monitor object has changed its status (i.e. is requested, acquired, released).
	 * 
	 * @param floor
	 *            The administration floor representing the monitor object whose status has changed.
	 */
	public void onMonitorStatusChanged(final AdministrationFloor floor) {

		super.notifyObservers(ObservableAction.MONITOR_STATUS_CHANGED, floor);
	}

}
