/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package models.view;

import java.util.LinkedList;
import java.util.List;

import javax.media.j3d.BadTransformException;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Node;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Text3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

import org.apache.log4j.Logger;

import com.sun.j3d.utils.geometry.Primitive;

import configuration.VisualProperties;

import data.dynamics.STrace;
import data.statics.SPackage;

import metaphor.entities.AdministrationBase;
import metaphor.entities.AdministrationBuilding;
import metaphor.entities.AdministrationDistrict;
import metaphor.entities.AdministrationFloor;
import metaphor.entities.Cable;
import metaphor.entities.Floor;
import metaphor.entities.RegularBase;
import metaphor.entities.RegularBuilding;
import metaphor.entities.RegularDistrict;
import metaphor.entities.RegularFloor;
import metaphor.entities.Status;
import metaphor.entities.Street;
import metaphor.types.VisualEntity;

import models.metaphor.MetaphorUtils;

import observer.AbstractObservable;
import observer.IObserver;
import observer.ObservableAction;

import visual.metaphor.BendableArrow;
import visual.metaphor.BorderedBox;
import visual.metaphor.DashedArrow;
import visual.metaphor.PosterBorderedBox;
import visual.metaphor.SymbolSign;
import visual.metaphor.SymbolSign.Symbol;
import visual.metaphor.utils.AppearanceUtils;
import visual.metaphor.utils.TextUtils;
import visual.metaphor.utils.TransformUtils;

/**
 * A view importer to convert given metaphor model objects into Java3D objects and import them into a {@link ViewModel} instance.
 * 
 * @author pdo
 * 
 */
public class ViewImporter implements IObserver {

	// ============================= Attributes ===============================

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(ViewImporter.class);

	/** The diameter of all street and cable representing arrows in the final visualization. */
	private float lineDiameter = VisualProperties.getInstance().getLineDiameter();

	/** The additive value when computing the radius of a {@link SymbolSign}. */
	private float radiusAdditive = VisualProperties.getInstance().getBuildingInterspace() / 3;

	/** The horizontal and vertical size difference between two consecutive districts. */
	private float districtBorder = VisualProperties.getInstance().getDistrictBorder();

	/** The minimal horizontal distance from a street's bendpoint to a floor's wall. */
	private float streetOffset = Math.max(Math.min(VisualProperties.getInstance().getLineArrowheadDiameter(), VisualProperties.getInstance()
			.getBuildingInterspace()), VisualProperties.getInstance().getBuildingInterspace()) / 2 - 0.01f;

	/** The horizontal distance between two streets having the same source and target floor. */
	private float parallelStreetOffset = VisualProperties.getInstance().getLineArrowheadDiameter() + 0.01f;

	/** The minimal vertical distance from a cable's bendpoint to a floor's roof. */
	private float cableOffset = Math.max(0.1f, VisualProperties.getInstance().getLineArrowheadHeight() * 1.5f);

	/** The vertical distance from a {@link SymbolSign} to a floor's roof. */
	private final float symbolSignOffset = 0.15f;

	/** The {@link ViewModel} instance to add all imported objects to. */
	private final ViewModel viewModel;

	/** The {@link SymbolSign} instances to be displayed on screen. */
	private final SymbolSign[] signs = new SymbolSign[3];

	/** A list of all {@link Cable} instances currently visible in the visualization, i.e. as well imported by this view importer. */
	private List<Cable> cables = new LinkedList<Cable>();

	// ========================== Constructor =================================

	/**
	 * Creates a new {@link ViewImporter} instance importing converting received metaphor model instances into Java3D objects and adding them to the given scene
	 * model.
	 * 
	 * @param viewModel
	 *            The view model to import all Java3d objects into.
	 * @throws IllegalArgumentException
	 *             If the view model passed in is {@code null}.
	 */
	public ViewImporter(final ViewModel viewModel) throws IllegalArgumentException {

		if (viewModel == null) {
			throw new IllegalArgumentException("You cannot pass in a null scene model.");
		}

		this.viewModel = viewModel;
	}

	// ============================ Methods ===================================

	@Override
	public void update(final AbstractObservable observable, final ObservableAction action, final Object object) {

		switch (action) {
		case ADMINISTRATION_FLOOR_ADDED:
			this.onAdministrationFloorAdded((AdministrationFloor) object);
			break;
		case ALL_CLEARED:
			this.onAllCleared();
			break;
		case CABLE_ADDED:
			this.onCableAdded((Cable) object);
			break;
		case DYNAMICS_LOADED:
			this.onDynamicsLoaded((AdministrationBase) object);
			break;
		case FLOOR_ADDED:
			this.onFloorAdded((RegularFloor) object);
			break;
		case STREET_ADDED:
			this.onStreetAdded((Street) object);
			break;
		case STATICS_LOADED:
			this.onStaticsLoaded((RegularBase) object);
			break;
		case ADMINISTRATION_FLOOR_REMOVED:
			this.onAdministrationFloorRemoved((AdministrationFloor) object);
			break;
		case CABLE_REMOVED:
		case FLOOR_REMOVED:
		case STREET_REMOVED:
			this.onEntityRemoved((VisualEntity) object);
			break;
		case THREAD_TERMINATION_ADDED:
			this.onThreadTerminationAdded((AdministrationFloor) object);
			break;
		case THREAD_TERMINATION_REMOVED:
			this.onThreadTerminationRemoved((AdministrationFloor) object);
			break;
		case MONITOR_STATUS_CHANGED:
			this.onMonitorStatusChanged((AdministrationFloor) object);
			break;
		default:
			LOGGER.warn("ObservableAction '" + action + "' not supported.");
		}
	}

	/**
	 * Updates all visual settings attributes of this instance.
	 */
	private void updateAttributes() {

		this.lineDiameter = VisualProperties.getInstance().getLineDiameter();
		this.radiusAdditive = VisualProperties.getInstance().getBuildingInterspace() / 3;
		this.districtBorder = VisualProperties.getInstance().getDistrictBorder();
		this.streetOffset = Math.max(Math.min(VisualProperties.getInstance().getLineArrowheadDiameter(), VisualProperties.getInstance()
				.getBuildingInterspace()), VisualProperties.getInstance().getBuildingInterspace()) / 2 - 0.01f;
		this.parallelStreetOffset = VisualProperties.getInstance().getLineArrowheadDiameter() + 0.01f;
		this.cableOffset = Math.max(0.1f, VisualProperties.getInstance().getLineArrowheadHeight() * 1.5f);

		AppearanceUtils.reloadAttributes();
		PosterBorderedBox.reloadAttributes();
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#CABLE_REMOVED}, {@link ObservableAction#FLOOR_REMOVED} or
	 * {@link ObservableAction#STREET_REMOVED} has been received. This method removes the Java 3D object related to the specified entity from the view model.
	 * 
	 * @param entity
	 *            The entity to remove its Java 3D object from the view model.
	 */
	private void onEntityRemoved(final VisualEntity entity) {

		if (entity instanceof Cable) {
			this.cables.remove((Cable) entity);
		}

		this.viewModel.removeNode(entity.getJava3dObject());
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#ADMINISTRATION_FLOOR_REMOVED} has been received. This method removes the
	 * Java 3D object related to the specified administration floor from the view model and updates the position of the {@link SymbolSign} objects in the
	 * visualization accordingly.
	 * 
	 * @param adminFloor
	 *            The administration floor to remove its Java 3D object from the view model.
	 */
	private void onAdministrationFloorRemoved(final AdministrationFloor adminFloor) {

		// remove floor
		this.onEntityRemoved(adminFloor);

		// handle symbol sign
		final AdministrationBuilding building = adminFloor.getBuilding();
		final int index = building.getDistrict().getBuildings().indexOf(building);

		this.shiftSymbolSign(this.signs[index], -adminFloor.getHeight());
		if (building.getFloors().isEmpty()) {
			this.viewModel.removeNode(this.signs[index]);
		}
	}

	/**
	 * The action to be performed when the static data has been loaded into the metaphor model. It creates a new {@link BorderedBox} for each district and base
	 * instance in the given regular base and adds it to the view model.
	 * 
	 * @param base
	 *            The regular base to be imported.
	 */
	private void onStaticsLoaded(final RegularBase base) {

		// import regular base
		final BorderedBox box = new BorderedBox(base.getLength(), base.getHeight(), base.getWidth());
		box.setColor(base.getColor());
		this.map(base, box);

		// add to model
		final TransformGroup group = TransformUtils.createTransformGroup(base.getPosition());
		group.addChild(box);
		this.viewModel.addNode(group);

		// import districts
		for (RegularDistrict district : base.getDistricts()) {
			this.importRegularDistrict(district);
		}
	}

	/**
	 * Part of the action to be performed when the static data has been loaded into the metaphor model. It creates a new {@link BorderedBox} for the district passed
	 * in and adds it to the view model.
	 * 
	 * @param district
	 *            The regular district to be imported.
	 */
	private void importRegularDistrict(final RegularDistrict district) {

		// import district
		final BorderedBox box = new BorderedBox(district.getLength(), district.getHeight(), district.getWidth());
		box.setColor(district.getColor());
		this.map(district, box);

		final TransformGroup group = TransformUtils.createTransformGroup(district.getPosition());
		group.addChild(box);

		// add label
		final TransformGroup board = TextUtils.createDistrictText(((SPackage) district.getDataObject()).getName(), new Point3f(0, 0, 0));
		final BranchGroup branch = new BranchGroup();
		branch.addChild(board);

		// arrange label
		final Point3d bounds = TextUtils.computeDimensions((Text3D) ((Shape3D) board.getChild(0)).getGeometry());
		final Transform3D trans = new Transform3D();
		trans.setScale(Math.min(1, (district.getWidth() - this.districtBorder) / (bounds.x * 0.045d))); // scale to make it fit on district width
		trans.setScale(Math.min(trans.getScale(), this.districtBorder * 0.7f / (bounds.y * 0.045d))); // scale to make it fit on district length

		trans.setTranslation(new Vector3f((float) (district.getLength() / 2 - (this.districtBorder + bounds.y) * 0.0225d), (float) (district.getHeight() / 2 - 0.008
				* trans.getScale()), (float) district.getWidth() / 2 - this.districtBorder / 2));

		final TransformGroup childGroup = new TransformGroup();
		childGroup.setTransform(trans);
		childGroup.addChild(branch);

		group.addChild(childGroup);
		this.viewModel.addNode(group);

		// import sub-districts
		for (RegularDistrict child : district.getChildDistricts()) {
			this.importRegularDistrict(child);
		}

		// import static class instances
		for (RegularBuilding building : district.getBuildungs()) {
			for (RegularFloor floor : building.getFloors()) {
				this.onFloorAdded(floor);
			}
		}
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#STREET_ADDED} has been received. This method creates a new
	 * {@link BendableArrow} or {@link DashedArrow} object (depending on the given street's characteristics) representing the specified street, sets it as the given
	 * objects associated Java 3D object (if not already existing) and adds it to the view model (i.e. makes it visible on screen).
	 * 
	 * @param street
	 *            The street to import into the view model.
	 */
	private void onStreetAdded(final Street street) {

		final Floor source = street.getSource();
		final Floor target = street.getTarget();

		try {
			final Node node;
			if (source == target) {
				// street goes between the same floor
				node = this.addSelfStreet(street);
			} else if (source.getClass().equals(target.getClass())) {
				if (source instanceof RegularFloor && ((RegularFloor) source).getBuilding() == ((RegularFloor) target).getBuilding()) {
					// street goes between same building
					node = this.addVerticalStreet(street);
				} else if (source instanceof AdministrationFloor && ((AdministrationFloor) source).getBuilding() == ((AdministrationFloor) target).getBuilding()) {
					node = this.addVerticalStreet(street);
				} else {
					// street goes between different buildings
					if (this.numParallelStreets(street) == 0) {
						node = this.addHorizontalStreet(street);
					} else {
						node = this.addParallelHorizontalStreet(street);
					}
				}
			} else {
				// street goes between different buildings
				if (this.numParallelStreets(street) == 0) {
					node = this.addHorizontalStreet(street);
				} else {
					node = this.addParallelHorizontalStreet(street);
				}
			}
			this.viewModel.addNode(node);
		} catch (BadTransformException e) {
			// just a little safety barrier
			LOGGER.debug("Non affine transformation detected.", e);
		}
	}

	/**
	 * Creates and adds a new {@link BendableArrow} to the view model (i.e. to the visualization) representing the {@link Street} instance passed in having the same
	 * {@link RegularFloor} as source and target floor.
	 * 
	 * @param street
	 *            The street to be imported into the view model.
	 * @return The node to be added to the view model.
	 */
	private Node addSelfStreet(final Street street) {
		
		// create new bendable arrow: compute bendpoints
		final float buildingLength;
		final float buildingWidth;
		if (street.getSource() instanceof RegularFloor) {
			final RegularBuilding buildung = ((RegularFloor) street.getSource()).getBuilding();
			buildingLength = buildung.getLength();
			buildingWidth = buildung.getWidth();
		} else {
			final AdministrationBuilding buildung = ((AdministrationFloor) street.getSource()).getBuilding();
			buildingLength = buildung.getLength();
			buildingWidth = buildung.getWidth();
		}

		final Vector3d temp = new Vector3d(street.getSource().getPosition());
		final int parallel = this.numParallelStreets(street);
		temp.x += (buildingLength / 2) + this.streetOffset;
		temp.y += parallel * this.parallelStreetOffset;
		temp.z += (buildingWidth / 2) + this.streetOffset;
		final Vector3d bend1 = new Vector3d(temp);

		temp.x += this.streetOffset;
		temp.z -= this.streetOffset;
		final Vector3d bend2 = new Vector3d(temp);

		temp.z -= buildingWidth;
		final Vector3d bend3 = new Vector3d(temp);

		temp.x -= this.streetOffset;
		temp.z -= this.streetOffset;
		final Vector3d bend4 = new Vector3d(temp);

		// create new arrow instance
		final BendableArrow arrow = new BendableArrow(street.getSource().getPosition(), street.getTarget().getPosition(), this.lineDiameter, bend1, bend2, bend3,
				bend4);
		arrow.setColor(street.getColor());
		this.map(street, arrow);

		return arrow;
	}

	/**
	 * Creates and adds a new {@link BendableArrow} to the view model (i.e. to the visualization) representing the {@link Street} instance passed in having the same
	 * {@link RegularBuilding} as source and target building, but not the same floor.
	 * 
	 * @param street
	 *            The street to be imported into the view model.
	 * @return The node to be added to the view model.
	 */
	private Node addVerticalStreet(final Street street) {

		// create new bendable arrow: compute bendpoints
		final float buildingLength;
		if (street.getSource() instanceof RegularFloor) {
			buildingLength = ((RegularFloor) street.getSource()).getBuilding().getLength();
		} else {
			buildingLength = ((AdministrationFloor) street.getSource()).getBuilding().getLength();
		}

		final Vector3d temp = new Vector3d(street.getSource().getPosition());
		final int parallel = this.numParallelStreets(street);
		temp.x += (buildingLength / 2) + this.streetOffset;
		temp.y += parallel * this.parallelStreetOffset;
		final Vector3d bend1 = new Vector3d(temp);

		temp.set(street.getTarget().getPosition());
		temp.x += (buildingLength / 2) + this.streetOffset;
		final Vector3d bend2 = new Vector3d(temp);

		// create new arrow instance
		final BendableArrow arrow = new BendableArrow(street.getSource().getPosition(), street.getTarget().getPosition(), this.lineDiameter, bend1, bend2);
		arrow.setColor(street.getColor());
		this.map(street, arrow);

		return arrow;
	}

	/**
	 * Creates and adds a new {@link BendableArrow} to the view model (i.e. to the visualization) representing the {@link Street} instance passed in having a
	 * different {@link RegularBuilding} (or {@link AdministrationBuilding}) as source and target building.
	 * 
	 * @param street
	 *            The street to be imported into the view model.
	 * @return The node to be added to the view model.
	 */
	private Node addHorizontalStreet(final Street street) {

		final Primitive arrow;
		// create new arrow
		if (street.getStatus() == Status.OPENED) {
			arrow = new BendableArrow(street.getSource().getPosition(), street.getTarget().getPosition(), this.lineDiameter);
			((BendableArrow) arrow).setColor(street.getColor());
		} else if (street.getStatus() == Status.CLOSED) {
			arrow = new DashedArrow(street.getSource().getPosition(), street.getTarget().getPosition(), this.lineDiameter);
			((DashedArrow) arrow).setColor(street.getColor());
		} else {
			final String error = "The given street has an unsupported status: " + street.getStatus().getName();
			LOGGER.error(error);
			throw new IllegalStateException(error);
		}

		this.map(street, arrow);
		return arrow;
	}

	/**
	 * Creates and adds a new {@link BendableArrow} to the view model (i.e. to the visualization) representing the {@link Street} instance passed in having a
	 * different {@link RegularBuilding} (or {@link AdministrationBuilding}) as source and target building and has the same source and target floor as another street
	 * currently visualized.
	 * 
	 * @param street
	 *            The street to be imported into the view model.
	 * @return The node to be added to the view model.
	 */
	private Node addParallelHorizontalStreet(final Street street) {

		final Primitive arrow;
		final int parallel = this.numParallelStreets(street);

		// compute bendpoints
		final Vector3d connector = new Vector3d(street.getTarget().getPosition());
		connector.sub(street.getSource().getPosition());
		connector.scale(0.2f);

		final Vector3d bend1 = new Vector3d(street.getSource().getPosition());
		bend1.add(connector);
		bend1.y += parallel * this.parallelStreetOffset;

		final Vector3d bend2 = new Vector3d(street.getTarget().getPosition());
		bend2.sub(connector);
		bend2.y += parallel * this.parallelStreetOffset;

		// create new arrow
		if (street.getStatus() == Status.OPENED) {
			arrow = new BendableArrow(street.getSource().getPosition(), street.getTarget().getPosition(), this.lineDiameter, bend1, bend2);
			((BendableArrow) arrow).setColor(street.getColor());
		} else if (street.getStatus() == Status.CLOSED) {
			arrow = new DashedArrow(street.getSource().getPosition(), street.getTarget().getPosition(), this.lineDiameter, bend1, bend2);
			((DashedArrow) arrow).setColor(street.getColor());
		} else {
			final String error = "The given street has an unsupported status: " + street.getStatus().getName();
			LOGGER.error(error);
			throw new IllegalStateException(error);
		}

		this.map(street, arrow);
		return arrow;
	}

	/**
	 * Computes and returns the number of parallel streets relative to the specified {@link Street} instance, which is defined as the number of streets having the
	 * same source and target {@link Floor} (or crosswise reverse) currently being visible on screen. Those might overlap if they are not handled correctly by
	 * slightly shifting them apart.
	 * 
	 * @param street
	 *            The street to compute the number of its parallel streets.
	 * @return The number of parallel streets relative to the given instance.
	 */
	private int numParallelStreets(final Street street) {

		int retval = -1; // since given street is already part of the streets list
		for (Street str : street.getSource().getStreets()) {
			if (str.getTarget().equals(street.getTarget())) {
				retval++;
			}
		}

		if (street.getSource() != street.getTarget()) {
			for (Street str : street.getTarget().getStreets()) {
				if (str.getTarget().equals(street.getSource())) {
					retval++;
				}
			}
		}

		return retval;
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#FLOOR_ADDED} has been received. This method creates a new
	 * {@link BorderedBox} object representing the specified regular floor, sets it as the given objects associated Java 3D object (if not already existing) and adds
	 * it to the view model (i.e. makes it visible on screen).
	 * 
	 * @param floor
	 *            The regular floor to import into the view model.
	 */
	private void onFloorAdded(final RegularFloor floor) {

		// create new bordered Box
		final RegularBuilding building = floor.getBuilding();
		final BorderedBox box = new BorderedBox(building.getLength(), floor.getHeight(), building.getWidth());
		box.setColor(floor.getColor());
		this.map(floor, box);

		// add to model
		final TransformGroup group = TransformUtils.createTransformGroup(floor.getPosition());
		group.addChild(box);
		this.viewModel.addNode(group);
	}

	/**
	 * The action to be performed when the dynamics data has been loaded into the metaphor model. It creates a new {@link BorderedBox} for each district and base
	 * instance in the given administration base and adds it to the view model.
	 * 
	 * @param adminBase
	 *            The administration base to be imported.
	 */
	private void onDynamicsLoaded(final AdministrationBase adminBase) {

		// import administration base
		final BorderedBox box = new BorderedBox(adminBase.getLength(), adminBase.getHeight(), adminBase.getWidth());
		box.setColor(adminBase.getColor());
		this.map(adminBase, box);

		// add to model
		final TransformGroup group = TransformUtils.createTransformGroup(adminBase.getPosition());
		group.addChild(box);
		this.viewModel.addNode(group);

		// import administration district
		final AdministrationDistrict district = adminBase.getDistrict();
		final BorderedBox districtBox = new BorderedBox(district.getLength(), district.getHeight(), district.getWidth());
		districtBox.setColor(district.getColor());
		this.map(district, districtBox);

		// add to model
		final TransformGroup districtGroup = TransformUtils.createTransformGroup(district.getPosition());
		districtGroup.addChild(districtBox);
		this.viewModel.addNode(districtGroup);

		// create symbol signs
		int index = 0;
		for (AdministrationBuilding building : district.getBuildings()) {
			final Vector3d position = new Vector3d(building.getPosition());
			final float radius = building.getWidth() + this.radiusAdditive;
			position.y += radius + this.symbolSignOffset;
			this.signs[index] = new SymbolSign(position, radius, Symbol.values()[index]);
			index++;
		}
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#CABLE_ADDED} has been received. This method creates a new
	 * {@link BendableArrow} object representing the specified cable, sets it as the given object's associated Java 3D object (if not already existing) and adds it
	 * to the view model (i.e. makes it visible on screen).
	 * 
	 * @param cable
	 *            The cable to import into the view model.
	 */
	private void onCableAdded(final Cable cable) {

		// create new bendable arrow: compute bendpoints
		final double sourceHeight = cable.getSource().getPosition().getY() + MetaphorUtils.computeHeigth(cable.getSource());
		final double targetHeight = cable.getTarget().getPosition().getY() + MetaphorUtils.computeHeigth(cable.getTarget());
		final double height = Math.max(sourceHeight, targetHeight);

		final Vector3d temp = new Vector3d(cable.getSource().getPosition());
		temp.y = height + this.computeCableOffset(cable);
		final Vector3d bend1 = new Vector3d(temp);

		temp.set(cable.getTarget().getPosition());
		temp.y = height + this.computeCableOffset(cable);
		final Vector3d bend2 = new Vector3d(temp);

		final Vector3d diff = new Vector3d(bend1);
		diff.sub(bend2);

		if (cable.getSource().equals(cable.getTarget())) {
			// self-relation
			bend1.x += cable.getSource().getLength() / 2;
			bend1.z += cable.getSource().getWidth() / 2;

			bend2.x -= cable.getSource().getLength() / 2;
			bend2.z -= cable.getSource().getWidth() / 2;
		} else {
			double length = Math.min(cable.getSource().getLength(), cable.getSource().getWidth()) / 2 / diff.length();
			bend1.x -= diff.x * length;
			bend1.z -= diff.z * length;

			length = Math.min(cable.getTarget().getLength(), cable.getTarget().getWidth()) / 2 / diff.length();
			bend2.x += diff.x * length;
			bend2.z += diff.z * length;
		}

		// create new arrow instance
		final Vector3d sourcePos = new Vector3d(cable.getSource().getPosition());
		sourcePos.y = sourceHeight;

		final Vector3d targetPos = new Vector3d(cable.getTarget().getPosition());
		targetPos.y = targetHeight;

		final BendableArrow arrow = new BendableArrow(sourcePos, targetPos, this.lineDiameter, bend1, bend2);
		arrow.setColor(cable.getColor());
		this.map(cable, arrow);

		// add to model
		this.cables.add(cable);
		this.viewModel.addNode(arrow);
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#ALL_CLEARED} has been received. This method removes all contents from the
	 * view model.
	 */
	private void onAllCleared() {

		this.cables.clear();
		this.viewModel.clearAll();
		this.updateAttributes();
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#ADMINISTRATION_FLOOR_ADDED} has been received. This method creates a new
	 * {@link PosterBorderedBox} object representing the specified administration floor, sets it as the given objects associated Java 3D object (if not already
	 * existing) and adds it to the view model (i.e. makes it visible on screen).
	 * 
	 * @param adminFloor
	 *            The administration floor to import into the view model.
	 */
	private void onAdministrationFloorAdded(final AdministrationFloor adminFloor) {

		// create new poster bordered Box
		final AdministrationBuilding building = adminFloor.getBuilding();
		final PosterBorderedBox box = new PosterBorderedBox(building.getLength(), adminFloor.getHeight(), building.getWidth());
		box.setColor(adminFloor.getColor());
		this.map(adminFloor, box);

		// set poster color
		if (adminFloor.getDataObject() instanceof STrace) {
			box.setPosterColor(adminFloor.getPoster().getColor());
			box.setPostersVisible(true);
		}

		// handle symbol sign
		final int index = building.getDistrict().getBuildings().indexOf(building);
		this.shiftSymbolSign(this.signs[index], adminFloor.getHeight());
		if (building.getFloors().get(0).equals(adminFloor)) {
			this.viewModel.addNode(this.signs[index]);
		}

		// add to model
		final TransformGroup group = TransformUtils.createTransformGroup(adminFloor.getPosition());
		group.addChild(box);
		this.viewModel.addNode(group);
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#THREAD_TERMINATION_ADDED} has been received. This method hides the posters
	 * of the
	 * Java 3D object representing the specified {@link AdministrationFloor} instance.
	 * 
	 * @param adminFloor
	 *            The administration floor representing the terminated thread.
	 */
	private void onThreadTerminationAdded(final AdministrationFloor adminFloor) {

		final PosterBorderedBox box = (PosterBorderedBox) adminFloor.getJava3dObject();
		box.setPostersVisible(false);
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#THREAD_TERMINATION_REMOVED} has been received. This method makes the
	 * posters of the
	 * Java 3D object representing the specified {@link AdministrationFloor} instance visible.
	 * 
	 * @param adminFloor
	 *            The administration floor representing the terminated thread.
	 */
	private void onThreadTerminationRemoved(final AdministrationFloor adminFloor) {

		final PosterBorderedBox box = (PosterBorderedBox) adminFloor.getJava3dObject();
		box.setPostersVisible(true);
	}

	/**
	 * The action to be performed when a change notification with {@link ObservableAction#MONITOR_STATUS_CHANGED} has been received. This method updates the Java 3D
	 * poster object mapped to the {@link AdministrationFloor} passed in according to the new status.
	 * 
	 * @param floor
	 *            The administration floor representing the monitor object whose status has changed.
	 */
	private void onMonitorStatusChanged(final AdministrationFloor floor) {

		final PosterBorderedBox box = (PosterBorderedBox) floor.getJava3dObject();
		if (floor.getPoster().getColor() == null) {
			box.setPostersVisible(false);
		} else {
			box.setPosterColor(floor.getPoster().getColor());
			box.setPostersVisible(true);
		}
	}

	/**
	 * Maps the given objects to each other. It sets the given {@code entity} as the {@code node}'s user data object and the {@code node} as {@code entity}'s
	 * Java3d object.
	 * 
	 * @param entity
	 *            The visual entity to map to the given {@link VisualizableElement} object.
	 * @param node
	 *            The Java3D node to map to the given {@link VisualEntity} instance.
	 */
	private void map(final VisualEntity entity, final Primitive node) {

		entity.setJava3dObject(node);
		node.setUserData(entity);
	}

	/**
	 * Computes the distance of the specified cable's height to the source or target building's roof with which the cable is supposed to be drawn.
	 * 
	 * @param cable
	 *            The cable to compute its vertical offset.
	 * @return The computed cable's offset.
	 */
	private double computeCableOffset(final Cable cable) {

		double offset = this.cableOffset;
		for (Cable c : this.cables) {
			if (c.getSource().equals(cable.getSource()) && c.getTarget().equals(cable.getTarget()) || c.getTarget().equals(cable.getSource())
					&& c.getSource().equals(cable.getTarget())) {
				offset += this.cableOffset;
			}
		}

		return offset;
	}

	/**
	 * Shifts the specified {@link SymbolSign} instance by the given distance in the vertical axis (y-axis).
	 * 
	 * @param sign
	 *            The sign to shift.
	 * @param distance
	 *            The distance by which the specified symbol sign is to be translated. A positive value to shift it up, a negative value to shift it down.
	 */
	private void shiftSymbolSign(final SymbolSign sign, final float distance) {

		// retrieve transform3d object
		final Transform3D transform3d = new Transform3D();
		final TransformGroup group = (TransformGroup) sign.getChild(0);
		group.getTransform(transform3d);

		// shift sign up / down
		final Transform3D shiftTG = new Transform3D();
		shiftTG.setTranslation(new Vector3d(0, 0, -distance));
		transform3d.mul(shiftTG);
		group.setTransform(transform3d);
	}

}
