/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package observer;

/**
 * An enumeration indicating the kind of or the action causing a change in the {@link AbstractObservable} object.
 * 
 * @author pdo
 * 
 * @see IObserver
 */
public enum ObservableAction {

	/*
	 * Data model related actions. To be handled by the metaphor importer.
	 */

	/** Indicates that all static program data has been successfully loaded into the observed object. */
	STATICS_LOADED,

	/** Indicates that all dynamic program data has been successfully loaded into the observed object. */
	DYNAMICS_LOADED,

	/** Indicates that all objects have been removed from the observed object. */
	ALL_CLEARED,

	/** Indicates that a new {@link SClassInstance} has been added to the observed object. The added instance will be passed by the optional parameter. */
	CLASS_INSTANCE_ADDED,

	/** Indicates that a {@link SClassInstance} has been removed from the observed object. The removed instance will be passed by the optional parameter. */
	CLASS_INSTANCE_REMOVED,

	/** Indicates that a new {@link SThreadEvent} has been added to the observed object. The added event will be passed by the optional parameter. */
	THREAD_EVENT_ADDED,

	/** Indicates that a {@link SThreadEvent} has been removed from the observed object. The removed event will be passed by the optional parameter. */
	THREAD_EVENT_REMOVED,

	/** Indicates that a {@link MonitorObject} has been added to the observed object. The added instance will be passed by the optional parameter. */
	MONITOR_OBJECT_ADDED,

	/** Indicates that a {@link MonitorObject} has been removed from the observed object. The removed instance will be passed by the optional parameter. */
	MONITOR_OBJECT_REMOVED,

	/** Indicates that a {@link WaitNotifyObject} has been added to the observed object. The added instance will be passed by the optional parameter. */
	WAIT_NOTIFY_OBJECT_ADDED,

	/** Indicates that a {@link WaitNotifyObject} has been removed from the observed object. The removed instance will be passed by the optional parameter. */
	WAIT_NOTIFY_OBJECT_REMOVED,

	/** Indicates that a new {@link STrace} instance has been added to the observed object. The added instance will be passed by the optional parameter. */
	THREAD_INSTANCE_ADDED,

	/** Indicates that a {@link STrace} instance has been removed from the observed object. The removed instance will be passed by the optional parameter. */
	THREAD_INSTANCE_REMOVED,

	/** Indicates that a new {@link Relation} has been added to the observed object. The added instance will be passed by the optional parameter. */
	RELATION_ADDED,

	/** Indicates that a {@link Relation} has been removed from the observed object. The removed instance will be passed by the optional parameter. */
	RELATION_REMOVED,

	/*
	 * Metaphor model related actions. To be handled by the view importer.
	 */

	/** Indicates that a {@link Floor} has been removed from the observed object. The removed instance will be passed by the optional parameter. */
	FLOOR_REMOVED,

	/** Indicates that a new {@link Floor} has been added to the observed object. The added instance will be passed by the optional parameter. */
	FLOOR_ADDED,

	/** Indicates that a new {@link Cable} has been added to the observed object. The added instance will be passed by the optional parameter. */
	CABLE_ADDED,

	/** Indicates that a {@link Cable} has been removed from the observed object. The removed instance will be passed by the optional parameter. */
	CABLE_REMOVED,

	/** Indicates that a new {@link Street} has been added to the observed object. The added instance will be passed by the optional parameter. */
	STREET_ADDED,

	/** Indicates that a {@link Street} has been removed from the observed object. The removed instance will be passed by the optional parameter. */
	STREET_REMOVED,

	/** Indicates that a new {@link AdministrationFloor} has been added to the observed object. The added instance will be passed by the optional parameter. */
	ADMINISTRATION_FLOOR_ADDED,

	/** Indicates that a {@link AdministrationFloor} has been removed from the observed object. The removed instance will be passed by the optional parameter. */
	ADMINISTRATION_FLOOR_REMOVED,

	/*
	 * Model-unspecific actions. To be handled by every model importer.
	 */

	/**
	 * Indicates that a {@link STrace} has reached its end, i.e. the corresponding thread has terminated. The corresponding {@link AdministrationFloor} will be
	 * passed by the optional parameter.
	 */
	THREAD_TERMINATION_ADDED,

	/**
	 * Indicates, that a previous {@link #THREAD_TERMINATION_ADDED} event has to be undone. The corresponding {@link AdministrationFloor} will be passed by
	 * the optional parameter.
	 */
	THREAD_TERMINATION_REMOVED,

	/**
	 * Indicates that the status of a {@link MonitorObject} has changed. The {@link AdministrationFloor} representing the monitor object will be passed by the
	 * optional parameter.
	 */
	MONITOR_STATUS_CHANGED,
}
