/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package utils;

import data.dynamics.STrace;
import data.dynamics.STraceEvent;

/**
 * A utility class whose instances are used as the {@link DefaultMutableTreeNode}'s user objects in order to map each tree node to its representing trace
 * event in {@link TraceTreeViewController}.
 * 
 * @author pdo
 * 
 */
public final class NodeObject {

	// =========================== Attributes ================================

	/** The trace event this node object contains. */
	private final STraceEvent event;

	/** The trace this node object contains. */
	private final STrace trace;

	/** The name of this node object. */
	private String name;

	// ========================== Constructor =================================

	/**
	 * Constructs a new {@link NodeObject} instance with the given name wrapping the specified trace event.
	 * 
	 * @param name
	 *            The name of this node object.
	 * @param event
	 *            The trace event this node object is supposed to wrap. May be {@code null}.
	 * @param trace
	 *            The trace this node object is supposed to wrap. May be {@code null}.
	 * @throws IllegalArgumentException
	 *             If the name passed in is {@code null}.
	 */
	public NodeObject(final String name, final STrace trace, final STraceEvent event) throws IllegalArgumentException {

		if (name == null) {
			throw new IllegalArgumentException("A node object cannot have a null name.");
		}

		this.name = name;
		this.event = event;
		this.trace = trace;
	}

	// ============================ Methods ===================================

	/**
	 * Returns the wrapped trace of this node object.
	 * 
	 * @return The trace wrapped by this node object.
	 */
	public STrace getTrace() {
		return this.trace;
	}

	/**
	 * Returns the wrapped trace event of this node object.
	 * 
	 * @return The trace event wrapped by this node object.
	 */
	public STraceEvent getTraceEvent() {
		return this.event;
	}

	/**
	 * Sets the name of this {@link NodeObject} instance.
	 * 
	 * @param newName
	 *            The new name to set.
	 * @throws IllegalArgumentException
	 *             If the name passed in is {@code null}.
	 */
	public void setName(final String newName) throws IllegalArgumentException {

		if (newName == null) {
			throw new IllegalArgumentException("A node object cannot have a null name.");
		}

		this.name = newName;
	}

	@Override
	public String toString() {
		return this.name;
	}
}
