/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.gui;

import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JPanel;

import org.apache.log4j.Logger;

import configuration.Language;



import logic.Logic;

/**
 * A city control bar allowing the user to control the current visualization. It offers buttons to start and stop the automatic replay and to jump one (or more)
 * event forward or backwards in the currently imported Kieker trace.
 * 
 * @author pdo
 * 
 */
public final class CityControlBar extends JPanel {

	// =========================== Attributes =================================

	/** The generated {@code serialVersionUID} for serialization. */
	private static final long serialVersionUID = -1126294691588281776L;

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(CityControlBar.class);

	/** The {@link JButton} objects displayed in this city control bar. See {@link ControlBarButton} for their order in this array. */
	private JButton[] buttons = new JButton[ControlBarButton.values().length];

	/** The {@link CityActionListener} to perform the action associated to a button. */
	private final CityActionListener listener = new CityActionListener();

	// ========================== Constructor =================================

	/**
	 * Creates a new {@link CityControlBar} instance.
	 */
	public CityControlBar() {

		super();
		Logic.getInstance().getCityController().init(this);
		this.setLayout(new GridLayout(1, 6));
		this.rebuild();

		// Initially, all buttons are disabled
		for (ControlBarButton button : ControlBarButton.values()) {
			this.buttons[button.ordinal()].setEnabled(false);
		}
	}

	// ============================ Methods ===================================

	/**
	 * Rebuilds the entire panel. All buttons will be replaced by new ones. However, all enabled conditions will be adopted from the old buttons.
	 */
	public void rebuild() {

		for (ControlBarButton button : ControlBarButton.values()) {
			boolean enabled = true;
			if (this.buttons[button.ordinal()] != null) {
				this.remove(this.buttons[button.ordinal()]);
				enabled = this.buttons[button.ordinal()].isEnabled();
			}
			this.buttons[button.ordinal()] = new JButton(Language.getInstance().get("citycontrolbar.button." + button.name().replace('_', '.').toLowerCase()));
			this.buttons[button.ordinal()].addActionListener(this.listener);
			this.buttons[button.ordinal()].setEnabled(enabled);
			this.add(this.buttons[button.ordinal()]);
		}
	}

	/**
	 * Enables or disables the specified button.
	 * 
	 * @param button
	 *            The button to enable or disable represented by its corresponding {@link ControlBarButton} object.
	 * @param enabled
	 *            {@code True} to enable, {@code false} to disable.
	 * @throws IllegalArgumentException
	 *             If the enumeration object representing the specified button is {@code null}.
	 */
	public void setButtonEnabled(final ControlBarButton button, final boolean enabled) throws IllegalArgumentException {

		if (button == null) {
			throw new IllegalArgumentException("You cannot set the enabled status of a null button.");
		}

		this.buttons[button.ordinal()].setEnabled(enabled);
	}

	/**
	 * An enumeration of all buttons displayed in a {@link CityControlBar}.
	 * 
	 * @author pdo
	 * 
	 */
	public static enum ControlBarButton {

		/** The button to jump X milliseconds backwards in time. */
		PREVIOUS_TIME,

		/** The button to jump to the previous event. */
		PREVIOUS_EVENT,

		/** The button to start the replay. */
		REPLAY_START,

		/** The button to pause the replay. */
		REPLAY_PAUSE,

		/** The button to jump to the next event. */
		NEXT_EVENT,

		/** The button to jump X milliseconds forwards in time. */
		NEXT_TIME,
	}

	/**
	 * A custom action listener to perform the appropriate action related to the {@link CityControlBar} button the user clicked on.
	 * 
	 * @author pdo
	 * 
	 */
	private class CityActionListener implements ActionListener {

		/**
		 * Creates a new {@link CityActionListener} instance.
		 */
		protected CityActionListener() {/* nothing to do */}

		@Override
		public void actionPerformed(final ActionEvent event) {

			final JButton source = (JButton) event.getSource();
			if (source.equals(CityControlBar.this.buttons[ControlBarButton.PREVIOUS_TIME.ordinal()])) {
				Logic.getInstance().getCityController().oneIntervalBackwards();
			} else if (source.equals(CityControlBar.this.buttons[ControlBarButton.PREVIOUS_EVENT.ordinal()])) {
				Logic.getInstance().getCityController().oneStepBackwards();
			} else if (source.equals(CityControlBar.this.buttons[ControlBarButton.REPLAY_START.ordinal()])) {
				Logic.getInstance().getCityController().startReplay();
			} else if (source.equals(CityControlBar.this.buttons[ControlBarButton.REPLAY_PAUSE.ordinal()])) {
				Logic.getInstance().getCityController().pauseReplay();
			} else if (source.equals(CityControlBar.this.buttons[ControlBarButton.NEXT_EVENT.ordinal()])) {
				Logic.getInstance().getCityController().oneStepForwards();
			} else if (source.equals(CityControlBar.this.buttons[ControlBarButton.NEXT_TIME.ordinal()])) {
				Logic.getInstance().getCityController().oneIntervalForwards();
			} else {
				final String error = "Unsupported control bar button detected: " + source.getName();
				LOGGER.error(error);
				throw new IllegalStateException(error);
			}
		}
	}

}
