/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.gui;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;

import configuration.Configuration;
import configuration.Language;

import logic.Logic;
import logic.SettingsListeners;
import logic.TraceTreeViewController.TreeStructure;

/**
 * The trace tree view allowing the user to quickly change the trace view layout.
 * 
 * @author pdo
 * 
 */
public class TraceViewPanel extends JPanel {

	/** The generated {@code serialVersionUID} for serialization. */
	private static final long serialVersionUID = 8992842225461138070L;

	/** The {@link GridBagConstraints} object used to communicate the position and layout of all components to add to this {@link SettingsWindow}. */
	private GridBagConstraints constants = new GridBagConstraints();

	/** A list of all components currently added to this dialog window. */
	private final List<Component> components = new LinkedList<Component>();

	private JComboBox<String> comboBox;

	/**
	 * Creates a new {@link TraceViewPanel} instance.
	 */
	public TraceViewPanel() {

		super();

		// set layout
		final GridBagLayout layout = new GridBagLayout();
		this.setLayout(layout);

		this.setMinimumSize(new Dimension(20, 30));
		this.setMaximumSize(new Dimension(20, 30));

		this.constants.fill = GridBagConstraints.HORIZONTAL;

		// register at controller
		Logic.getInstance().getTraceViewPanelController().init(this);

		// add components
		this.rebuild();
	}

	/**
	 * Rebuilds this entire {@link TraceViewPanel} object.
	 */
	public void rebuild() {

		// remove old components
		for (Component comp : this.components) {
			this.remove(comp);
		}
		this.components.clear();

		// add new components
		this.addTraceTreeViewSettings();
		this.revalidate();
		this.repaint();
	}

	/**
	 * Adds the check box allowing the user to set the trace tree views layout at the position defined in {@link #constants}.
	 */
	private void addTraceTreeViewSettings() {

		// reset grid constants
		this.constants.gridx = 0;
		this.constants.gridy = 0;

		// add label
		this.addLabel(Language.getInstance().get("menubar.settings.traceviewer.layout.title") + "     ");

		// add combo box
		this.comboBox = new JComboBox<String>();
		for (TreeStructure structure : TreeStructure.values()) {
			this.comboBox.addItem(Language.getInstance().get("menubar.settings.traceviewer.layout." + structure.name().toLowerCase()));
		}
		this.comboBox.setSelectedIndex(Configuration.getInstance().getTraceViewerLayout().ordinal());
		this.comboBox.addActionListener(SettingsListeners.createTraceViewListener(this.comboBox));

		this.add(this.comboBox, this.constants);
		this.components.add(this.comboBox);
	}

	/**
	 * Adds a new label containing the specified text at the position defined in {@link #constants}.
	 * 
	 * @param text
	 *            The text of the label to display.
	 */
	private void addLabel(final String text) {

		final JLabel label = new JLabel(text);
		this.add(label, this.constants);
		this.components.add(label);
		this.constants.gridx++;
	}

}
