/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.metaphor;

import java.awt.Color;

import javax.media.j3d.TransformGroup;
import javax.vecmath.Color3f;
import javax.vecmath.Vector3d;

import com.sun.j3d.utils.geometry.Box;

import visual.metaphor.utils.AppearanceUtils;
import visual.metaphor.utils.TransformUtils;

/**
 * A regular {@link com.sun.j3d.utils.geometry.Box Box}, which differs from the original implementation in the way, that
 * its edges are emphasized by solid (dyeable) lines, its {@link javax.media.j3d.Appearance Appearance} is readable and setable
 * and pick reporting is enabled.
 * 
 * @author pdo
 * 
 */
public class BorderedBox extends Box {
	
	// =========================== Attributes =================================
	
	/** The default interspace between the box and its surrounding border. */
	private static final float BORDER_OFFSET = 0.0005f;

	// ========================== Constructors ================================

	/**
	 * Constructs a new {@link BorderedBox} instance with a given length, width and height.
	 * 
	 * @param length
	 *            The length (x-dimension) of the new bordered box.
	 * @param height
	 *            The height (y-dimension) of the new bordered box.
	 * @param width
	 *            The width (z-dimension) of the new bordered box.
	 * @throws IllegalArgumentException
	 *             If either the given width, height or length is negative.
	 */
	public BorderedBox(final float length, final float height, final float width) throws IllegalArgumentException {

		super(length / 2, height / 2, width / 2, AppearanceUtils.createReadWriteAppearance());

		if (length < 0 || height < 0 || width < 0) {
			throw new IllegalArgumentException("A box cannot have negative dimension attributes.");
		}

		this.setCapability(ENABLE_PICK_REPORTING);
		this.setCapability(ALLOW_PARENT_READ);
		this.setCollidable(true);

		// create a new cuboid whose edges are visible only
		final Cuboid border = new Cuboid(this.getXdimension() + BORDER_OFFSET, this.getYdimension(), this.getZdimension() + BORDER_OFFSET);
		final TransformGroup trans = TransformUtils.createTransformGroup(new Vector3d(0, BORDER_OFFSET, 0));
		trans.addChild(border);
		border.setAppearance(AppearanceUtils.createBorderAppearance());
		border.setPickable(false);
		this.addChild(trans);
	}

	/**
	 * Construct a new {@link BorderedBox} instance with the specified length, width, height and color.
	 * 
	 * @param length
	 *            The length (x-dimension) of the bordered box.
	 * @param height
	 *            The height (y-dimension) of the bordered box.
	 * @param width
	 *            The width (z-dimension) of the bordered box.
	 * @param color
	 *            The {@link Color} of the bordered box.
	 * @throws IllegalArgumentException
	 *             If the color passed in is {@code null} or if either the given width, height, or length is negative.
	 */
	public BorderedBox(final float length, final float height, final float width, final Color color) throws IllegalArgumentException {

		this(length, height, width);
		this.setColor(color);
	}

	// ============================ Methods ===================================

	/**
	 * Set the {@link Color} of this {@link BorderedBox} instance.
	 * 
	 * @param color
	 *            The new color of this bordered box.
	 * @throws IllegalArgumentException
	 *             If the color passed in is {@code null}.
	 */
	public final void setColor(final Color color) throws IllegalArgumentException {

		if (color == null) {
			throw new IllegalArgumentException("You cannot set a null color.");
		}

		for (int i = FRONT; i <= BOTTOM; i++) {
			this.getAppearance(i).getColoringAttributes().setColor(new Color3f(color));
		}
	}
}
