/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.metaphor;

import java.awt.Color;

import javax.media.j3d.Appearance;
import javax.media.j3d.Shape3D;
import javax.vecmath.Color3f;

import com.sun.j3d.utils.geometry.Primitive;

/**
 * A regular cuboid created with a given length, width, and height centered at the origin located at {@code (0,0,0)}. In contrast to Java3D's regular
 * {@link com.sun.j3d.utils.geometry.Box Box} implementation, the faces of this cuboid are comprised of rectangulars instead of triangulars.
 * 
 * @author pdo
 */
public class Cuboid extends Primitive {

	// =========================== Attributes =================================

	/** The id representing the front face of this {@link Cuboid} instance. */
	public static final int FRONT = 0;

	/** The id representing the back face of this {@link Cuboid} instance. */
	public static final int BACK = 1;

	/** The id representing the right face of this {@link Cuboid} instance. */
	public static final int RIGHT = 2;

	/** The id representing the left face of this {@link Cuboid} instance. */
	public static final int LEFT = 3;

	/** The id representing the top face of this {@link Cuboid} instance. */
	public static final int TOP = 4;

	/** The id representing the bottom face of this {@link Cuboid} instance. */
	public static final int BOTTOM = 5;

	/** The six faces of this cuboid. */
	private final Shape3D[] faces = new Shape3D[6];

	/** The {@link Appearance} of this {@link Cuboid}. */
	private Appearance app;

	// ========================== Constructor =================================

	/**
	 * Constructs a new {@link Cuboid} instance with the given dimension attributes centered at the origin {@code (0,0,0)}.
	 * 
	 * @param length
	 *            The x-dimension size (length) of the cuboid to create.
	 * @param heigth
	 *            The y-dimension size (height) of the cuboid to create.
	 * @param width
	 *            the z-dimension size (width) of the cuboid to create.
	 * @throws IllegalArgumentException
	 *             If at least one dimension value passed in is negative.
	 */
	public Cuboid(final float length, final float heigth, final float width) throws IllegalArgumentException {

		super();
		if (length < 0 || heigth < 0 || width < 0) {
			throw new IllegalArgumentException("You cannot create a cuboid with negative dimension sizes.");
		}

		for (int i = FRONT; i <= BOTTOM; i++) {
			final Shape3D face = CuboidBuilder.createFace(length, heigth, width, i);
			face.setAppearance(this.app);
			face.setCollidable(true);
			this.faces[i] = face;

			this.addChild(face);
		}
	}

	// ============================ Methods ===================================

	/**
	 * Sets the color of this {@link Cuboid} instance.
	 * 
	 * @param color
	 *            The new color of this cuboid.
	 * @throws IllegalArgumentException
	 *             If the color passed in is {@code null}.
	 */
	public final void setColor(final Color color) throws IllegalArgumentException {

		if (color == null) {
			throw new IllegalArgumentException("You cannot set a null color.");
		}

		for (int i = FRONT; i <= BOTTOM; i++) {
			this.faces[i].getAppearance().getColoringAttributes().setColor(new Color3f(color));
		}
	}

	/**
	 * Returns the specified face's ({@link Shape3D}) object of this {@link Cuboid}.
	 * 
	 * @param face
	 *            The Id of the face's shape to return. Use {@value #FRONT} for the front, {@value #BACK} for the back, {@value #LEFT} for the left, {@value #RIGHT}
	 *            for the right, {@value #TOP} for the top, and {@value #BOTTOM} for the bottom face.
	 * @return The Shape3D object associated with the specified partID. If an invalid partID is passed in, {@code null} is returned.
	 */
	@Override
	public Shape3D getShape(final int face) {

		if (face >= FRONT && face <= BOTTOM) {
			return (Shape3D) this.faces[face];
		}
		return null;
	}

	/**
	 * Sets the appearance of this {@link Cuboid} instance. This method will set each face of the cuboid to the same appearance. To set each face's appearance
	 * separately, use {@link #getShape(int)} to get the individual shape and call {@code shape.setAppearance(Appearance)}.
	 * 
	 * @param appearance
	 *            The new appearance to set.
	 * @throws IllegalArgumentException
	 *             If the specified appearance is {@code null}.
	 */
	@Override
	public void setAppearance(final Appearance appearance) throws IllegalArgumentException {

		if (appearance == null) {
			throw new IllegalArgumentException("You cannot set a null appearance.");
		}

		for (int i = FRONT; i <= BOTTOM; i++) {
			((Shape3D) this.getChild(i)).setAppearance(appearance);
		}
	}

	/**
	 * Get the appearance of the specified part of this cuboid.
	 * 
	 * @param faceId
	 *            The faceId of the part of this cuboid to get its appearance. Use {@value #FRONT} for the front, {@value #BACK} for the back, {@value #LEFT} for the
	 *            left, {@value #RIGHT} for the right, {@value #TOP} for the top, and {@value #BOTTOM} for the bottom face.
	 * @return The appearance object associated with the face. If an invalid faceId is passed in, {@code null} is returned.
	 */
	@Override
	public Appearance getAppearance(final int faceId) {

		if (faceId >= FRONT && faceId <= BOTTOM) {
			return this.getShape(faceId).getAppearance();
		}
		return null;
	}

}
