/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.metaphor;

import java.awt.Color;
import java.util.LinkedList;
import java.util.List;

import javax.media.j3d.Appearance;
import javax.media.j3d.Shape3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Color3f;
import javax.vecmath.Vector3d;

import com.sun.j3d.utils.geometry.Primitive;

import visual.metaphor.utils.AppearanceUtils;

/**
 * A dashed arrow used to represent monitor lock releases and notify() / notifyAll() method invocations in the final visualization.
 * 
 * @author pdo
 * 
 */
public class DashedArrow extends Primitive {

	// =========================== Attributes =================================

	/** The number of visible line segments per length unit. */
	private static final int SEGMENTS_PER_LENGTH_UNIT = 70;

	/** The default interspace between a pipe and its surrounding border. */
	private static final float BORDER_OFFSET = 0.0005f;

	/** The common {@link TransformGroup} instance to add all components of this dashed arrow to. */
	private final TransformGroup group = new TransformGroup();

	/** The solid appearance to be applied on all visible components of this dashed arrow. */
	private final Appearance app = AppearanceUtils.createReadWriteAppearance();

	/** The border appearance to be applied on all visible components of this dashed arrow. */
	private final Appearance borderApp = AppearanceUtils.createBorderAppearance();

	/** A list of all visible pipe components of this dashed arrow with their transform groups. */
	private final List<TransformGroup> visiblePipes = new LinkedList<TransformGroup>();

	// ========================== Constructor =================================

	/**
	 * Creates a new {@link DashedArrow} instance with the given diameter connecting the specified source with the given target point.
	 * 
	 * @param source
	 *            The source point of the dashed arrow to create. This will be the point in 3d space, where the arrow points from.
	 * @param target
	 *            The target point of the dashed arrow to create. This will be the point in 3d space, where the arrow points at.
	 * @param diameter
	 *            The diameter of the dashed arrow to create.
	 * @throws IllegalArgumentException
	 *             If either the source or target vector passed is {@code null} or the given diameter is negative.
	 */
	public DashedArrow(final Vector3d source, final Vector3d target, final float diameter) throws IllegalArgumentException {

		super();

		if (source == null || target == null) {
			throw new IllegalArgumentException("You cannot pass in a null vector.");
		}
		if (diameter < 0) {
			throw new IllegalArgumentException("A dashed arrow cannot have a negative diameter.");
		}

		this.setCollidable(true);
		this.setCapability(ALLOW_PARENT_READ);
		this.createComponents(source, target, diameter, true);
		this.addChild(this.group);
	}

	/**
	 * Creates a new {@link DashedArrow} instance with the given diameter connecting the specified source with the given target point.
	 * 
	 * @param source
	 *            The source point of the dashed arrow to create. This will be the point in 3d space, where the arrow points from.
	 * @param target
	 *            The target point of the dashed arrow to create. This will be the point in 3d space, where the arrow points at.
	 * @param diameter
	 *            The diameter of the dashed arrow to create.
	 * @param bend1
	 *            The first bendpoint of the dashed arrow to be created.
	 * @param bend2
	 *            The second bendpoint of the dashed arrow to be created.
	 * @throws IllegalArgumentException
	 *             If either the source, target, bend1, or bend2 vector passed is {@code null} or the given diameter is negative.
	 */
	public DashedArrow(final Vector3d source, final Vector3d target, final float diameter, final Vector3d bend1, final Vector3d bend2)
		throws IllegalArgumentException {

		this(bend1, bend2, diameter);

		if (source == null || target == null) {
			throw new IllegalArgumentException("You cannot pass in a null vector.");
		}

		this.createComponents(source, bend1, diameter, false);
		this.createComponents(bend2, target, diameter, false);
	}

	// ============================ Methods ===================================

	/**
	 * Creates all components (i.e. pipes and arrowhead) of this dashed arrow and adds them as child nodes to this instance.
	 * 
	 * @param source
	 *            The source point of the dashed arrow to create. This will be the point in 3d space, where the dashed arrow points from.
	 * @param target
	 *            The target point of the dashed arrow to create. This will be the point in 3d space, where the dashed arrow points at.
	 * @param diameter
	 *            The diameter of the dashed arrow components to create.
	 * @param directed
	 *            {@code True} to direct the dashed line (i.e. display an arrowhead), {@code false} otherwise.
	 */
	private void createComponents(final Vector3d source, final Vector3d target, final float diameter, final boolean directed) {

		if (directed) {
			// create arrow component
			final Vector3d connector = new Vector3d(target);
			connector.sub(source);
			connector.scale(0.42f);

			final Vector3d from = new Vector3d(source);
			from.add(connector);

			final Vector3d to = new Vector3d(target);
			to.sub(connector);

			final TransformGroup arrowTG = BendableArrowBuilder.createArrowComponent(from, to, diameter, this.app);
			this.visiblePipes.add(arrowTG);
			this.group.addChild(arrowTG);

			final TransformGroup arrowBorderTG = BendableArrowBuilder.createArrowComponent(from, to, diameter + BORDER_OFFSET * 2, this.app);
			final Arrow arrow = (Arrow) arrowBorderTG.getChild(0);
			arrow.getShape(Arrow.BOTTOM).setAppearance(this.borderApp);
			arrow.getShape(Arrow.TOP).setAppearance(this.borderApp);
			arrow.setAppearance(Arrow.ARROWHEAD, this.borderApp);
			arrow.setPickable(false);
			this.group.addChild(arrowBorderTG);

			// create pipe components
			this.createDashedPipes(source, from, diameter, false);
			this.createDashedPipes(to, target, diameter, true);
		} else {
			this.createDashedPipes(source, target, diameter, false);
		}
	}

	/**
	 * Creates a series of dashed pipes (i.e. a connected series of visible / invisible pipes) reaching from the given source to target point in space.
	 * 
	 * @param from
	 *            The point in space the created dashed pipe sequence is supposed to start.
	 * @param to
	 *            The point in space the created dashed pipe sequence is supposed to end.
	 * @param withInterspace
	 *            {@code True} to add one {@code 1 / #SEGMENTS_PER_LENGTH_UNIT} as interspace in front of the first dashed pipe element being placed. {@code False}
	 *            otherwise.
	 * @param diameter
	 *            The diameter of the pipes to create.
	 */
	private void createDashedPipes(final Vector3d from, final Vector3d to, final float diameter, final boolean withInterspace) {

		// compute length
		final Vector3d connector = new Vector3d(to);
		connector.sub(from);
		final double length = connector.length();
		final float increment = 2f / SEGMENTS_PER_LENGTH_UNIT;

		connector.normalize();
		connector.scale(increment / 2f);
		double curLength = 0;
		
		final Vector3d temp1 = new Vector3d(from);
		final Vector3d temp2 = new Vector3d(from);
		temp2.add(connector);

		// handle interspace
		if (withInterspace) {
			curLength = increment;
			temp1.add(connector);
			temp2.add(connector);
		}

		while (curLength + increment < length) {
			// create solid pipe
			TransformGroup pipeTG = BendableArrowBuilder.createPipeComponent(temp1, temp2, diameter, this.app);
			this.group.addChild(pipeTG);

			// add border
			pipeTG = BendableArrowBuilder.createPipeComponent(temp1, temp2, diameter + BORDER_OFFSET * 2, this.app);
			final Pipe pipe = (Pipe) pipeTG.getChild(0);
			pipe.getShape(Pipe.TOP).setAppearance(this.borderApp);
			pipe.getShape(Pipe.BOTTOM).setAppearance(this.borderApp);
			this.group.addChild(pipeTG);

			// update bendpoints
			temp1.add(connector);
			temp1.add(connector);
			temp2.add(connector);
			temp2.add(connector);
			curLength += increment;
		}

		if (curLength < length) {
			// connect remaining space
			final TransformGroup solidTG = BendableArrowBuilder.createPipeComponent(temp1, temp2, diameter, this.app);
			this.visiblePipes.add(solidTG);
			this.group.addChild(solidTG);

			// add border
			final TransformGroup pipeTG = BendableArrowBuilder.createPipeComponent(temp1, temp2, diameter + BORDER_OFFSET * 2, this.app);
			final Pipe pipe = (Pipe) pipeTG.getChild(0);
			pipe.getShape(Pipe.TOP).setAppearance(this.borderApp);
			pipe.getShape(Pipe.BOTTOM).setAppearance(this.borderApp);
			this.group.addChild(pipeTG);
		}
	}

	/**
	 * Returns the body {@link Shape3D} of the visible component with the specified part id or {@code null}, if no such component exists.
	 * 
	 * @param partId
	 *            The id of the visible component to get its body shape.
	 * @return The body shape of the specified component.
	 */
	@Override
	public Shape3D getShape(final int partId) {

		if (partId >= 0 && partId < this.visiblePipes.size()) {
			return ((Pipe) ((TransformGroup) this.visiblePipes.get(partId)).getChild(0)).getShape(Pipe.BODY);
		}

		return null;
	}

	/**
	 * Returns the {@link Appearance} of the visible component with the specified part id or {@code null}, if no such component exists.
	 * 
	 * @param partId
	 *            The id of the visible component to get its appearance.
	 * @return The appearance of the specified component.
	 */
	@Override
	public Appearance getAppearance(final int partId) {

		if (partId >= 0 && partId < this.visiblePipes.size()) {
			return ((Pipe) ((TransformGroup) this.visiblePipes.get(partId)).getChild(0)).getAppearance();
		}

		return null;
	}

	/**
	 * Sets the {@link Appearance} instance of all visible components of this dashed arrow.
	 * 
	 * @param appearance
	 *            The new appearance to set.
	 * @throws IllegalArgumentException
	 *             If the appearance passed in is {@code null}.
	 */
	@Override
	public void setAppearance(final Appearance appearance) throws IllegalArgumentException {

		if (appearance == null) {
			throw new IllegalArgumentException("The appearance passed in is null.");
		}

		for (TransformGroup transTG : this.visiblePipes) {
			((Pipe) transTG.getChild(0)).setAppearance(appearance);
		}
	}

	/**
	 * Sets the color of all visible components of this dashed arrow.
	 * 
	 * @param color
	 *            The color to be applied to all visible components.
	 * @throws IllegalArgumentException
	 *             If the color passed in is {@code null}.
	 */
	public void setColor(final Color color) throws IllegalArgumentException {

		if (color == null) {
			throw new IllegalArgumentException("You cannot set a null color.");
		}

		for (TransformGroup transTG : this.visiblePipes) {
			((Pipe) transTG.getChild(0)).getAppearance().getColoringAttributes().setColor(new Color3f(color));
		}
	}

}
