/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.metaphor;

import java.awt.Color;

import javax.media.j3d.Appearance;
import javax.vecmath.Color3f;

import com.sun.j3d.utils.geometry.Cylinder;

import visual.metaphor.utils.AppearanceUtils;

/**
 * A regular {@link Cylinder} with pick reporting and read/write access to its {@link Appearance} and {@link ColoringAttributes} enabled.
 * 
 * @author pdo
 * 
 */
public class Pipe extends Cylinder {

	// ========================== Constructors ================================

	/**
	 * Constructs a new {@link Pipe} instance with a given radius and height.
	 * 
	 * @param diameter
	 *            The diameter of the new pipe.
	 * @param length
	 *            The length of the new pipe.
	 * @throws IllegalArgumentException
	 *             If either the given radius or length value is negative.
	 */
	public Pipe(final float diameter, final float length) throws IllegalArgumentException {

		super(diameter / 2, length, GENERATE_NORMALS | GENERATE_TEXTURE_COORDS, AppearanceUtils.createReadWriteAppearance());

		if (diameter < 0 || length < 0) {
			throw new IllegalArgumentException("You cannot pass in negative dimension sizes.");
		}

		super.getAppearance().setCapability(ENABLE_PICK_REPORTING);
		this.setCapability(ALLOW_PARENT_READ);
		this.setCollidable(true);
	}

	/**
	 * Constructs a new {@link Pipe} instance with a given radius, height, and color.
	 * 
	 * @param radius
	 *            The radius of the new pipe.
	 * @param length
	 *            The length of the new pipe.
	 * @param color
	 *            The color of the new pipe.
	 * @throws IllegalArgumentException
	 *             If the given radius or length value is negative or the color passed in is {@code null}.
	 */
	public Pipe(final float radius, final float length, final Color color) throws IllegalArgumentException {

		this(radius, length);
		this.getAppearance().getColoringAttributes().setColor(new Color3f(color));
	}

	// ============================ Methods ===================================

	/**
	 * Set the {@link Color} of this pipe.
	 * 
	 * @param color
	 *            The new color of this pipe.
	 * @throws IllegalArgumentException
	 *             If the color passed in is {@code null}.
	 */
	public void setColor(final Color color) throws IllegalArgumentException {

		if (color == null) {
			throw new IllegalArgumentException("The color passed in is null.");
		}
		this.getAppearance().getColoringAttributes().setColor(new Color3f(color));
		for (int i = BODY; i <= BOTTOM; i++) {
			super.getShape(i).getAppearance().getColoringAttributes().setColor(new Color3f(color));
		}
	}

	/**
	 * Sets the {@link Appearance} instance of this pipe.
	 * 
	 * @param app
	 *            The new appearance to set.
	 * @throws IllegalArgumentException
	 *             If the appearance passed in is {@code null}.
	 */
	public void setAppearance(final Appearance app) throws IllegalArgumentException {

		if (app == null) {
			throw new IllegalArgumentException("The appearance passed in is null.");
		}
		super.setAppearance(app);
	}

}
