/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.metaphor;

import java.awt.Color;

import javax.media.j3d.Appearance;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Texture;
import javax.media.j3d.TextureAttributes;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Color4f;
import javax.vecmath.Vector3d;


import com.sun.j3d.utils.geometry.Cylinder;
import com.sun.j3d.utils.geometry.Primitive;
import com.sun.j3d.utils.image.TextureLoader;

import configuration.FileSystem;


import visual.metaphor.utils.AppearanceUtils;
import visual.metaphor.utils.TransformUtils;

/**
 * A symbol sign to uniquely identify the thread, monitor, and wait-/notify building in the visualization.
 * 
 * @author pdo
 * 
 */
public class SymbolSign extends Primitive {

	// =========================== Attributes =================================

	/** The default length of this {@link SymbolSign}. */
	private static final float LENGTH = 0.01f;

	/** The default {@link Color} of this {@link SymbolSign}. */
	private static final Color COLOR = Color.WHITE;
	
	/** The pipe used to model the background of this symbol sign. */
	private final Pipe background;

	/** The pipe used to model the sign of this symbol sign itself. */
	private final Pipe sign;

	// ========================== Constructor =================================

	/**
	 * Creates a new {@link SymbolSign} instance located on the given position with the specified symbol on its top and bottom shape.
	 * 
	 * @param position
	 *            The position of this symbol sign as {@link Vector3d} object.
	 * @param radius
	 *            The radius of the symbol sign to create.
	 * @param symbol
	 *            The {@link Symbol} to be displayed on this symbol sign.
	 * @throws IllegalArgumentException
	 *             If either the position vector or the symbol passed in is {@code null} or the radius is not positive.
	 */
	public SymbolSign(final Vector3d position, final float radius, final Symbol symbol) throws IllegalArgumentException {

		super();

		if (position == null) {
			throw new IllegalArgumentException("You cannot pass in a null position");
		}
		if (radius <= 0) {
			throw new IllegalArgumentException("A symbol sign cannot have a non-positive radius.");
		}
		if (symbol == null) {
			throw new IllegalArgumentException("You cannot set a null symbol");
		}

		// create background
		this.background = new Pipe(radius - 0.001f, LENGTH - 0.001f, COLOR);

		// create sign
		this.sign = new Pipe(radius, LENGTH, COLOR);
		this.sign.setAppearance(Cylinder.BODY, AppearanceUtils.createBorderAppearance());

		this.setCapability(ALLOW_PARENT_READ);
		this.setSymbol(symbol);

		final TransformGroup transform = TransformUtils.createTransformGroup(position, TransformUtils.computeRotationMatrix(new Vector3d(1, 0, 0)));
		transform.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		transform.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

		transform.addChild(this.background);
		transform.addChild(this.sign);
		
		// align sign by rotation
		final Transform3D trans = new Transform3D();
		transform.getTransform(trans);

		final Transform3D transZ = new Transform3D();
		transZ.setRotation(TransformUtils.computeRotationMatrix(new Vector3d(0, 0, 1)));

		final Transform3D transX = new Transform3D();
		transX.setRotation(TransformUtils.computeRotationMatrix(new Vector3d(1, 0, 0)));
		
		trans.mul(transZ);
		trans.mul(transX);
		transform.setTransform(trans);
		this.addChild(transform);
	}

	// ============================ Methods ===================================

	/**
	 * Sets the symbol of this symbol sign. It will display the icon mapped to the specified symbol on both sides (front and back) of this sign instance.
	 * 
	 * @param symbol
	 *            The symbol to be displayed.
	 * @throws IllegalArgumentException
	 *             If the symbol passed in is {@code null} or is neither {@link Symbol#THREAD}, nor {@link Symbol#MONITOR} nor {@link Symbol#WAIT_NOTIFY}.
	 */
	private void setSymbol(final Symbol symbol) throws IllegalArgumentException {

		if (symbol == null) {
			throw new IllegalArgumentException("You cannot set a null symbol.");
		}

		switch (symbol) {
		case THREAD:
			this.setTexture(FileSystem.getInstance().getIconsFolder() + "Pfeil.jpg");
			break;
		case MONITOR:
			this.setTexture(FileSystem.getInstance().getIconsFolder() + "Vorhaengeschloss.jpg");
			break;
		case WAIT_NOTIFY:
			this.setTexture(FileSystem.getInstance().getIconsFolder() + "Hand.jpg");
			break;
		default:
			throw new IllegalArgumentException("You passed in an unsupported symbol." + symbol.name() + ".");
		}
	}

	/**
	 * Applies the texture given by the specified filename to this {@link SymbolSign}.
	 * 
	 * @param filename
	 *            The name of the image file to set as texture.
	 */
	private void setTexture(final String filename) {

		final Appearance topApp = AppearanceUtils.createReadWriteAppearance();
		final Appearance bottomApp = AppearanceUtils.createReadWriteAppearance();
		final Texture tex = new TextureLoader(filename, null).getTexture();

		// add front texture
		TextureAttributes texAtts = new TextureAttributes(TextureAttributes.REPLACE, new Transform3D(), new Color4f(), TextureAttributes.NICEST);
		topApp.setTextureAttributes(texAtts);
		topApp.setTexture(tex);
		this.sign.setAppearance(Cylinder.TOP, topApp);

		// add back texture		
		final Transform3D trans = new Transform3D();
		trans.setRotation(TransformUtils.computeRotationMatrix(new Vector3d(0, -1, 0)));
		texAtts = new TextureAttributes(TextureAttributes.REPLACE, trans, new Color4f(), TextureAttributes.NICEST);

		bottomApp.setTextureAttributes(texAtts);
		bottomApp.setTexture(tex);
		this.sign.setAppearance(Cylinder.BOTTOM, bottomApp);
	}

	@Override
	public Shape3D getShape(final int partId) {

		return this.background.getShape(partId);
	}

	@Override
	public Appearance getAppearance(final int partId) {

		return this.background.getAppearance(partId);
	}

	@Override
	public void setAppearance(final Appearance app) {

		this.background.setAppearance(app);
	}

	/**
	 * An enumeration of all symbols currently displayable on a {@link SymbolSign}.
	 * 
	 * @author pdo
	 * 
	 */
	public static enum Symbol {

		/** The symbol representing a thread building. */
		THREAD,

		/** The symbol representing a wait notify building. */
		WAIT_NOTIFY,
		
		/** The symbol representing a monitor building. */
		MONITOR,
	}
}
