/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.metaphor.ambience;

import java.awt.Color;

import javax.media.j3d.AmbientLight;
import javax.media.j3d.BoundingSphere;
import javax.media.j3d.DirectionalLight;
import javax.media.j3d.Material;
import javax.vecmath.Color3f;
import javax.vecmath.Point3d;

/**
 * A utility class creating special ambient and directional lights to make the visualization nice and shiny.
 * 
 * @author pdo
 * 
 */
public final class Lightning {

	// =========================== Attributes =================================

	/** The default bounding point used for each light object created by any method in this class. */
	private static final Point3d BOUNDING_POINT = new Point3d(0, 0, 0);

	/** The ambient color used in {@link #DEFAULT_MATERIAL}. */
	private static final Color3f AMBIENT_COLOR = new Color3f(Color.WHITE);

	/** The diffuse color used in {@link #DEFAULT_MATERIAL}. */
	private static final Color3f DIFFUSE_COLOR = new Color3f(Color.WHITE);

	/** The emissive color used in {@link #DEFAULT_MATERIAL}. */
	private static final Color3f EMISSIVE_COLOR = new Color3f(Color.WHITE);

	/** The spectacular color used in {@link #DEFAULT_MATERIAL}. */
	private static final Color3f SPECTACULAR_COLOR = new Color3f(Color.WHITE);

	/** The shininess setting used in {@link #DEFAULT_MATERIAL}. */
	private static final float SHININESS = 1;

	/** The default {@link Material} to be applied to any Java 3D object which interferes with lightning. It defines their behavior when light shines on them. */
	private static final Material DEFAULT_MATERIAL = new Material(AMBIENT_COLOR, EMISSIVE_COLOR, DIFFUSE_COLOR, SPECTACULAR_COLOR, SHININESS);

	/** The default radius used for each light object created by any method in this class. */
	private static final double RADIUS = 100;

	// ========================== Constructor =================================

	/**
	 * Private constructor. This class is not intended for instantiation, since it contains static methods only.
	 */
	private Lightning() {}

	// ============================ Methods ===================================

	/**
	 * Creates a new {@link AmbientLight} instance with a bounding point as defined in {@link #BOUNDING_POINT}, a radius of {@value #RADIUS} and given color.
	 * 
	 * @param color
	 *            The color of the ambient light to create.
	 * @return A ambient light instance with given color.
	 * @throws IllegalArgumentException
	 *             If the color passed in is {@code null}.
	 */
	public static AmbientLight createAmbienceLight(final Color color) throws IllegalArgumentException {

		if (color == null) {
			throw new IllegalArgumentException("You may not pass in a null color.");
		}

		final AmbientLight light = new AmbientLight();
		light.setInfluencingBounds(new BoundingSphere(BOUNDING_POINT, RADIUS));
		light.setColor(new Color3f(color));
		return light;
	}

	/**
	 * Creates a new {@link DirectionalLight} instance with a bounding point as defined in {@link #BOUNDING_POINT}, a radius of {@value #RADIUS}, and given
	 * {@link Color}.
	 * 
	 * @param color
	 *            The color of the directional light to create.
	 * @return A directional light instance with given color.
	 * @throws IllegalArgumentException
	 *             If the color passed in is {@code null}.
	 */
	public static DirectionalLight createDirectionalLight(final Color color) throws IllegalArgumentException {

		if (color == null) {
			throw new IllegalArgumentException("You may not pass in a null color.");
		}

		final DirectionalLight light = new DirectionalLight();
		light.setInfluencingBounds(new BoundingSphere(BOUNDING_POINT, RADIUS));
		light.setColor(new Color3f(color));
		return light;
	}

	/**
	 * Returns the default {@link Material} which is supposed to applied to any Java3D node instance which interferes with lightning.
	 * 
	 * @return The default material as specified in {@link #DEFAULT_MATERIAL}.
	 */
	public static Material getDefaultMaterial() {

		return DEFAULT_MATERIAL;
	}

}
