/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.metaphor.utils;

import java.awt.Color;

import javax.media.j3d.Appearance;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.LineAttributes;
import javax.media.j3d.PolygonAttributes;
import javax.media.j3d.Shape3D;
import javax.media.j3d.TransparencyAttributes;
import javax.vecmath.Color3f;

import configuration.VisualProperties;



/**
 * A utility class offering methods to create commonly used {@link Appearance}s.
 * 
 * @author pdo
 * 
 */
public final class AppearanceUtils {
	
	// ========================== Attributes ==================================
	
	/** The {@link Color} of the bordering lines. */
	private static Color borderColor = VisualProperties.getInstance().getBorderColor();

	// ========================== Constructor =================================

	/** Private constructor. This class should not be instantiated since it contains static methods only. */
	private AppearanceUtils() {}

	// ============================ Methods ===================================
	
	/**
	 * Create an {@link Appearance} instance, which causes a {@link Primitive} object to visualize its outer edges only, i.e. its faces will be totally invisible.
	 * 
	 * @return An appearance witch causes a primitive to show its edges only.
	 */
	public static Appearance createBorderAppearance() {

		// define appearance
		final Appearance appearance = new Appearance();
		appearance.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
		appearance.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
		appearance.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_READ);
		appearance.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);

		// define polygon attributes
		final PolygonAttributes polygonAttributes = new PolygonAttributes();
		polygonAttributes.setPolygonMode(PolygonAttributes.POLYGON_LINE);
		polygonAttributes.setCullFace(PolygonAttributes.CULL_NONE);
		appearance.setPolygonAttributes(polygonAttributes);

		// define coloring attributes
		final ColoringAttributes colorAttributes = new ColoringAttributes();
		colorAttributes.setColor(new Color3f(borderColor));
		colorAttributes.setShadeModel(ColoringAttributes.NICEST);
		appearance.setColoringAttributes(colorAttributes);

		// define line attributes
		final LineAttributes lineAttributes = new LineAttributes();
		lineAttributes.setLineWidth(VisualProperties.getInstance().getBorderWidth());
		lineAttributes.setLinePattern(LineAttributes.PATTERN_SOLID);
		appearance.setLineAttributes(lineAttributes);

		// add material
		appearance.setCapability(Appearance.ALLOW_MATERIAL_READ);
		appearance.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		//appearance.setMaterial(Lightning.getDefaultMaterial());

		// allow transparency read / write
		appearance.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
		appearance.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);

		return appearance;
	}

	/**
	 * Create an {@link Appearance} instance, which grants read / write access to itself and to its {@link ColoringAttributes}.
	 * 
	 * @return An appearance with read / write access enabled.
	 */
	public static Appearance createReadWriteAppearance() {
		
		// define appearance
		final Appearance appearance = new Appearance();
		appearance.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_READ);
		appearance.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
		appearance.setCapability(Appearance.ALLOW_TEXTURE_ATTRIBUTES_READ);
		appearance.setCapability(Appearance.ALLOW_TEXTURE_ATTRIBUTES_WRITE);

		// define coloring attributes
		final ColoringAttributes coloring = new ColoringAttributes();
		coloring.setShadeModel(ColoringAttributes.NICEST);
		coloring.setCapability(ColoringAttributes.ALLOW_COLOR_READ);
		coloring.setCapability(ColoringAttributes.ALLOW_COLOR_WRITE);
		appearance.setColoringAttributes(coloring);

		// add material
		appearance.setCapability(Appearance.ALLOW_MATERIAL_READ);
		appearance.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		//appearance.setMaterial(Lightning.getDefaultMaterial());

		// allow transparency read / write
		appearance.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
		appearance.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);

		return appearance;
	}

	/**
	 * Creates an {@link Appearance} instance, which equals those {@link #createReadWriteAppearance()} creates, but adds {@link TransparencyAttributes}, which cause
	 * the object, to which this appearance is applied, to be totally invisible (i.e. has maximum transparency).
	 * 
	 * @return An appearance with read / write access enabled and maximum transparency set.
	 */
	public static Appearance createTransparentAppearance() {

		final Appearance appearance = createReadWriteAppearance();
		appearance.setMaterial(null);

		// define transparency attributes
		final TransparencyAttributes transparancy = new TransparencyAttributes();
		transparancy.setTransparency(1); // totally transparent, makes the object, to which this appearance is applied, completely invisible
		transparancy.setTransparencyMode(TransparencyAttributes.FASTEST); // here, we do not need quality, since the object is not visible anyways
		appearance.setTransparencyAttributes(transparancy);

		return appearance;
	}
	
	/**
	 * Reloads the {@link #borderColor} attribute of this class.
	 */
	public static void reloadAttributes() {
		
		borderColor = VisualProperties.getInstance().getBorderColor();
	}
}
