/***************************************************************************
 * This file is part of Synchrovis - A prototypical implementation of the 
 * visualization approach described in the author's master's thesis titled 
 * 			
 * 							 Visualisierung von 
 * 					Synchronisationspunkten in Kombination 
 * 						  mit der Statik und Dynamik 
 *							 eines Softwaresystems
 *
 * written at Software Engineering Group, Department of Computer Science at 
 * Christian-Albrechts-University of Kiel.
 * 
 * Author: Philipp Dhring, October 2012
 ***************************************************************************/

package visual.metaphor.utils;

import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Matrix3d;
import javax.vecmath.Point3d;
import javax.vecmath.SingularMatrixException;
import javax.vecmath.Vector3d;

import org.apache.log4j.Logger;

/**
 * A utility class offering commonly used transformation related methods.
 * 
 * @author pdo
 * 
 */
public final class TransformUtils {

	// =========================== Attributes =================================

	/** A logger to log relevant actions in this class. */
	private static final Logger LOGGER = Logger.getLogger(TransformUtils.class);

	// ========================== Constructor =================================

	/** Private constructor. This class should not be instantiated since it contains static methods only. */
	private TransformUtils() {}

	// ============================ Methods ===================================

	/**
	 * Computes the rotation matrix equivalent to the given direction vector assuming that visual element to be rotated is currently oriented in direction (0,1,0).
	 * 
	 * @param direction
	 *            The direction to compute the equivalent rotation matrix for.
	 * @return The rotation matrix equivalent to the given direction.
	 * @throws IllegalArgumentException
	 *             If the given direction vector is {@code null}.
	 * @throws SingularMatrixException
	 *             If the computed rotation matrix could not be inverted.
	 */
	public static Matrix3d computeRotationMatrix(final Vector3d direction) throws IllegalArgumentException, SingularMatrixException {

		if (direction == null) {
			throw new IllegalArgumentException("Given direction vector is null.");
		}

		final Vector3d yDim = (Vector3d) direction.clone();
		yDim.normalize();

		Vector3d xDim;
		if (yDim.getX() == 0 && yDim.getZ() == 0) {
			xDim = new Vector3d(-yDim.getY(), 0f, 0f);
		} else {
			xDim = new Vector3d(0f, 1f, 0f);
		}

		final Vector3d zDim = new Vector3d();
		zDim.cross(xDim, yDim);
		zDim.normalize();

		xDim.cross(zDim, yDim);
		xDim.normalize();
		xDim.negate();

		final Matrix3d rotation = new Matrix3d(xDim.getX(), xDim.getY(), xDim.getZ(), yDim.getX(), yDim.getY(), yDim.getZ(), zDim.getX(), zDim.getY(), zDim.getZ());

		try {
			rotation.invert();
		} catch (SingularMatrixException e) {
			if (LOGGER.isDebugEnabled()) {
				System.out.println("SingularMatrixException");
				LOGGER.debug("Singular matrix detected: " + rotation.toString() + ". Unable to compute inverse matrix.", e);
			}
		}

		return rotation;
	}

	/**
	 * Creates a new {@link TransformGroup} performing the rotations specified by the given rotation angles.
	 * 
	 * @param rotX
	 *            The rotation around the x-axis to perform.
	 * @param rotY
	 *            The rotation around the y-axis to perform.
	 * @param rotZ
	 *            The rotation around the z-axis to perform.
	 * @return A transform group performing the specified rotations.
	 */
	public static TransformGroup createTransformation(final double rotX, final double rotY, final double rotZ) {

		final Transform3D transform = new Transform3D();
		final Transform3D transHelp1 = new Transform3D();
		final Transform3D transHelp2 = new Transform3D();

		transform.rotX(rotX);
		transHelp1.rotY(rotY);
		transHelp2.rotZ(rotZ);

		transform.mul(transHelp1);
		transform.mul(transHelp2);

		return new TransformGroup(transform);
	}

	/**
	 * Instantiates a {@link TransformGroup} making any visual object on which this group is applied to point from the {@code source} to the {@code target} point in
	 * space. I.e. the object's bottom end will start at {@code source} and its top will point and and at the given {@code target} coordinate.
	 * 
	 * @param source
	 *            The point in space a visual object this transform group is applied to will start.
	 * @param target
	 *            The point in space a visual object this transform group is applied to will point at.
	 * @return A transform group making any object go from the given {@code source} to the {@code target} point.
	 * @throws IllegalArgumentException
	 *             If at least one vector passed in is {@code null}.
	 */
	public static TransformGroup computeTransformation(final Vector3d source, final Vector3d target) throws IllegalArgumentException {

		if (source == null || target == null) {
			throw new IllegalArgumentException("You may not pass in null vectors.");
		}

		// computing length
		final Vector3d diff = new Vector3d();
		diff.sub(target, source);

		// computing position
		final Vector3d pos = new Vector3d();
		pos.scaleAdd(0.5, diff, source);

		// creating transform group
		final Matrix3d rot = TransformUtils.computeRotationMatrix(diff);
		final TransformGroup trans = TransformUtils.createTransformGroup(pos, rot);

		return trans;
	}

	/**
	 * Creates a new {@link TransformGroup} instance performing the given translation and rotation on any object this group is applied to. I.e. it shifts an attached
	 * object to the defined point in space and rotates it using the given rotation matrix.
	 * 
	 * @param translation
	 *            The translation vector to add to the transform group being created.
	 * @param rotation
	 *            The rotation matrix to add to the transform group being created.
	 * @return A new transform group performing the given translation and rotation.
	 * @throws IllegalArgumentException
	 *             If at least one argument is {@code null}.
	 */
	public static TransformGroup createTransformGroup(final Vector3d translation, final Matrix3d rotation) throws IllegalArgumentException {

		if (translation == null) {
			throw new IllegalArgumentException("The given translation vector is null.");
		}
		if (rotation == null) {
			throw new IllegalArgumentException("The given rotation matrix is null.");
		}

		final Transform3D trans = new Transform3D();
		trans.setTranslation(translation);
		trans.setRotation(rotation);
		return new TransformGroup(trans);
	}

	/**
	 * Creates a new {@link TransformGroup} instance performing the given translation on any object this group is applied to. I.e. it shifts an attached object to
	 * the defined point in space.
	 * 
	 * @param translation
	 *            The translation vector to add to the transform group being created.
	 * @return A new transform group performing the given translation and rotation.
	 * @throws IllegalArgumentException
	 *             If the given translation is {@code null}.
	 */
	public static TransformGroup createTransformGroup(final Vector3d translation) throws IllegalArgumentException {

		if (translation == null) {
			throw new IllegalArgumentException("The given translation vector is null.");
		}

		final Transform3D trans = new Transform3D();
		trans.setTranslation(translation);
		return new TransformGroup(trans);
	}

	/**
	 * Extracts the rotation angles set in the given {@link Transform3D} instance as {@link Point3d} object for the purpose of this documentation named {@code rot},
	 * in which {@code rot.x} equals the rotation around the x-axis, {@code rot.y} the rotation around the y-axis, and {@code rot.z} the rotation around the z-axis.
	 * 
	 * @param transform
	 *            The transform3d object to extract the rotation angles from.
	 * @return The rotation angles set in the given transform3D object.
	 * @throws IllegalArgumentException
	 *             If the given transform3D object is {@code null}.
	 */
	public static Point3d extractRotationAngles(final Transform3D transform) throws IllegalArgumentException {

		if (transform == null) {
			throw new IllegalArgumentException("The given transform3d object is null.");
		}

		// extract rotation angles
		final Point3d rotation = new Point3d();

		final Matrix3d trans = new Matrix3d();
		transform.get(trans);

		rotation.y = -Math.asin(trans.getElement(2, 0));
		final double cosY = Math.cos(rotation.y);

		if (rotation.y == 0) {
			rotation.x = 0.0;
			rotation.z = -Math.atan2(trans.getElement(0, 1), trans.getElement(1, 1));
		} else {
			rotation.x = -Math.atan2(-trans.getElement(2, 1) / cosY, trans.getElement(2, 2) / cosY);
			rotation.z = -Math.atan2(-trans.getElement(1, 0) / cosY, trans.getElement(0, 0) / cosY);
		}

		return rotation;
	}
}
